-- -------------------------------------------------------------------------- --
-- CharScanner by kunda                                                       --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- CharScanner scans Basic Character Data, Talent Specifications, Equipment,  --
-- Achievements and Boss Kill Statistics from target unit or group and shows  --
-- this data in a movable frame. Any scanned unit data is temporarily saved   --
-- (in memory not in SavedVariables) and can be viewed or deleted any time.   --
--                                                                            --
-- ONE click and ONE look shows the most important character data!            --
--                                                                            --
-- CharScanner does not rate, score or measure!                               --
-- CharScanner simply displays available ingame information in a nice way.    --
-- CharScanner GUI: Simple, distinctive, clearly recognizable design.         --
--                                                                            --
-- What you get in ~0.300 Seconds:                                            --
--  - A complete Talent scan (primary and secondary talent specification).    --
--  - A complete Gear scan.                                                   --
--  - A dungeon and raid achievement scan.                                    --
--  - A dungeon and raid statistic scan.                                      --
-- What you get in 10 Seconds if you scan your 25 player group:               --
--  - All from above.                                                         --
--  - A nice Compare Tool.                                                    --
--                                                                            --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- NOTES:                                                                     --
-- - The scanned unit MUST BE in inspect range!                               --
-- - There is no 'Inspect Glyphs API'! :-(                                    --
-- - Maximum time for one scan is: 1 Second (default).                        --
--    ! Change this value in the options menu if you have a high latency or   --
--      if you get a 'Scan incomplete!' status message if you scan yourself.  --
-- - GroupScan:                                                               --
--    - Raid roster changes between a 'GroupScan'-Button click and end of     --
--      groupscan are discarded!                                              --
--    - GroupScan will stop if:                                               --
--        All group/raid members has been scanned. or                         --
--        You enter combat! or                                                --
--        You click on: 'STOP SCAN'                                           --
-- - Mainrole: (TANK check)                                                   --
--    - DEATHKNIGHT TANK: >=10 points in Blade Barrier+Toughness+Anticipation.--
--    - DRUID TANK: >=4 points in Thick Hide+Protector of the Pack.           --
--    Sure, this is not perfect, but all Deathknight or Druid Tanks I scanned --
--    in 25 player raids used this talent combinations.                       --
-- - LDB is supported.                                                        --
--    But: LibStub, CallbackHandler-1.0 & LibDataBroker-1.1 is not included.  --
--         This is intended and I will not change this!                       --
-- - Entering combat or entering/leaving an instance or zoning (with loading  --
--   screen) will stop any scan.                                              --
-- - Memory usage:                                                            --
--    - Base memory usage without a scan: ~610KB                              --
--    - Base memory with populated class talent data: ~940KB (~33KB each)     --
--    - One unit: ~30KB (incl. data for Compare Tool)                         --
--    = 25 scanned units: ~1690KB                                             --
--    CharScanner is designed for fast temporary scans.                       --
--    Delete the data you do not need anymore!                                --
-- - Scan duration:                                                           --
--    - This is mainly latency dependent!                                     --
--      My tests:                                                             --
--      - ~0.300 Seconds for one unit                                         --
--      - A group with 25 player is normally scanned in less than 10 Seconds  --
--        if all members are in inspect range.                                --
--      = This is really really fast, because each scan requires to wait      --
--        until 2 events trigger:                                             --
--        INSPECT_TALENT_READY and INSPECT_ACHIEVEMENT_READY                  --
--                                                                            --
-- TODO:                                                                      --
--  - Shitty Gem Scan.                                                        --
--                                                                            --
-- KNOWN BUGS:                                                                --
-- - statistic id 4046/4045 (Anub'arak)                                       --
--   4046 'Times completed the Trial of the Crusader (25 player)'             --
--   also counts for                                                          --
--   4045 'Times completed the Trial of the Grand Crusader (10 player)'       --
--   -> incorrect return value from GetComparisonStatistic()/GetStatistic()   --
--   -> check workaround in BossStatistic()                                   --
--   This is a bug in WoW Client 3.3.0.11159!                                 --
--                                                                            --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- CharScanner will never ever save scanned data to SavedVariables.           --
--  - Most data will change everyday, so it's completely useless to save such --
--    data to SavedVariables.                                                 --
--    The sheer amount of data makes such a feature a no go.                  --
-- CharScanner will never ever autoscan a character or a group.               --
--  - You must actively initialize a single or group scan by clicking a       --
--    button.                                                                 --
--    Triggering INSPECT_TALENT_READY or INSPECT_ACHIEVEMENT_READY without    --
--    user interaction is very bad and can cause massive lags.                --
-- CharScanner will never ever use SendAddonMessage()!!!                      --
--  - I hate SendAddonMessage()!                                              --
--  - EVERYBODY can fake any message that is send via SendAddonMessage().     --
--  - A Messages from SendAddonMessage() is ALWAYS! an untrustable source!    --
--  - There is only one trustable source: The WoW server!                     --
--                                                ...ok, never say never ;)...--
--                                                                            --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- - slash commands: /cs - /scan - /charscanner                               --
--          example: /scan         : CharScanner GUI & SINGLESCAN             --
--          example: /scan options : CharScanner Options                      --
--                                                                            --
-- -------------------------------------------------------------------------- --

-- ---------------------------------------------------------------------------------------------------------------------
CharScanner = CreateFrame("Frame")    -- container
CharScanner_Options = {}              -- SavedVariable options table

local _G                           = _G
local type                         = _G.type
local mod                          = _G.mod
local pairs                        = _G.pairs
local select                       = _G.select
local tonumber                     = _G.tonumber
local math_floor                   = _G.math.floor
local string_sub                   = _G.string.sub
local GetTalentTabInfo             = _G.GetTalentTabInfo
local GetTalentInfo                = _G.GetTalentInfo
local GetTalentLink                = _G.GetTalentLink
local GetStatistic                 = _G.GetStatistic
local GetComparisonStatistic       = _G.GetComparisonStatistic
local GetAchievementComparisonInfo = _G.GetAchievementComparisonInfo
local GetAchievementInfo           = _G.GetAchievementInfo
local GetCategoryNumAchievements   = _G.GetCategoryNumAchievements
local UnitGUID                     = _G.UnitGUID
local GetTime                      = _G.GetTime

local L = CharScanner_Locales         -- localization table
local DATA = CharScanner_Data         -- achievement & statistic data table

local mult2 = 10^(2 or 0)             -- x.xx
local mult4 = 10^(4 or 0)             -- x.xxxx

local AchievementIDTable = {}         -- Achievement ID->name table (build after event PLAYER_LOGIN triggered)
local StatisticIDTable = {}           -- Statistic ID->name table (build after event PLAYER_LOGIN triggered)
local inCombat = false                -- false = out of combat | true = in combat
local inCombatGrpChk = false          -- party/raid member check if inCombat invite
local inPartyOrRaid = false           -- false = NOT in party or raid | true = in party or raid
local rosterhack = 0                  -- execute RAID_ROSTER_UPDATE block only if GetNumRaidMembers() has changed
local bugfix                          -- -> workaround for achievementFunctions.selectedCategory bug (Blizzard_AchievementUI.lua)
local memusage = false                -- ugly -for testing only-

local DataStep = 0                    -- upvalue for the different information sources (eg. events)
local SingleScanIsActive = false      -- false = SingleScan is NOT active | true = SingleScan is active
local SingleScanTimerStart = 0        -- Timer (single scan) start (time for one scan)
local GroupScanIsActive = false       -- false = GroupScan is NOT active | true = GroupScan is active
local GroupScanTimerStart = 0         -- Timer (group scan) start (time for whole group scan)
local GroupScanNext = false           --

local GUI_menu_values = {}                -- === Achievement/Statistic collapse menu values
local GUI_history_data = {}               -- === History (scanned units) data
local GUI_history_maxNum = 25             -- max. scroll frame size. if GUI_history_maxNum > 25 = scrollframe 
local GUI_history_Count = 0               -- count
local GUI_history_SortBy = 1              -- sort by: 1 = class | 2 = name | 3 = time | 4 = mainrole | 5 = talents1 | 6 = talents2 | 7 = ItemLevel | 8 = Achievement | 9 = Boss | 10 = Kills
local GUI_history_SortOrder = 1           -- sort order: 0 = asc | 1 = desc
local GUI_history_RaidL1 = 0              -- raid level1
local GUI_history_RaidL2                  -- raid level2
local GUI_history_RaidL3                  -- raid level3
local GUI_history_Role = {}               -- role
local GUI_achievementdetail_data = {}     -- === Achievement Detail data
local GUI_achievementdetail_maxNum = 25   -- max. scroll frame size. if GUI_achievementdetail_maxNum > 25 = scrollframe 
local GUI_achievementdetail_Count = 0     -- count
local GUI_achievementdetail_SortBy = 1    -- sort by: 1 = id | 2 = name
local GUI_achievementdetail_SortOrder = 1 -- sort order: 0 = asc | 1 = desc
local GUI_achievementdetail_L1            -- level1
local GUI_achievementdetail_L2            -- level2
local GUI_achievementdetail_L3            -- level3
local GUI_achievementdetail_Select = 1    -- 1 = ALL | 2 = COMPLETE | 3 = INCOMPLETE

local GUIselectedGUID                 -- current shown char
local NumScannedChars = 0             -- number of scanned characters
local playerGUID = ""                 -- current character GUID (set after event PLAYER_LOGIN triggered)
local IsUnitInpected = false          -- scanned unit: true: "Inspect" | false: "Character" (it's me)
local SCANGUID                        -- current scan GUID
local SCAN = {}                       -- scanned unit data (key = GUID)
local LatestScanIsComplete            -- true = last scan is complete | false/nil = last scan is NOT complete

local TalentTreeFrameIsShown          -- for mouseover
local GearFrameIsShown                -- for mouseover
local HistoryFrameIsShown             --
local CompareFrameIsShown             --
local AchievementDetailFrameIsShown   --

local myMAX_TIER = 11                 -- max. tier
local myMAX_COLUMN = 4                -- Talent Tree column
local myMAX_TALENT_TABS = 3           -- Blizzard default is 3 too 
local myMAX_NUM_TALENTS = 44          -- Blizzard default is 40 (how strange) -> we use 44, because: 11 (tier) * 4 (column) = 44 (button)

local Textures = {
	CharScanner       = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-button.tga"}, -- Textures.CharScanner.path
	CharScannerHL     = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-button-highlight.tga"}, -- Textures.CharScannerHL.path
	CharScannerGUI    = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-button-GUI.tga"}, -- Textures.CharScannerGUI.path
	MinimapButton     = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.MinimapButton.path
	                     coords  = {96/128, 128/128, 0/64, 32/64},
	                     width   = 32,
	                     height  = 32},
	SingleScan        = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.SingleScan.path
	                     coords  = {55/128, 86/128, 19/64, 36/64},
	                     width   = 31,
	                     height  = 17},
	GroupScan         = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.GroupScan.path
	                     coords  = {1/128, 43/128, 37/64, 54/64},
	                     width   = 42,
	                     height  = 17},
	GroupScanAdd      = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.GroupScanAdd.path
	                     coords  = {55/128, 86/128, 1/64, 18/64},
	                     width   = 31,
	                     height  = 17},
	Delete            = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.Delete.path
	                     coords  = {1/128, 18/128, 19/64, 36/64},
	                     width   = 17,
	                     height  = 17},
	History           = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.History.path
	                     coords  = {1/128, 18/128, 1/64, 18/64},
	                     width   = 17,
	                     height  = 17},
	Compare           = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.Compare.path
	                     coords  = {19/128, 36/128, 1/64, 18/64},
	                     width   = 17,
	                     height  = 17},
	Options           = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.Options.path
	                     coords  = {44/128, 64/128, 37/64, 54/64},
	                     width   = 20,
	                     height  = 17},
	Close             = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.Close.path
	                     coords  = {19/128, 36/128, 19/64, 36/64},
	                     width   = 17,
	                     height  = 17},
	Expand            = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.Expand.path
	                     coords  = {37/128, 54/128, 19/64, 36/64},
	                     width   = 17,
	                     height  = 17},
	Collapse          = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.Collapse.path
	                     coords  = {37/128, 54/128, 1/64, 18/64},
	                     width   = 17,
	                     height  = 17},
	Plus              = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.Plus.path
	                     coords  = {1/128, 8/128, 56/64, 63/64},
	                     width   = 7,
	                     height  = 7},
	Minus             = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.Minus.path
	                     coords  = {9/128, 16/128, 56/64, 63/64}},
	                     --width   = 7,  -- same as Plus
	                     --height  = 7}, -- same as Plus
	AchievementDetail = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.AchievementDetail.path
	                     coords  = {112/128, 128/128, 40/64, 64/64},
	                     width   = 8,  -- 16 scaled
	                     height  = 12},-- 24 scaled
	RaidSelect        = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.RaidSelect.path
	                     coords  = {19/128, 36/128, 1/64, 18/64},
	                     width   = 12, -- 17 scaled
	                     height  = 12},-- 17 scaled
	BossSelect        = {path    = "Interface\\TargetingFrame\\UI-TargetingFrame-Skull", -- Textures.BossSelect.path
	                     coords  = {3/32, 27/32, 0/32, 32/32},
	                     width   = 9,  -- 24 scaled
	                     height  = 12},-- 32 scaled
	SortArrow         = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.SortArrow.path
	                     coords1 = {17/128, 24/128, 55/64, 63/64},
	                     coords2 = {25/128, 32/128, 55/64, 63/64},
	                     width   = 7,
	                     height  = 8},
	Selection         = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.Selection.path
	                     coords  = {102/128, 111/128, 48/64, 63/64},
	                     width   = 6,  -- 9 scaled
	                     height  = 10},-- 15 scaled
	Trashcan          = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga", -- Textures.Trashcan.path
	                     coords  = {100/128, 111/128, 35/64, 47/64},
	                     width   = 11,
	                     height  = 12},
	--RoleIcon          = {path    = "Interface\\LFGFrame\\LFGRole", -- Textures.RoleIcon.path
	--                     coordsT = {32/64, 48/64, 0/16, 16/16}, -- Tank
	--                     coordsH = {48/64, 64/64, 0/16, 16/16}, -- Healer
	--                     coordsD = {16/64, 32/64, 0/16, 16/16}, -- Damage
	--                     width   = 12,  -- 16 scaled
	--                     height  = 12}, -- 16 scaled
	ClassIcon         = {path    = "Interface\\WorldStateFrame\\Icons-Classes"}, -- Textures.ClassIcon.path
	                     --coords  = auto
	                     --width   = depends,
	                     --height  = depends},
	TalentMiniIcons   = {--path    = auto, -- Textures.TalentMiniIcons.path
	                     coords  = {5/64, 59/64, 5/64, 59/64},
	                     width   = 12,
	                     height  = 12},
	TalentIconTexture = {--path    = auto -- Textures.TalentIconTexture.path
	                     --coords  = auto
	                     width   = 25,
	                     height  = 25},
	TalentRankBorder  = {path    = "Interface\\TalentFrame\\TalentFrame-RankBorder", -- Textures.TalentRankBorder.path
	                     coords  = {9/32, 23/32, 9/32, 22/32},
	                     width   = 14,
	                     height  = 14},
	GearIconTexture   = {--path    = auto -- Textures.GearIconTexture.path
	                     --coords  = not used - default = {0, 1, 0, 1}
	                     width   = 25,
	                     height  = 25},
	Highlight         = {path    = "Interface\\Buttons\\ButtonHilight-Square", -- Textures.Highlight.path
	                     --coords  = not used - default = {0, 1, 0, 1}
	                     width   = 25,
	                     height  = 25},
	Enchantment       = {path    = "Interface\\Icons\\Trade_Engraving", -- Textures.Enchantment.path
	                     --coords  = not used - default = {0, 1, 0, 1}
	                     width   = 12,
	                     height  = 12},
}

-- fonts
local fonts = {
	GameFontNormal         = {name = "GameFontNormal"},         -- fonts.GameFontNormal.name
	GameFontNormalSmall    = {name = "GameFontNormalSmall"},    -- fonts.GameFontNormalSmall.name
	GameFontNormalLarge    = {name = "GameFontNormalLarge"},    -- fonts.GameFontNormalLarge.name
	GameFontWhiteSmall     = {name = "GameFontWhiteSmall"},     -- fonts.GameFontWhiteSmall.name
	GameFontHighlight      = {name = "GameFontHighlight"},      -- fonts.GameFontHighlight.name
	GameFontHighlightSmall = {name = "GameFontHighlightSmall"}, -- fonts.GameFontHighlightSmall.name
	GameFontDisableSmall   = {name = "GameFontDisableSmall"},   -- fonts.GameFontDisableSmall.name
}
for key, value in pairs(fonts) do
	local _, fontHeight = _G[value.name]:GetFont()
	value.height = math_floor(fontHeight+0.5)
end

-- size General
local sizeGeneral = {}
sizeGeneral.MainWidth    = 420
sizeGeneral.Border       = 1
sizeGeneral.Space        = 2
sizeGeneral.Offset       = 5
sizeGeneral.SliderWidth  = 16
sizeGeneral.BarHeight    = fonts.GameFontNormalSmall.height + sizeGeneral.Space -- 12
sizeGeneral.StatusHeight = sizeGeneral.Offset + fonts.GameFontNormal.height + sizeGeneral.Offset

sizeGeneral.MainOffset       = sizeGeneral.Offset -- 5
sizeGeneral.MainButtonHeight = sizeGeneral.Border + sizeGeneral.Space + Textures.SingleScan.height + sizeGeneral.Space + sizeGeneral.Border -- 23 - SingleScan.height is default button height
sizeGeneral.MainHeight       = sizeGeneral.Border + sizeGeneral.MainOffset + sizeGeneral.MainButtonHeight + sizeGeneral.MainOffset + sizeGeneral.Border -- 35

-- size BaseCharFrame
local sizeBaseCharFrame_BaseHeight = sizeGeneral.Offset + fonts.GameFontNormal.height + fonts.GameFontNormalSmall.height + fonts.GameFontNormalSmall.height + sizeGeneral.Offset

-- size TalentFrame
local sizeTalentFrame = {}
sizeTalentFrame.BaseHeight = sizeGeneral.Offset + Textures.TalentMiniIcons.height + Textures.TalentMiniIcons.height + sizeGeneral.Offset
sizeTalentFrame.Offset     = sizeGeneral.Offset -- 5
sizeTalentFrame.GroupSelectionWidth  = (sizeGeneral.MainWidth/2) - (sizeTalentFrame.Offset+(sizeTalentFrame.Offset/2)) -- (mainwidth/number of max specs) - (offset+(offset/2))
sizeTalentFrame.GroupSelectionHeight = sizeGeneral.Border + sizeGeneral.Space + fonts.GameFontNormalSmall.height + sizeGeneral.Space + sizeGeneral.Border -- 16
sizeTalentFrame.GroupSelection1Left  = sizeTalentFrame.Offset
sizeTalentFrame.GroupSelection2Left  = sizeTalentFrame.Offset + sizeTalentFrame.GroupSelectionWidth + sizeTalentFrame.Offset
sizeTalentFrame.TreeHeight           = (myMAX_TIER*Textures.TalentIconTexture.height) + ((myMAX_TIER+1)*sizeTalentFrame.Offset) -- 323 -- (11*25) + (12*4) -- (maxTier*talentbuttonheight) + ((maxTier+1)*offset)
sizeTalentFrame.TreeWidth            = (myMAX_COLUMN*Textures.TalentIconTexture.width) + (myMAX_COLUMN+1)*sizeTalentFrame.Offset -- 120 --(4*25) + (5*4) --
sizeTalentFrame.TreeOffset           = 10 -- offset between the 3 talent trees -> autocalc | tree1 | TreeOffset | tree2 | TreeOffset | tree3 | autocalc
sizeTalentFrame.Tree1Left            = (sizeGeneral.MainWidth - ((myMAX_TALENT_TABS*sizeTalentFrame.TreeWidth) + ((myMAX_TALENT_TABS-1)*sizeTalentFrame.TreeOffset)) ) /2 -- 10 --(400 - ((3*120) + ((3-1)*10)) ) /2
sizeTalentFrame.Tree2Left            = sizeTalentFrame.Tree1Left + sizeTalentFrame.TreeWidth + sizeTalentFrame.TreeOffset -- 140 -- Tree1 + 120 + 10
sizeTalentFrame.Tree3Left            = sizeTalentFrame.Tree2Left + sizeTalentFrame.TreeWidth + sizeTalentFrame.TreeOffset -- 270 -- Tree2 + 120 + 10
sizeTalentFrame.TreeBaseHeight       = sizeTalentFrame.GroupSelectionHeight + sizeTalentFrame.Offset + fonts.GameFontWhiteSmall.height + fonts.GameFontWhiteSmall.height + sizeTalentFrame.TreeHeight -- 367 -- 16 + 4 + 12 + 12 + (11*25) + (12*4)

-- size Equipment
local sizeEquipmentFrame = {}
sizeEquipmentFrame.BaseHeight    = sizeGeneral.Offset + fonts.GameFontNormalSmall.height + sizeGeneral.Space + sizeGeneral.Space + fonts.GameFontNormalSmall.height + sizeGeneral.Offset -- 30
sizeEquipmentFrame.BaseHeightXXL = sizeGeneral.Offset + fonts.GameFontNormalSmall.height + sizeGeneral.Space + sizeGeneral.Space + fonts.GameFontNormalSmall.height + fonts.GameFontNormalSmall.height + sizeGeneral.Offset -- 40
sizeEquipmentFrame.Offset        = sizeGeneral.Offset -- 5
sizeEquipmentFrame.GearHeight    = sizeEquipmentFrame.Offset + (8*Textures.GearIconTexture.height) + (Textures.GearIconTexture.height/2) + sizeEquipmentFrame.Offset + fonts.GameFontNormalSmall.height + sizeEquipmentFrame.Offset + Textures.Enchantment.height + sizeEquipmentFrame.Offset-- 260 -- 5 + (8*25) + 12.5 + 5 + 12 + 5 + 12 + 5
sizeEquipmentFrame.FirstSlotLeft = (sizeGeneral.MainWidth-(7*Textures.GearIconTexture.width))/2 -- 112.5 -- (400-(7*25))/2

-- size Achievement & Statistics
local sizeStatFrame = {}
sizeStatFrame.Offset     = sizeGeneral.Offset -- 5
sizeStatFrame.Space      = sizeGeneral.Space -- 2
sizeStatFrame.BaseHeight = sizeStatFrame.Offset + ((#DATA+1)*sizeGeneral.BarHeight) + sizeStatFrame.Offset -- 58 -- 5 + ((3+1)*12) + 5
sizeStatFrame.LevelModifier         = Textures.Plus.width + sizeStatFrame.Space -- 9 10
sizeStatFrame.RaidSelectRight       = 0
sizeStatFrame.RaidSelectWidth       = Textures.RaidSelect.width
sizeStatFrame.BossesTotalRight      = sizeStatFrame.RaidSelectRight + sizeStatFrame.RaidSelectWidth + sizeStatFrame.Space
sizeStatFrame.BossesTotalWidth      = 50
sizeStatFrame.StatisticRight        = sizeStatFrame.BossesTotalRight + sizeStatFrame.BossesTotalWidth + sizeStatFrame.Space
sizeStatFrame.StatisticWidth        = 60
sizeStatFrame.AchievementDetailRight = sizeStatFrame.StatisticRight + sizeStatFrame.StatisticWidth + sizeStatFrame.Space
sizeStatFrame.AchievementDetailWidth = Textures.AchievementDetail.width
sizeStatFrame.AchievementRight      = sizeStatFrame.AchievementDetailRight + sizeStatFrame.AchievementDetailWidth + sizeStatFrame.Space
sizeStatFrame.AchievementWidth      = 60
sizeStatFrame.AchievementRightTitle = sizeStatFrame.AchievementDetailRight + sizeStatFrame.AchievementDetailWidth + sizeStatFrame.Space -- only title
sizeStatFrame.AchievementWidthTitle = 80 -- only title
sizeStatFrame.ButtonWidth = sizeGeneral.MainWidth-(2 * sizeStatFrame.Offset)
sizeStatFrame.NameLeft       = 0
sizeStatFrame.NameWidth      = sizeStatFrame.ButtonWidth - (sizeStatFrame.Space + sizeStatFrame.AchievementWidth + sizeStatFrame.AchievementRight)
sizeStatFrame.NameWidthTitle = sizeStatFrame.ButtonWidth - (sizeStatFrame.Space + sizeStatFrame.AchievementWidthTitle + sizeStatFrame.AchievementRightTitle)
sizeStatFrame.NameLeft_L1c   = sizeStatFrame.LevelModifier
sizeStatFrame.NameWidth_L1c  = sizeStatFrame.NameWidth - sizeStatFrame.LevelModifier
sizeStatFrame.NameLeft_L2    = sizeStatFrame.LevelModifier
sizeStatFrame.NameWidth_L2   = sizeStatFrame.NameWidth_L1c + sizeStatFrame.Space + sizeStatFrame.AchievementWidth + sizeStatFrame.Space + sizeStatFrame.AchievementDetailWidth + sizeStatFrame.Space + sizeStatFrame.StatisticWidth
sizeStatFrame.NameLeft_L2c   = sizeStatFrame.LevelModifier * 2
sizeStatFrame.NameWidth_L2c  = sizeStatFrame.NameWidth_L1c - sizeStatFrame.LevelModifier
sizeStatFrame.NameLeft_L3    = sizeStatFrame.LevelModifier * 2
sizeStatFrame.NameWidth_L3   = sizeStatFrame.NameWidth_L2 - sizeStatFrame.LevelModifier
sizeStatFrame.NameLeft_L3c   = sizeStatFrame.LevelModifier * 3
sizeStatFrame.NameWidth_L3c  = sizeStatFrame.NameWidth_L2c - sizeStatFrame.LevelModifier
sizeStatFrame.NameLeft_L4    = sizeStatFrame.LevelModifier * 3
sizeStatFrame.NameWidth_L4   = sizeStatFrame.NameWidth_L3 - sizeStatFrame.LevelModifier

-- size AchievementDetailFrame
local sizeAchievementDetail = {}
sizeAchievementDetail.Offset        = sizeGeneral.Offset -- 5
sizeAchievementDetail.Space         = sizeGeneral.Space -- 2
sizeAchievementDetail.SortHeight    = 12
sizeAchievementDetail.IdRight       = 0
sizeAchievementDetail.IdWidth       = 35
sizeAchievementDetail.NameRight     = sizeAchievementDetail.IdRight + sizeAchievementDetail.IdWidth + sizeAchievementDetail.Space
sizeAchievementDetail.NameWidth     = 220
sizeAchievementDetail.ButtonWidth   = sizeAchievementDetail.NameRight + sizeAchievementDetail.NameWidth + sizeAchievementDetail.Space + sizeGeneral.SliderWidth
sizeAchievementDetail.Width         = sizeAchievementDetail.Offset + sizeAchievementDetail.ButtonWidth + sizeAchievementDetail.Offset
sizeAchievementDetail.HighlightLeft = sizeAchievementDetail.Space + sizeGeneral.SliderWidth
sizeAchievementDetail.SortLine1Top    = sizeAchievementDetail.Offset
sizeAchievementDetail.SortLine1Height = 22
sizeAchievementDetail.SortLine2Top    = sizeAchievementDetail.SortLine1Top + sizeAchievementDetail.SortLine1Height + sizeAchievementDetail.Space
sizeAchievementDetail.SortLine2Height = sizeGeneral.BarHeight
sizeAchievementDetail.SortLineXTop    = sizeAchievementDetail.SortLine2Top + sizeAchievementDetail.SortLine2Height + sizeAchievementDetail.Space
sizeAchievementDetail.Height          = sizeAchievementDetail.SortLineXTop + (GUI_achievementdetail_maxNum*sizeGeneral.BarHeight) + sizeAchievementDetail.Offset
sizeAchievementDetail.InfoTexture       = 24
sizeAchievementDetail.InfoTextWidth     = sizeAchievementDetail.Width - (3*sizeGeneral.Offset) - sizeAchievementDetail.InfoTexture
sizeAchievementDetail.InfoTextminHeight = (sizeAchievementDetail.InfoTexture + (2*sizeGeneral.Offset))

-- size SummaryFrame
local sizeSummaryFrame_BaseHeight = sizeGeneral.Offset + sizeGeneral.BarHeight + sizeGeneral.BarHeight + sizeGeneral.Offset

-- size ProtectionFrame
local sizeProtectionFrame_BarWidth = 300

-- size History/CompareFrame
local sizeCompareFrame = {}
sizeCompareFrame.Offset        = sizeGeneral.Offset -- 5
sizeCompareFrame.Space         = sizeGeneral.Space -- 2
sizeCompareFrame.HighlightLeft = sizeCompareFrame.Space + sizeGeneral.SliderWidth + sizeCompareFrame.Offset
sizeCompareFrame.TrashcanLeft     = sizeCompareFrame.Offset
sizeCompareFrame.TrashcanWidth    = Textures.Trashcan.width
sizeCompareFrame.ArrowLeft        = sizeCompareFrame.TrashcanLeft + sizeCompareFrame.TrashcanWidth + sizeCompareFrame.Space
sizeCompareFrame.ArrowWidth       = Textures.Selection.width
sizeCompareFrame.ClassiconLeft    = sizeCompareFrame.ArrowLeft + sizeCompareFrame.ArrowWidth + sizeCompareFrame.Space
sizeCompareFrame.ClassiconWidth   = sizeGeneral.BarHeight
sizeCompareFrame.NameLeft         = sizeCompareFrame.ClassiconLeft + sizeCompareFrame.ClassiconWidth + sizeCompareFrame.Space
sizeCompareFrame.NameWidth        = 120
sizeCompareFrame.TimeLeft         = sizeCompareFrame.NameLeft + sizeCompareFrame.NameWidth + sizeCompareFrame.Space
sizeCompareFrame.TimeWidth        = 40
sizeCompareFrame.MainroleLeft     = sizeCompareFrame.TimeLeft + sizeCompareFrame.TimeWidth + sizeCompareFrame.Space
sizeCompareFrame.MainroleWidth    = 15
sizeCompareFrame.Talents1Left     = sizeCompareFrame.MainroleLeft + sizeCompareFrame.MainroleWidth + sizeCompareFrame.Space
sizeCompareFrame.Talents1Width    = 150
sizeCompareFrame.Talents2Left     = sizeCompareFrame.Talents1Left + sizeCompareFrame.Talents1Width + sizeCompareFrame.Space
sizeCompareFrame.Talents2Width    = 150
sizeCompareFrame.EquipLeft        = sizeCompareFrame.Talents2Left + sizeCompareFrame.Talents2Width + sizeCompareFrame.Space
sizeCompareFrame.EquipWidth       = 50
sizeCompareFrame.AchievementLeft  = sizeCompareFrame.EquipLeft + sizeCompareFrame.EquipWidth + sizeCompareFrame.Space
sizeCompareFrame.AchievementWidth = 80
sizeCompareFrame.BossLeft         = sizeCompareFrame.AchievementLeft + sizeCompareFrame.AchievementWidth + sizeCompareFrame.Space
sizeCompareFrame.BossWidth        = 80
sizeCompareFrame.KillsLeft        = sizeCompareFrame.BossLeft + sizeCompareFrame.BossWidth + sizeCompareFrame.Space
sizeCompareFrame.KillsWidth       = 50
sizeCompareFrame.HistoryWidth = sizeCompareFrame.MainroleLeft + sizeCompareFrame.MainroleWidth + sizeCompareFrame.Space + sizeGeneral.SliderWidth + sizeCompareFrame.Offset
sizeCompareFrame.CompareWidth = sizeCompareFrame.KillsLeft    + sizeCompareFrame.KillsWidth    + sizeCompareFrame.Space + sizeGeneral.SliderWidth + sizeCompareFrame.Offset
sizeCompareFrame.SortLine1Top    = sizeCompareFrame.Offset
sizeCompareFrame.SortLine1Height = 22
sizeCompareFrame.SortLine2Top    = sizeCompareFrame.SortLine1Top + sizeCompareFrame.SortLine1Height + sizeCompareFrame.Space
sizeCompareFrame.SortLine2Height = sizeGeneral.BarHeight
sizeCompareFrame.SortLineXTop    = sizeCompareFrame.SortLine2Top + sizeCompareFrame.SortLine2Height + sizeCompareFrame.Space
sizeCompareFrame.HistoryHeight = sizeCompareFrame.SortLineXTop + (GUI_history_maxNum*sizeGeneral.BarHeight) + sizeCompareFrame.Offset

local DefaultStatsViewData = { -- WotLK default view | customizable by user TODO
	[1] = {lvl = 1, name1 = "WotLK", index1 = 3, name2 = nil,       index2 = nil, name3 = nil, index3 = nil},
	[2] = {lvl = 2, name1 = "WotLK", index1 = 3, name2 = "raid10p", index2 = 2,   name3 = nil, index3 = nil},
	[3] = {lvl = 2, name1 = "WotLK", index1 = 3, name2 = "raid25p", index2 = 3,   name3 = nil, index3 = nil}
}

local color = { -- e.g.: color.White.hex | color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4]
	DarkGrey            = {rgba = {0.2,  0.2,  0.2,  1    }, hex = "333333"}, -- color.DarkGrey.
	Grey                = {rgba = {0.5,  0.5,  0.5,  1    }, hex = "808080"}, -- color.Grey.
	Red                 = {rgba = {1,    0,    0,    1    }, hex = "ff0000"}, -- color.Red.
	White               = {rgba = {1,    1,    1,    1    }, hex = "ffffff"}, -- color.White.
	Level1              = {rgba = {1,    1,    1,    1    }, hex = "ffffff"}, -- color.Level1.
	Level2              = {rgba = {0.8,  0.8,  0.8,  1    }, hex = "cccccc"}, -- color.Level2.
	Level3              = {rgba = {0.6,  0.6,  0.6,  1    }, hex = "999999"}, -- color.Level3.
	LightYellow         = {rgba = {1,    1,    0.49, 1    }, hex = "ffff7f"}, -- color.LightYellow.
	Yellow              = {rgba = {1,    0.82, 0,    1    }, hex = "ffd100"}, -- color.Yellow.
	Green               = {rgba = {0,    0.75, 0,    1    }, hex = "00bf00"}, -- color.Green.
	Brown               = {rgba = {0.78, 0.52, 0.28, 1    }, hex = "c78547"}, -- color.Brown.
	Firebrick           = {rgba = {0.8,  0.00, 0.00, 1    }, hex = "cd0000"}, -- color.Firebrick.
	FrameBG             = {rgba = {0,    0,    0,    1    }},                 -- color.FrameBG.
	FrameBorder         = {rgba = {0.4,  0.4,  0.4,  1    }},                 -- color.FrameBorder.
	ButtonBG            = {rgba = {0,    0,    0,    1    }},                 -- color.ButtonBG.
	ButtonBGActive      = {rgba = {0.25, 0.25, 0.25, 1    }},                 -- color.ButtonBGActive.
	ButtonHighlight     = {rgba = {1,    1,    1,    0.2  }},                 -- color.ButtonHighlight.
	ButtonHighlightLow  = {rgba = {0.6,  0.6,  0.6,  0.2  }},                 -- color.ButtonHighlightLow.
	ButtonHighlightRed  = {rgba = {0.55, 0,    0,    0.5  }},                 -- color.ButtonHighlightRed.
	ButtonBorder        = {rgba = {0.4,  0.4,  0.4,  1    }},                 -- color.ButtonBorder.
	ButtonBorderDisable = {rgba = {0.2,  0.2,  0.2,  1    }},                 -- color.ButtonBorderDisable.
	ButtonSortBG        = {rgba = {1,    1,    1,    0.1  }},                 -- color.ButtonSortBG.
	ButtonLineMarker    = {rgba = {0.06, 0.06, 0.06, 1    }},                 -- color.ButtonLineMarker.
}

local factionimgAlliance = "Interface\\TargetingFrame\\UI-PVP-Alliance"
local factionimgHorde    = "Interface\\TargetingFrame\\UI-PVP-Horde"

local classcolors = RAID_CLASS_COLORS
local classes = {-- the classes table is populated with talent data if a class is scanned the first time
	DEATHKNIGHT = {icon = {0.265625, 0.484375, 0.515625, 0.734375}, -- (68/256, 124/256, 132/256, 188/256)
	               role = {[1] = "?",   -- TANK/DAMAGE? | Blood  \
	                       [2] = "?",   -- TANK/DAMAGE? | Frost   > A DEATHKNIGHT is classified as TANK with at least 10 points in Blade Barrier+Toughness+Anticipation.
	                       [3] = "?"}}, -- TANK/DAMAGE? | Unholy /
	DRUID       = {icon = {0.7578125, 0.9765625, 0.015625, 0.234375}, -- (194/256, 250/256, 4/256, 60/256)
	               role = {[1] = "D",   -- DAMAGE | Balance
	                       [2] = "T",   -- TANK   | FeralCombat  -> A DRUID is classified as TANK with at least 4 points in Thick Hide+Protector of the Pack.
	                       [3] = "H"}}, -- HEAL   | Restoration
	HUNTER      = {icon = {0.01953125, 0.23828125, 0.265625, 0.484375}, -- (5/256, 61/256, 68/256, 124/256)
	               role = {[1] = "D",   -- DAMAGE | BeastMastery
	                       [2] = "D",   -- DAMAGE | Marksmanship
	                       [3] = "D"}}, -- DAMAGE | Survival
	MAGE        = {icon = {0.265625, 0.484375, 0.015625, 0.234375}, -- (68/256, 124/256, 4/256, 60/256)
	               role = {[1] = "D",   -- DAMAGE | Arcane
	                       [2] = "D",   -- DAMAGE | Fire
	                       [3] = "D"}}, -- DAMAGE | Frost
	PALADIN     = {icon = {0.01953125, 0.23828125, 0.515625, 0.734375}, -- (5/256, 61/256, 132/256, 188/256)
	               role = {[1] = "H",   -- HEAL   | Holy
	                       [2] = "T",   -- TANK   | Protection
	                       [3] = "D"}}, -- DAMAGE | Combat
	PRIEST      = {icon = {0.51171875, 0.73046875, 0.265625, 0.484375}, -- (131/256, 187/256, 68/256, 124/256)
	               role = {[1] = "H",   -- HEAL   | Discipline
	                       [2] = "H",   -- HEAL   | Holy
	                       [3] = "D"}}, -- DAMAGE | Shadow
	ROGUE       = {icon = {0.51171875, 0.73046875, 0.015625, 0.234375}, -- (131/256, 187/256, 4/256, 60/256)
	               role = {[1] = "D",   -- DAMAGE | Assassination
	                       [2] = "D",   -- DAMAGE | Combat
	                       [3] = "D"}}, -- DAMAGE | Subtlety
	SHAMAN      = {icon = {0.265625, 0.484375, 0.265625, 0.484375}, -- (68/256, 124/256, 68/256, 124/256)
	               role = {[1] = "D",   -- DAMAGE | ElementalCombat
	                       [2] = "D",   -- DAMAGE | Enhancement
	                       [3] = "H"}}, -- HEAL   | Restoration
	WARLOCK     = {icon = {0.7578125, 0.9765625, 0.265625, 0.484375}, -- (194/256, 250/256, 68/256, 124/256)
	               role = {[1] = "D",   -- DAMAGE | Curses
	                       [2] = "D",   -- DAMAGE | Summoning
	                       [3] = "D"}}, -- DAMAGE | Destruction
	WARRIOR     = {icon = {0.01953125, 0.23828125, 0.015625, 0.234375}, -- (5/256, 61/256, 4/256, 60/256)
	               role = {[1] = "D",   -- DAMAGE | Arms
	                       [2] = "D",   -- DAMAGE | Fury
	                       [3] = "T"}}  -- TANK   | Protection
}

local guildRank = {
	[1] = GUILD_RANK0_DESC, -- 0 | Guild Master
	[2] = GUILD_RANK1_DESC, -- 1 | Officer
	[3] = GUILD_RANK2_DESC, -- 2 | Veteran
	[4] = GUILD_RANK3_DESC, -- 3 | Member
	[5] = GUILD_RANK4_DESC  -- 4 | Initiate
}

-- Equipment
local slots = {
	[1] = "HeadSlot",
	[2] = "NeckSlot",
	[3] = "ShoulderSlot",
	[4] = "BackSlot",
	[5] = "ChestSlot",
	[6] = "WristSlot",
	[7] = "HandsSlot",
	[8] = "WaistSlot",
	[9] = "LegsSlot",
	[10] = "FeetSlot",
	[11] = "Finger0Slot",
	[12] = "Finger1Slot",
	[13] = "Trinket0Slot",
	[14] = "Trinket1Slot",
	[15] = "MainHandSlot",
	[16] = "SecondaryHandSlot",
	[17] = "RangedSlot"	
}
local slotsNum = #slots

local slotButton = {
	[1]  = {id =  1, sid =  1, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Head"},
	[2]  = {id =  2, sid =  2, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Neck"},
	[3]  = {id =  3, sid =  3, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Shoulder"},
	[4]  = {id =  4, sid = 15, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Chest"}, -- Back
	[5]  = {id =  5, sid =  5, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Chest"},
	[6]  = {id = -1, sid = -1, IQ = false, texture="Interface\\PaperDollInfoFrame\\UI-GearManager-LeaveItem-Opaque"}, -- Interface\\Paperdoll\\UI-PaperDoll-Slot-Shirt
	[7]  = {id = -1, sid = -1, IQ = false, texture="Interface\\PaperDollInfoFrame\\UI-GearManager-LeaveItem-Opaque"}, -- Interface\\Paperdoll\\UI-PaperDoll-Slot-Tabart
	[8]  = {id =  6, sid =  9, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Wrists"},
	[9]  = {id = 15, sid = 16, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-MainHand"}, -- middle
	[10] = {id = 16, sid = 17, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-SecondaryHand"}, -- middle
	[11] = {id = 17, sid = 18, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Ranged"}, -- middle
	[12] = {id = -1, sid = -1, IQ = false, texture="Interface\\PaperDollInfoFrame\\UI-GearManager-LeaveItem-Opaque"}, -- Interface\\Paperdoll\\UI-PaperDoll-Slot-Ammo" -- middle
	[13] = {id = 14, sid = 14, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Trinket"},
	[14] = {id = 13, sid = 13, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Trinket"},
	[15] = {id = 12, sid = 12, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Finger"},
	[16] = {id = 11, sid = 11, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Finger"},
	[17] = {id = 10, sid =  8, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Feet"},
	[18] = {id =  9, sid =  7, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Legs"},
	[19] = {id =  8, sid =  6, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Waist"},
	[20] = {id =  7, sid = 10, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Hands"}
}

local itemQuality = {
	[1] = {name = ITEM_QUALITY0_DESC, color = "9d9d9d"}, -- 0 | poor
	[2] = {name = ITEM_QUALITY1_DESC, color = "ffffff"}, -- 1 | common
	[3] = {name = ITEM_QUALITY2_DESC, color = "1eff00"}, -- 2 | uncommon
	[4] = {name = ITEM_QUALITY3_DESC, color = "0070dd"}, -- 3 | rare
	[5] = {name = ITEM_QUALITY4_DESC, color = "a335ee"}, -- 4 | epic
	[6] = {name = ITEM_QUALITY5_DESC, color = "ff8000"}, -- 5 | legendary
	[7] = {name = ITEM_QUALITY6_DESC, color = "e6cc80"}, -- 6 | artifact (not in game)
	[8] = {name = ITEM_ACCOUNTBOUND,  color = "e6cc80"}  -- 7 | account bound
}
local itemQualityNum = #itemQuality
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
local function Print(...)
	print("|cff"..color.LightYellow.hex.."CharScanner:|r", ...)
end

local function ClassHexColor(class)
	local hex = string.format("%.2x%.2x%.2x", classcolors[class].r*255, classcolors[class].g*255, classcolors[class].b*255)
	return hex or "cccccc"
end

local function cnum(number)
	local a = tostring(number)
	local p = string.find(a, ".", 1, true)
	if p then
		local r = string.sub(a, p+1)
		if string.len(r) == 1 then
			number = number.."0"
		end
	else
		number = number..".00"
	end
	return number -- string:)
end

local function BuildIDTable()
	local temp = {}
	-- achievements START
	for i=1, GetCategoryNumAchievements(-1) do
		local id, name = GetAchievementInfo(-1, i)
		if id and name then
			temp[id] = name
			--TEST code for future Achievement updates START
			--if CharScanner_Options.AchievementTest == nil then CharScanner_Options.AchievementTest = {} end
			--if not CharScanner_Options.AchievementTest[id] then
			--	CharScanner_Options.AchievementTest[id] = name
			--end
			--TEST code for future Achievement updates END
		end
	end
	for i = 1, #DATA do
		if DATA[i].achievement then
			for key, value in pairs(DATA[i].achievement) do
				AchievementIDTable[value] = temp[value]
			end
		end
		for j = 1, #DATA[i] do
			if DATA[i][j].achievement then
				for key, value in pairs(DATA[i][j].achievement) do
					AchievementIDTable[value] = temp[value]
				end
			end
			for k = 1, #DATA[i][j] do
				if DATA[i][j][k].achievement then
					for key, value in pairs(DATA[i][j][k].achievement) do
						AchievementIDTable[value] = temp[value]
					end
				end
			end
		end
	end
	-- --------------------------------------------
	-- DUMMY_FactionFix START
	-- --------------------------------------------
	if temp[4297] then
		AchievementIDTable[4297] = temp[4297] -- HORDE    (same text) -> Heroic: Trial of the Champion
		AchievementIDTable[4298] = temp[4297] -- ALLIANCE (same text) -> Heroic: Trial of the Champion
	else
		AchievementIDTable[4297] = temp[4298] -- HORDE    (same text) -> Heroic: Trial of the Champion
		AchievementIDTable[4298] = temp[4298] -- ALLIANCE (same text) -> Heroic: Trial of the Champion
	end
	-- --------------------------------------------
	-- DUMMY_FactionFix END
	-- --------------------------------------------
	table.wipe(temp)
	-- achievements END

	-- statistic START
	for i=1, GetCategoryNumAchievements(-2) do
		local id, name = GetAchievementInfo(-2, i)
		if id and name then
			temp[id] = name
			--TEST code for future Statistic updates START
			--if CharScanner_Options.StatisticTest == nil then CharScanner_Options.StatisticTest = {} end
			--if not CharScanner_Options.StatisticTest[id] then
			--	CharScanner_Options.StatisticTest[id] = name
			--end
			--TEST code for future Statistic updates END
		end
	end
	for i = 1, #DATA do
		if DATA[i].statistic then
			for key, value in pairs(DATA[i].statistic) do
				StatisticIDTable[value] = temp[value]
			end
		end
		for j = 1, #DATA[i] do
			if DATA[i][j].statistic then
				for key, value in pairs(DATA[i][j].statistic) do
					StatisticIDTable[value] = temp[value]
				end
			end
			for k = 1, #DATA[i][j] do
				if DATA[i][j][k].statistic then
					for key, value in pairs(DATA[i][j][k].statistic) do
						StatisticIDTable[value] = temp[value]
					end
				end
			end
		end
	end
	table.wipe(temp)
	-- statistic END
	--local c, d = 0, 0 -- TEST
	--for key in pairs(AchievementIDTable) do c = c + 1 end
	--for key in pairs(StatisticIDTable) do d = d + 1 end
	--print("AchievementIDTable:", c, "StatisticIDTable:", d)
end

local function CreateScanDataByGUID(guid)
	SCAN[guid] = {}
	SCAN[guid].name = nil                      -- scanned unit: base info - name
	SCAN[guid].level = nil                     -- scanned unit: base info - level
	SCAN[guid].race = nil                      -- scanned unit: base info - race
	SCAN[guid].class = nil                     -- scanned unit: base info - class (localized)
	SCAN[guid].classEN = nil                   -- scanned unit: base info - class (english)
	SCAN[guid].faction = nil                   -- scanned unit: base info - faction (english)
	SCAN[guid].guildName = nil                 -- scanned unit: base info - guild name
	SCAN[guid].guildRankName = nil             -- scanned unit: base info - guild rank name
	SCAN[guid].guildRankIndex = nil            -- scanned unit: base info - guild rank index
	SCAN[guid].talentGroups = nil              -- scanned unit: talent - talent groups (count)
	SCAN[guid].activeTalentGroup = nil         -- scanned unit: talent - active talent group

	SCAN[guid].talents = {}                    -- scanned unit: all spend talent points
	SCAN[guid].talentGroup1name = nil          -- scanned unit: talent - talent group 1 name with most points (eg: Affliction)
	SCAN[guid].talentGroup1role = nil          -- scanned unit: talent - talent group 1 role (class role: eg: PRIEST 2 -> HEAL)
	SCAN[guid].talentGroup1pointsString = nil  -- scanned unit: talent - talent group 1 points (eg: 55/11/6)
	SCAN[guid].talentGroup1pointsUnspent = nil -- scanned unit: talent - talent group 1 points unspend (eg: 0)
	SCAN[guid].talentGroup1icon = nil          -- scanned unit: talent - talent group 1 icon
	SCAN[guid].talentGroup2name = nil          -- scanned unit: talent - talent group 2 name with most points (eg: Destruction)
	SCAN[guid].talentGroup2role = nil          -- scanned unit: talent - talent group 2 role (class role: eg: PRIEST 3 -> DAMAGE)
	SCAN[guid].talentGroup2pointsString = nil  -- scanned unit: talent - talent group 2 points (eg: 3/10/58)
	SCAN[guid].talentGroup2pointsUnspent = nil -- scanned unit: talent - talent group 2 points unspend (eg: 1)
	SCAN[guid].talentGroup2icon = nil          -- scanned unit: talent - talent group 2 icon

	SCAN[guid].gear = {}                       -- scanned unit: itemLink from every gear slot
	--SCAN[guid].gearStats = {}                  -- scanned unit: basic gear stats from every gear slot
	SCAN[guid].itemlevelAvg = 0                -- scanned unit: itemlevel - average item level
	SCAN[guid].itemlevelMin = 0                -- scanned unit: itemlevel - lowest item level
	SCAN[guid].itemlevelMax = 0                -- scanned unit: itemlevel - highest item level
	SCAN[guid].itemlevelMiss = 0               -- scanned unit: itemlevel - number of unused Slots
	SCAN[guid].itemlevelQuality = nil          -- scanned unit: itemlevel - quality

	SCAN[guid].DATA = {}                       -- scanned unit: achievement & statistic data
	SCAN[guid].statsAchievementTotal = 0       -- scanned unit: summary: achievements - total achievement points
	SCAN[guid].statsAchievementCount = 0       -- scanned unit: summary: achievements - total achievement points
	SCAN[guid].statsStatisticTotal = 0         -- scanned unit: summary: statistics - total achievement points
	SCAN[guid].statsStatisticCount = 0         -- scanned unit: summary: statistics - total achievement points
	SCAN[guid].statsStatisticBossKills = 0     -- scanned unit: summary: statistics - total achievement points

	SCAN[guid].time = nil                      -- scanned unit: scan time (eg.: '22:19')
	SCAN[guid].scanTime = nil                  -- scanned unit: time that elapsed for scan (eg.: '0.362 seconds')
end

local function BossStatistic(id, inspect)
	-- bugfix START
	-- bug       : incorrect return value from GetComparisonStatistic()/GetStatistic()
	-- workaround: check related achievement and return 1 or 0
	if id == 4045 then -- Times completed the Trial of the Grand Crusader (10 player)
		local x = GetAchievementComparisonInfo(3918, 1)
		if x then
			return 1
		else
			return 0
		end
	end
	-- bugfix END

	if inspect then
		return tonumber(GetComparisonStatistic(id)) or 0
	else
		return tonumber(GetStatistic(id)) or 0
	end
end

local function NodeDisplay(button, state)
	if state then
		_G[button]:Show()
	else
		_G[button]:Hide()
	end
	for i = 1, #GUI_menu_values do
		if GUI_menu_values[i].name == button then
			GUI_menu_values[i].collapse = state
			break
		end
	end

	local height = 0
	local visibleNode = {}
	local si = 1
	for i = 1, #GUI_menu_values do
		if GUI_menu_values[i].collapse then
			if GUI_menu_values[i].level == 2 then
				visibleNode[si] = i
				si = si + 1
			elseif GUI_menu_values[i].level == 3 then
				for j = 1, #GUI_menu_values do
					if GUI_menu_values[j].level == 2 and GUI_menu_values[j].index1 == GUI_menu_values[i].index1 and GUI_menu_values[j].collapse then
						visibleNode[si] = i
						si = si + 1
					end
				end
			elseif GUI_menu_values[i].level == 4 then

				for j = 1, #GUI_menu_values do
					if GUI_menu_values[j].level == 2 and GUI_menu_values[j].index1 == GUI_menu_values[i].index1 and GUI_menu_values[j].collapse then--and
						for k = 1, #GUI_menu_values do
							if GUI_menu_values[k].level == 3 and GUI_menu_values[k].index2 == GUI_menu_values[i].index2 and GUI_menu_values[k].collapse then
								visibleNode[si] = i
								si = si + 1
							end
						end
					end
				end

			end
		end
	end
	for i = 1, #visibleNode do
		height = height + GUI_menu_values[visibleNode[i]].height
	end
	table.wipe(visibleNode)
	return height
end

local function CreateGUIhistoryRole(role, isActive, isSame)
	if not GUI_history_Role[role] then
		GUI_history_Role[role] = {}
		GUI_history_Role[role].active = isActive
		GUI_history_Role[role].count = 1
	else
		if isActive == 1 then
			GUI_history_Role[role].active = GUI_history_Role[role].active + 1
		end
		if not isSame then
			GUI_history_Role[role].count = GUI_history_Role[role].count + 1
		end
	end
end

local function ItemData(unit, index)
	local slotID = GetInventorySlotInfo(slots[index])
	local itemLink = GetInventoryItemLink(unit, slotID)
	if not SCAN[SCANGUID].gear[index] then SCAN[SCANGUID].gear[index] = {} end

	if itemLink then
		local _, _, linkID, enchID = itemLink:find("item:(%d+):(%d+)")
		local _, _, itemRarity, iLevel = GetItemInfo(linkID)

		SCAN[SCANGUID].gear[index].iLink = itemLink or ""
		SCAN[SCANGUID].gear[index].iLvl  = iLevel or 0
		SCAN[SCANGUID].gear[index].iRar  = itemRarity or 0
		SCAN[SCANGUID].gear[index].iEnch = enchID or "0"

		--for key, value in pairs(GetItemStats(itemLink)) do
		--	if key and value and type(value) == "number" then
		--		if not SCAN[SCANGUID].gearStats[key] then
		--			SCAN[SCANGUID].gearStats[key] = value
		--		else
		--			SCAN[SCANGUID].gearStats[key] = SCAN[SCANGUID].gearStats[key] + value
		--		end
		--	end
		--end

		return iLevel or 0, itemRarity or 0
	else

		SCAN[SCANGUID].gear[index].iLink = ""
		SCAN[SCANGUID].gear[index].iLvl  = -1
		SCAN[SCANGUID].gear[index].iRar  = -1
		SCAN[SCANGUID].gear[index].iEnch = "0"

		return -1, -1
	end
end

local function TwoHandWeaponCheck(unit, who)
	local id = GetInventoryItemLink(unit, _G[who.."MainHandSlot"]:GetID())
	local twohandweapon = false
	if id then
		local linkID = select(3, id:find("item:(%d+)"))
		local itemEquipLoc = select(9, GetItemInfo(linkID))
		if itemEquipLoc == "INVTYPE_2HWEAPON" then
			local id2 = GetInventoryItemLink(unit, _G[who.."SecondaryHandSlot"]:GetID())
			if not id2 then
				twohandweapon = true
			end
		end
	end
	return twohandweapon
end

local function GearTooltipHandler(self)
	if not self.comparing and IsModifiedClick("COMPAREITEMS") then
		self.shoppingTooltips = {ShoppingTooltip1, ShoppingTooltip2, ShoppingTooltip3}
		GameTooltip_ShowCompareItem()
		self.comparing = true
	elseif self.comparing and not IsModifiedClick("COMPAREITEMS") then
		for _, frame in pairs(self.shoppingTooltips) do
			frame:Hide()
		end
		self.comparing = false
	end
	if IsModifiedClick("DRESSUP") then
		ShowInspectCursor()
	else
		ResetCursor()
	end
end

local MiniOptions_TimerStart
local function MiniOptionsTimer(self)
	--print(MiniOptions_TimerStart, GetTime()) -- TEST
	if GetTime() >= MiniOptions_TimerStart+3 then
		CharScanner_MiniOptionsFrame:Hide()
		self:SetScript("OnUpdate", nil)
	end
end

local function SlashHandler(msg)
	msg = strtrim(string.lower(msg or ""))
	if msg == "options" then
		InterfaceOptionsFrame_OpenToCategory("CharScanner")
	elseif msg == "mem" then
		memusage = not memusage
		if memusage then
			Print("Memory usage display activated. For testing only!!!")
		else
			Print("Memory usage display deactivated.")
		end		
		CharScanner:UpdateHistoryStatus()
	else
		CharScanner:MainFrameToggle(true)
		if not SingleScanIsActive and not GroupScanIsActive then
			CharScanner:ScanUnit("target")
		end
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:InitOptions()
	SlashCmdList["CHARSCANNER"] = SlashHandler
	SLASH_CHARSCANNER1 = "/cs"
	SLASH_CHARSCANNER2 = "/scan"
	SLASH_CHARSCANNER3 = "/charscanner"

	-- version check | version number = internal options version
	if CharScanner_Options.version then
		if CharScanner_Options.version == 1 then
			CharScanner_Options = {}
			CharScanner_Options.version = 2
			Print("Options reset.")
		end
		if CharScanner_Options.version == 2 then
			CharScanner_Options.version = 3
			CharScanner_Options.scanButtonTrans = nil
		end
	end

	if CharScanner_Options.version          == nil then CharScanner_Options.version          = 3     end
	if CharScanner_Options.scanButton       == nil then CharScanner_Options.scanButton       = false end
	if CharScanner_Options.scanButtonLock   == nil then CharScanner_Options.scanButtonLock   = false end
	if CharScanner_Options.scanButtonSize   == nil then CharScanner_Options.scanButtonSize   = 24    end
	if CharScanner_Options.chatoutput       == nil then CharScanner_Options.chatoutput       = false end
	if CharScanner_Options.chatoutputGrp    == nil then CharScanner_Options.chatoutputGrp    = false end
	if CharScanner_Options.scanDuration     == nil then CharScanner_Options.scanDuration     = 1     end
	if CharScanner_Options.MinimapButton    == nil then CharScanner_Options.MinimapButton    = false end
	if CharScanner_Options.MinimapButtonPos == nil then CharScanner_Options.MinimapButtonPos = -90   end
	--CharScanner_Options.scanDuration     = 0.15 --TEST
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------

function CharScanner:LDBcheck()
	if LibStub and LibStub:GetLibrary("CallbackHandler-1.0", true) and LibStub:GetLibrary("LibDataBroker-1.1", true) then
		LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("CharScanner", {
			type = "launcher",
			icon = Textures.CharScannerGUI.path,
			OnClick = function(self, button)
				CharScanner:MiniOptionsToggle(button, 1)
			end,
		})
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:CreateInterfaceOptions()
	local options = CreateFrame("FRAME", "CharScanner_InterfaceOptions")
	options.name = "CharScanner"

	local title = options:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalLarge.name)
	title:SetPoint("TOPLEFT", 16, -16)
	title:SetText("CharScanner")
	title:SetJustifyH("LEFT")
	title:SetJustifyV("TOP")

	local GUI = CreateFrame("Button", "CharScanner_InterfaceOptions_GUI", CharScanner_InterfaceOptions, "UIPanelButtonTemplate")
	CharScanner_InterfaceOptions_GUI:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -10)
	CharScanner_InterfaceOptions_GUI:SetWidth(120)
	CharScanner_InterfaceOptions_GUI:SetHeight(24)
	CharScanner_InterfaceOptions_GUIText:SetText(L["Open GUI"])
	CharScanner_InterfaceOptions_GUI:SetScript("OnClick", function(self)
		InterfaceOptionsFrame_Show()  -- hide InterfaceOptionsFrame
		HideUIPanel(GameMenuFrame)    -- hide GameMenuFrame
		CharScanner:MainFrameToggle() -- show CharScanner GUI
	end)

	local slashcommandText = options:CreateFontString(nil, "ARTWORK", fonts.GameFontHighlightSmall.name)
	slashcommandText:SetPoint("LEFT", GUI, "RIGHT", 10, 0)
	slashcommandText:SetText("(|cff"..color.LightYellow.hex.."/cs|r - |cff"..color.LightYellow.hex.."/scan|r - |cff"..color.LightYellow.hex.."/charscanner|r)")
	slashcommandText:SetNonSpaceWrap(true)

	local ScanButton = CreateFrame("CheckButton", "CharScanner_InterfaceOptions_ScanButton", CharScanner_InterfaceOptions, "InterfaceOptionsCheckButtonTemplate")
	CharScanner_InterfaceOptions_ScanButton:SetPoint("TOPLEFT", GUI, "BOTTOMLEFT", 0, -10)
	CharScanner_InterfaceOptions_ScanButtonText:SetText(L["Show SCAN-Button"])
	CharScanner_InterfaceOptions_ScanButton:SetChecked(CharScanner_Options.scanButton)
	CharScanner_InterfaceOptions_ScanButton.tooltipText = L["buttonshow_desc"]

	local ScanButtonLock = CreateFrame("CheckButton", "CharScanner_InterfaceOptions_ScanButtonLock", CharScanner_InterfaceOptions, "InterfaceOptionsCheckButtonTemplate")
	CharScanner_InterfaceOptions_ScanButtonLock:SetPoint("TOPLEFT", ScanButton, "BOTTOMLEFT", 10, 0)
	CharScanner_InterfaceOptions_ScanButtonLockText:SetText(L["Lock SCAN-Button"])
	CharScanner_InterfaceOptions_ScanButtonLock:SetChecked(CharScanner_Options.scanButtonLock)

	CharScanner_InterfaceOptions_ScanButton:SetScript("OnClick", function(self)
		CharScanner_Options.scanButton = not CharScanner_Options.scanButton
		if CharScanner_Options.scanButton then
			CharScanner_ScanButtonFrame:Show()
			BlizzardOptionsPanel_CheckButton_Enable(CharScanner_InterfaceOptions_ScanButtonLock, true)
		else
			CharScanner_ScanButtonFrame:Hide()
			BlizzardOptionsPanel_CheckButton_Disable(CharScanner_InterfaceOptions_ScanButtonLock)
		end
	end)
	CharScanner_InterfaceOptions_ScanButtonLock:SetScript("OnClick", function(self)
		CharScanner_Options.scanButtonLock = not CharScanner_Options.scanButtonLock
		if CharScanner_Options.scanButtonLock then
			CharScanner_ScanButtonFrame_Mover:Hide()
		else
			CharScanner_ScanButtonFrame:Show()
			CharScanner_ScanButtonFrame_Mover:Show()
		end
	end)

	if CharScanner_Options.scanButton then
		BlizzardOptionsPanel_CheckButton_Enable(CharScanner_InterfaceOptions_ScanButtonLock, true)
	else
		BlizzardOptionsPanel_CheckButton_Disable(CharScanner_InterfaceOptions_ScanButtonLock)
	end

	local chatOutput = CreateFrame("CheckButton", "CharScanner_InterfaceOptions_chatOutput", CharScanner_InterfaceOptions, "InterfaceOptionsCheckButtonTemplate")
	CharScanner_InterfaceOptions_chatOutput:SetPoint("TOPLEFT", ScanButtonLock, "BOTTOMLEFT", -10, -10)
	CharScanner_InterfaceOptions_chatOutputText:SetText(L["Chat output"])
	CharScanner_InterfaceOptions_chatOutput:SetChecked(CharScanner_Options.chatoutput)
	CharScanner_InterfaceOptions_chatOutput.tooltipText = L["chat_desc"]

	local chatOutputGrp = CreateFrame("CheckButton", "CharScanner_InterfaceOptions_chatOutputGrp", CharScanner_InterfaceOptions, "InterfaceOptionsCheckButtonTemplate")
	CharScanner_InterfaceOptions_chatOutputGrp:SetPoint("TOPLEFT", chatOutput, "BOTTOMLEFT", 10, 0)
	CharScanner_InterfaceOptions_chatOutputGrpText:SetText(L["GROUP scan: Show only one chat message"])
	CharScanner_InterfaceOptions_chatOutputGrp:SetChecked(CharScanner_Options.chatoutputGrp)
	CharScanner_InterfaceOptions_chatOutputGrp.tooltipText = L["chatgrp_desc"]

	CharScanner_InterfaceOptions_chatOutput:SetScript("OnClick", function(self)
		CharScanner_Options.chatoutput = not CharScanner_Options.chatoutput
		if CharScanner_Options.chatoutput then
			BlizzardOptionsPanel_CheckButton_Enable(CharScanner_InterfaceOptions_chatOutputGrp, true)
		else
			BlizzardOptionsPanel_CheckButton_Disable(CharScanner_InterfaceOptions_chatOutputGrp)
		end
	end)
	CharScanner_InterfaceOptions_chatOutputGrp:SetScript("OnClick", function(self)
		CharScanner_Options.chatoutputGrp = not CharScanner_Options.chatoutputGrp
	end)

	if CharScanner_Options.chatoutput then
		BlizzardOptionsPanel_CheckButton_Enable(CharScanner_InterfaceOptions_chatOutputGrp, true)
	else
		BlizzardOptionsPanel_CheckButton_Disable(CharScanner_InterfaceOptions_chatOutputGrp)
	end

	local scanduration = CreateFrame("Slider", "CharScanner_InterfaceOptions_scanduration", CharScanner_InterfaceOptions, "OptionsSliderTemplate")
	CharScanner_InterfaceOptions_scanduration:SetWidth(200)
	CharScanner_InterfaceOptions_scanduration:SetHeight(16)
	CharScanner_InterfaceOptions_scanduration:SetPoint("TOPLEFT", chatOutputGrp, "BOTTOMLEFT", -6, -30)
	CharScanner_InterfaceOptions_scandurationText:SetPoint("LEFT")
	CharScanner_InterfaceOptions_scandurationText:SetText(L["Maximum time for one scan"]..":")
	CharScanner_InterfaceOptions_scandurationLow:SetText("1")
	CharScanner_InterfaceOptions_scandurationHigh:SetText("2")
	CharScanner_InterfaceOptions_scanduration:SetMinMaxValues(10, 20)
	CharScanner_InterfaceOptions_scanduration:SetValueStep(1)
	CharScanner_InterfaceOptions_scanduration:SetValue(CharScanner_Options.scanDuration*10)
	local valuetext = scanduration:CreateFontString("CharScanner_InterfaceOptions_scanduration_valuetext", "OVERLAY", fonts.GameFontNormal.name)
	valuetext:SetPoint("TOP", scanduration, "BOTTOM", 0, 0)
	valuetext:SetJustifyH("CENTER")
	valuetext:SetText(CharScanner_Options.scanDuration)
	CharScanner_InterfaceOptions_scanduration:SetScript("OnValueChanged", function(self, value)
		value = value/10
		CharScanner_InterfaceOptions_scanduration_valuetext:SetText(value)
		CharScanner_Options.scanDuration = value
	end)
	CharScanner_InterfaceOptions_scanduration.tooltipText = L["scanduration_desc"]

	local minimapButton = CreateFrame("CheckButton", "CharScanner_InterfaceOptions_minimapButton", CharScanner_InterfaceOptions, "InterfaceOptionsCheckButtonTemplate")
	CharScanner_InterfaceOptions_minimapButton:SetPoint("TOPLEFT", scanduration, "BOTTOMLEFT", -4, -20)
	CharScanner_InterfaceOptions_minimapButtonText:SetText(L["Minimap-Button"])
	CharScanner_InterfaceOptions_minimapButton:SetChecked(CharScanner_Options.MinimapButton)
	CharScanner_InterfaceOptions_minimapButton.tooltipText = L["minimap_desc"]
	CharScanner_InterfaceOptions_minimapButton:SetScript("OnClick", function(self)
		CharScanner_Options.MinimapButton = not CharScanner_Options.MinimapButton
		CharScanner:CreateMinimapButton()
	end)

	InterfaceOptions_AddCategory(options)
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:InitializeFrame()
	-- initial Achievement & Statistics collapse menu appearance
	local opt = DefaultStatsViewData
	for i = 1, #opt do
		CharScanner:NodeClick(opt[i].lvl, opt[i].name1, opt[i].index1, opt[i].name2, opt[i].index2, opt[i].name3, opt[i].index3)
	end
	-- initial History/Compare sort button appearance
	CharScanner:UpdateSortButtons()
	-- initial Achievement Detail sort button appearance
	CharScanner:UpdateachievementdetailSortButtons()
end

function CharScanner:MainFrameToggle(show)
	if show then
		CharScanner_MainFrame:Show()
	else
		if CharScanner_MainFrame:IsShown() then
			CharScanner_MainFrame:Hide()
		else
			CharScanner_MainFrame:Show()
		end
	end
end

function CharScanner:ScanFrameShow()
	CharScanner_MainFrame_Base:Show()
	CharScanner_MainFrame_Talent:Show()
	CharScanner_MainFrame_Equipment:Show()
	CharScanner_MainFrame_Stats:Show()
	CharScanner_MainFrame_Summary:Show()
	CharScanner_MainFrame_Status:Show()
	CharScanner_MainFrame_Headline_ButtonCollapseNormal:SetTexture(Textures.Collapse.path)
	CharScanner_MainFrame_Headline_ButtonCollapseNormal:SetTexCoord(Textures.Collapse.coords[1], Textures.Collapse.coords[2], Textures.Collapse.coords[3], Textures.Collapse.coords[4])
	CharScanner_MainFrame_Headline_ButtonCollapsePush:SetTexture(Textures.Collapse.path)
	CharScanner_MainFrame_Headline_ButtonCollapsePush:SetTexCoord(Textures.Collapse.coords[1], Textures.Collapse.coords[2], Textures.Collapse.coords[3], Textures.Collapse.coords[4])
	if NumScannedChars == 0 then
		CharScanner_MainFrame_Headline_ButtonCollapse:Disable()
		CharScanner_MainFrame_Headline_ButtonCollapseBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
		CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
		CharScanner_MainFrame_Headline_ButtonHistory:Disable()
		CharScanner_MainFrame_Headline_ButtonHistoryBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
		CharScanner_MainFrame_Headline_ButtonCompare:Disable()
		CharScanner_MainFrame_Headline_ButtonCompareBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
		HistoryFrameIsShown = false
		CompareFrameIsShown = false
	else
		CharScanner_MainFrame_Headline_ButtonCollapse:Enable()
		CharScanner_MainFrame_Headline_ButtonCollapseBackground:SetTexture(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
		CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])
		CharScanner_MainFrame_Headline_ButtonHistory:Enable()
		CharScanner_MainFrame_Headline_ButtonHistoryBorderColor:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])
		CharScanner_MainFrame_Headline_ButtonCompare:Enable()
		CharScanner_MainFrame_Headline_ButtonCompareBorderColor:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])
	end
end

function CharScanner:ScanFrameHide()
	CharScanner_MainFrame_Base:Hide()
	CharScanner_MainFrame_Talent:Hide()
	CharScanner_MainFrame_Equipment:Hide()
	CharScanner_MainFrame_Stats:Hide()
	CharScanner_MainFrame_Summary:Hide()
	CharScanner_MainFrame_Status:Hide()
	CharScanner_HistoryFrame:Hide()
	CharScanner_MainFrame_Headline_ButtonCollapseBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
	CharScanner_MainFrame_Headline_ButtonHistory:Disable()
	CharScanner_MainFrame_Headline_ButtonHistoryBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
	CharScanner_MainFrame_Headline_ButtonHistoryBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	CharScanner_MainFrame_Headline_ButtonCompare:Disable()
	CharScanner_MainFrame_Headline_ButtonCompareBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
	CharScanner_MainFrame_Headline_ButtonCompareBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	CharScanner_MainFrame_Headline_ButtonCollapseNormal:SetTexture(Textures.Expand.path)
	CharScanner_MainFrame_Headline_ButtonCollapseNormal:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])
	CharScanner_MainFrame_Headline_ButtonCollapsePush:SetTexture(Textures.Expand.path)
	CharScanner_MainFrame_Headline_ButtonCollapsePush:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])
	HistoryFrameIsShown = false
	CompareFrameIsShown = false
	if NumScannedChars == 0 then
		CharScanner_MainFrame_Headline_ButtonCollapse:Disable()
		CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	else
		CharScanner_MainFrame_Headline_ButtonCollapse:Enable()
		CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])
	end
end

function CharScanner:ScanFrameToggle(show, hide)
	if show then
		CharScanner:ScanFrameShow()
		return
	end
	if hide then
		CharScanner:ScanFrameHide()
		return
	end
	if CharScanner_MainFrame_Base:IsShown() then
		CharScanner:ScanFrameHide()
	else
		CharScanner:ScanFrameShow()
	end
end

function CharScanner:CheckStatusFramePosition()
	if CharScanner_MainFrame_Base:IsShown() then
		CharScanner_MainFrame_Status:SetPoint("TOPLEFT", CharScanner_MainFrame_Summary, "BOTTOMLEFT", 0, sizeGeneral.Border)
	else
		CharScanner_MainFrame_Status:SetPoint("TOPLEFT", CharScanner_MainFrame_Headline, "BOTTOMLEFT", 0, sizeGeneral.Border)
	end
end

function CharScanner:DeleteButtonToggle(show)
	if show then
		CharScanner_MainFrame_Headline_ButtonDelete:Enable()
		CharScanner_MainFrame_Headline_ButtonDeleteBorderColor:SetTexture(0.55, 0, 0, 1)
	else
		CharScanner_MainFrame_Headline_ButtonDelete:Disable()
		CharScanner_MainFrame_Headline_ButtonDeleteBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	end
end

function CharScanner:GroupScanToggle()
	if inPartyOrRaid then
		CharScanner_MainFrame_Headline_ButtonGroupScan:Enable()
		CharScanner_MainFrame_Headline_ButtonGroupScanBorderColor:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])
		if CharScanner_MiniOptionsFrame then
			CharScanner_MiniOptionsFrame_GroupScanAdd:Enable()
			CharScanner_MiniOptionsFrame_GroupScanAddText:SetTextColor(color.Yellow.rgba[1], color.Yellow.rgba[2], color.Yellow.rgba[3], color.Yellow.rgba[4])
			CharScanner_MiniOptionsFrame_GroupScan:Enable()
			CharScanner_MiniOptionsFrame_GroupScanText:SetTextColor(color.Yellow.rgba[1], color.Yellow.rgba[2], color.Yellow.rgba[3], color.Yellow.rgba[4])
		end
	else
		CharScanner_MainFrame_Headline_ButtonGroupScan:Disable()
		CharScanner_MainFrame_Headline_ButtonGroupScanBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
		if CharScanner_MiniOptionsFrame then
			CharScanner_MiniOptionsFrame_GroupScanAdd:Disable()
			CharScanner_MiniOptionsFrame_GroupScanAddText:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
			CharScanner_MiniOptionsFrame_GroupScan:Disable()
			CharScanner_MiniOptionsFrame_GroupScanText:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
		end
	end
end

function CharScanner:HistoryFrameToggle()
	if CharScanner_HistoryFrame:IsShown() then
		if CompareFrameIsShown then
			HistoryFrameIsShown = true
			CompareFrameIsShown = false
			CharScanner:HistoryCompareReorganize()
		else
			HistoryFrameIsShown = false
			CharScanner_MainFrame_Headline_ButtonHistoryBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
			CharScanner_MainFrame_Headline_ButtonCompareBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
			CharScanner_HistoryFrame:Hide()
		end
	else
		HistoryFrameIsShown = true
		CharScanner_HistoryFrame:Show()
		CharScanner:HistoryCompareReorganize()
	end
end

function CharScanner:CompareFrameToggle()
	if CharScanner_HistoryFrame:IsShown() then
		if HistoryFrameIsShown then
			CompareFrameIsShown = true
			HistoryFrameIsShown = false
			CharScanner:HistoryCompareReorganize(true)
		else
			CompareFrameIsShown = false
			CharScanner_MainFrame_Headline_ButtonHistoryBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
			CharScanner_MainFrame_Headline_ButtonCompareBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
			CharScanner_HistoryFrame:Hide()
		end
	else
		CompareFrameIsShown = true
		CharScanner_HistoryFrame:Show()
		CharScanner:HistoryCompareReorganize(true)
	end
end

function CharScanner:CompareFrameShow()
	HistoryFrameIsShown = false
	CompareFrameIsShown = true
	CharScanner_HistoryFrame:Show()
	CharScanner:HistoryCompareReorganize(true)
end

function CharScanner:HistoryCompareReorganize(show)
	if show then
		CharScanner_MainFrame_Headline_ButtonHistoryBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
		CharScanner_MainFrame_Headline_ButtonCompareBackground:SetTexture(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
		CharScanner_HistoryFrame:SetWidth(sizeCompareFrame.CompareWidth)
		CharScanner_HistoryFrame_Sort:SetWidth(sizeCompareFrame.CompareWidth)
		CharScanner_HistoryFrame_Status:SetWidth(sizeCompareFrame.CompareWidth)
		for i = 1, GUI_history_maxNum do
			_G["CharScanner_HistoryFrame_Button"..i]:SetWidth(sizeCompareFrame.CompareWidth)
			_G["CharScanner_HistoryFrame_Button"..i.."_CompareFrame"]:Show()
		end
		CharScanner_HistoryFrame:Hide()
		CharScanner_HistoryFrame:Show()
		CharScanner_HistoryFrame_Sort_CompareFrame:Show()
	else
		CharScanner_MainFrame_Headline_ButtonHistoryBackground:SetTexture(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
		CharScanner_MainFrame_Headline_ButtonCompareBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
		CharScanner_HistoryFrame:SetWidth(sizeCompareFrame.HistoryWidth)
		CharScanner_HistoryFrame_Sort:SetWidth(sizeCompareFrame.HistoryWidth)
		CharScanner_HistoryFrame_Status:SetWidth(sizeCompareFrame.HistoryWidth)
		for i = 1, GUI_history_maxNum do
			_G["CharScanner_HistoryFrame_Button"..i]:SetWidth(sizeCompareFrame.HistoryWidth)
			_G["CharScanner_HistoryFrame_Button"..i.."_CompareFrame"]:Hide()
		end
		CharScanner_HistoryFrame_Sort_CompareFrame:Hide()
	end
end

function CharScanner:AchievementDetailClickedShow(buttonname, show)
	if show then
		_G[buttonname.."AchievementDetailIcon"]:SetAlpha(1)
		_G[buttonname.."AchievementDetailHighlight"]:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
	else
		_G[buttonname.."AchievementDetailIcon"]:SetAlpha(0.25)
		_G[buttonname.."AchievementDetailHighlight"]:SetTexture(0, 0, 0, 0)
	end
end

function CharScanner:AchievementDetailHide()
	if CharScanner_MainFrame_Stats_Title.AchievementClicked ~= "" then
		CharScanner:AchievementDetailClickedShow(CharScanner_MainFrame_Stats_Title.AchievementClicked, false)
		CharScanner_MainFrame_Stats_Title.AchievementClicked = ""
	end
	CharScanner_AchievementDetailFrame:Hide()
	AchievementDetailFrameIsShown = false
end

-- -----------------------------------------------------------------------------
function CharScanner:GearFrame_OnClick()
	if GearFrameIsShown then
		CharScanner_MainFrame_Gear:SetFrameLevel( CharScanner_MainFrame_Equipment:GetFrameLevel() )
		CharScanner_MainFrame_Gear:Hide()
		CharScanner_MainFrame_Stats:SetPoint("TOPLEFT", CharScanner_MainFrame_Equipment, "BOTTOMLEFT", 0, sizeGeneral.Border)
		CharScanner_MainFrame_Equipment_ButtonCollapseBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
		CharScanner_MainFrame_Equipment_ButtonCollapseNormal:SetTexture(Textures.Expand.path)
		CharScanner_MainFrame_Equipment_ButtonCollapseNormal:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])
		CharScanner_MainFrame_Equipment_ButtonCollapsePush:SetTexture(Textures.Expand.path)
		CharScanner_MainFrame_Equipment_ButtonCollapsePush:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])
		GearFrameIsShown = false
	else
		CharScanner_MainFrame_Gear:SetFrameLevel( CharScanner_MainFrame_Equipment:GetFrameLevel()+5 )
		CharScanner_MainFrame_Gear:Show()
		CharScanner_MainFrame_Stats:SetPoint("TOPLEFT", CharScanner_MainFrame_Gear, "BOTTOMLEFT", 0, sizeGeneral.Border)
		CharScanner_MainFrame_Equipment_ButtonCollapseBackground:SetTexture(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
		CharScanner_MainFrame_Equipment_ButtonCollapseNormal:SetTexture(Textures.Collapse.path)
		CharScanner_MainFrame_Equipment_ButtonCollapseNormal:SetTexCoord(Textures.Collapse.coords[1], Textures.Collapse.coords[2], Textures.Collapse.coords[3], Textures.Collapse.coords[4])
		CharScanner_MainFrame_Equipment_ButtonCollapsePush:SetTexture(Textures.Collapse.path)
		CharScanner_MainFrame_Equipment_ButtonCollapsePush:SetTexCoord(Textures.Collapse.coords[1], Textures.Collapse.coords[2], Textures.Collapse.coords[3], Textures.Collapse.coords[4])
		GearFrameIsShown = true
	end
end

function CharScanner:GearFrame_OnEnter()
	if not GearFrameIsShown then
		CharScanner_MainFrame_Gear:SetFrameLevel( CharScanner_MainFrame_Equipment:GetFrameLevel()+5 )
		CharScanner_MainFrame_Gear:Show()
	end
end

function CharScanner:GearFrame_OnLeave()
	if not GearFrameIsShown then
		CharScanner_MainFrame_Gear:SetFrameLevel( CharScanner_MainFrame_Equipment:GetFrameLevel() )
		CharScanner_MainFrame_Gear:Hide()
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:TalentTreeFrame_OnClick()
	CharScanner_MainFrame_TalentTree_GroupSelection_Button2:EnableMouse(true) -- fast MouseOver Button2 hackfix
	if TalentTreeFrameIsShown then
		CharScanner_MainFrame_TalentTree:SetFrameLevel( CharScanner_MainFrame_Talent:GetFrameLevel() )
		CharScanner_MainFrame_TalentTree:Hide()
		CharScanner_MainFrame_Equipment:SetPoint("TOPLEFT", CharScanner_MainFrame_Talent, "BOTTOMLEFT", 0, sizeGeneral.Border)
		CharScanner_MainFrame_Talent_ButtonCollapseBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
		CharScanner_MainFrame_Talent_ButtonCollapseNormal:SetTexture(Textures.Expand.path)
		CharScanner_MainFrame_Talent_ButtonCollapseNormal:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])
		CharScanner_MainFrame_Talent_ButtonCollapsePush:SetTexture(Textures.Expand.path)
		CharScanner_MainFrame_Talent_ButtonCollapsePush:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])
		TalentTreeFrameIsShown = false
	else
		CharScanner_MainFrame_TalentTree:SetFrameLevel( CharScanner_MainFrame_Talent:GetFrameLevel()+7 )
		CharScanner_MainFrame_TalentTree:Show()
		CharScanner_MainFrame_Equipment:SetPoint("TOPLEFT", CharScanner_MainFrame_TalentTree, "BOTTOMLEFT", 0, sizeGeneral.Border)
		CharScanner_MainFrame_Talent_ButtonCollapseBackground:SetTexture(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
		CharScanner_MainFrame_Talent_ButtonCollapseNormal:SetTexture(Textures.Collapse.path)
		CharScanner_MainFrame_Talent_ButtonCollapseNormal:SetTexCoord(Textures.Collapse.coords[1], Textures.Collapse.coords[2], Textures.Collapse.coords[3], Textures.Collapse.coords[4])
		CharScanner_MainFrame_Talent_ButtonCollapsePush:SetTexture(Textures.Collapse.path)
		CharScanner_MainFrame_Talent_ButtonCollapsePush:SetTexCoord(Textures.Collapse.coords[1], Textures.Collapse.coords[2], Textures.Collapse.coords[3], Textures.Collapse.coords[4])
		TalentTreeFrameIsShown = true
	end
end

function CharScanner:TalentTreeFrame_OnEnter()
	if not TalentTreeFrameIsShown then
		CharScanner_MainFrame_TalentTree:SetFrameLevel( CharScanner_MainFrame_Talent:GetFrameLevel()+7 )
		CharScanner_MainFrame_TalentTree:Show()
		CharScanner_MainFrame_TalentTree_GroupSelection_Button2:EnableMouse(false) -- fast MouseOver Button2 hackfix
	end
end

function CharScanner:TalentTreeFrame_OnLeave()
	if not TalentTreeFrameIsShown then
		CharScanner_MainFrame_TalentTree:SetFrameLevel( CharScanner_MainFrame_Talent:GetFrameLevel() )
		CharScanner_MainFrame_TalentTree:Hide()
		CharScanner_MainFrame_TalentTree_GroupSelection_Button2:EnableMouse(true) -- fast MouseOver Button2 hackfix
	end
end
-- -----------------------------------------------------------------------------

function CharScanner:ProtectionFrameToggle(show)
	if show then
		if CompareFrameIsShown then
			CharScanner_ProtectionHistoryFrame:SetWidth(sizeCompareFrame.CompareWidth)
		else
			CharScanner_ProtectionHistoryFrame:SetWidth(sizeCompareFrame.HistoryWidth)
		end
		CharScanner_ProtectionHistoryFrame:SetFrameLevel( CharScanner_MainFrame:GetFrameLevel()+1 )
		CharScanner_ProtectionHistoryFrame:Show()
		CharScanner_ProtectionAchievementDetailFrame:SetFrameLevel( CharScanner_MainFrame:GetFrameLevel()+1 )
		CharScanner_ProtectionAchievementDetailFrame:Show()
		CharScanner_ProtectionFrame:SetFrameLevel( CharScanner_MainFrame:GetFrameLevel()+2 )
		CharScanner_ProtectionFrame:Show()
	else
		CharScanner_ProtectionFrame:Hide()
		CharScanner_ProtectionHistoryFrame:Hide()
		CharScanner_ProtectionAchievementDetailFrame:Hide()
	end
end

function CharScanner:MiniOptionsToggle(button, pos)
	if button == "RightButton" then
		CharScanner:CreateMiniOptionsFrame()
		if CharScanner_MiniOptionsFrame:IsShown() then
			if CharScanner_MiniOptionsFrame.position and CharScanner_MiniOptionsFrame.position ~= pos then

				if pos == 1 then
					local xpos, ypos = GetCursorPosition()
					CharScanner_MiniOptionsFrame:ClearAllPoints()
					CharScanner_MiniOptionsFrame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", xpos, ypos)
				else
					CharScanner_MiniOptionsFrame:ClearAllPoints()
					CharScanner_MiniOptionsFrame:SetPoint("TOP", CharScanner_MinimapButton, "BOTTOM", 0, 0)	
				end
				CharScanner_MiniOptionsFrame.position = pos
				MiniOptions_TimerStart = GetTime()
				return
			end

			CharScanner_MiniOptionsFrame:Hide()
			CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
		else

			if pos == 1 then
				local xpos, ypos = GetCursorPosition()
				CharScanner_MiniOptionsFrame:ClearAllPoints()
				CharScanner_MiniOptionsFrame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", xpos, ypos)
			else
				CharScanner_MiniOptionsFrame:ClearAllPoints()
				CharScanner_MiniOptionsFrame:SetPoint("TOP", CharScanner_MinimapButton, "BOTTOM", 0, 0)	
			end

			CharScanner_MiniOptionsFrame.position = pos

			CharScanner_MiniOptionsFrame:Show()
			MiniOptions_TimerStart = GetTime()
			CharScanner_MiniOptionsFrame:SetScript("OnUpdate", MiniOptionsTimer)
		end
	else
		if CharScanner_MiniOptionsFrame then
			CharScanner_MiniOptionsFrame:Hide()
			CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
		end
		CharScanner:MainFrameToggle()
	end
end

function CharScanner:MainFramePos()
	if CharScanner_Options.MainFrame_relP then
		CharScanner_MainFrame:ClearAllPoints()
		CharScanner_MainFrame:SetPoint(CharScanner_Options.MainFrame_relP, CharScanner_Options.MainFrame_posX, CharScanner_Options.MainFrame_posY)
	else
		CharScanner_MainFrame:ClearAllPoints()
		CharScanner_MainFrame:SetPoint("CENTER", 0, 0)
	end
end

function CharScanner:MainFrameHide()
	PlaySound("igQuestListClose")
	CharScanner_InterfaceOptions_GUI:Enable()
end

function CharScanner:MainFrameShow()
	PlaySound("igQuestListOpen")
	CharScanner_InterfaceOptions_GUI:Disable()
	CharScanner:MainFramePos()
	CharScanner:UpdateHistory()
end

function CharScanner:SaveMainFramePos()
	local point, relativeTo, relativePoint, x, y = CharScanner_MainFrame:GetPoint()
	CharScanner_Options.MainFrame_posX = x
	CharScanner_Options.MainFrame_posY = y
	CharScanner_Options.MainFrame_relP = relativePoint
end

function CharScanner:SaveScanButtonPos()
	local point, relativeTo, relativePoint, x, y = CharScanner_ScanButtonFrame:GetPoint()
	CharScanner_Options.ScanButton_posX = x
	CharScanner_Options.ScanButton_posY = y
	CharScanner_Options.ScanButton_relP = relativePoint
end

function CharScanner:ScanButtonPos()
	if CharScanner_Options.ScanButton_relP then
		CharScanner_ScanButtonFrame:ClearAllPoints()
		CharScanner_ScanButtonFrame:SetPoint(CharScanner_Options.ScanButton_relP, CharScanner_Options.ScanButton_posX, CharScanner_Options.ScanButton_posY)
	else
		CharScanner_ScanButtonFrame:ClearAllPoints()
		CharScanner_ScanButtonFrame:SetPoint("TOP", 0, -50)
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
-- frame/button templates START
local function TemplateBorderTRBL(frame) -- TRBL = Top-Right-Bottom-Left
	local texture = frame:CreateTexture(nil, "BORDER")
	texture:SetPoint("TOPLEFT", sizeGeneral.Border, -sizeGeneral.Border)
	texture:SetPoint("BOTTOMRIGHT", -sizeGeneral.Border, sizeGeneral.Border)
	texture:SetTexture(color.FrameBG.rgba[1], color.FrameBG.rgba[2], color.FrameBG.rgba[3], color.FrameBG.rgba[4])
	texture.Background = frame:CreateTexture(nil, "BACKGROUND")
	texture.Background:SetPoint("TOPLEFT", 0, 0)
	texture.Background:SetPoint("BOTTOMRIGHT", 0, 0)
	texture.Background:SetTexture(color.FrameBorder.rgba[1], color.FrameBorder.rgba[2], color.FrameBorder.rgba[3], color.FrameBorder.rgba[4])
end

local function TemplateBorderRBL(frame) -- RBL = Right-Bottom-Left
	local name = frame:GetName()
	local texture = frame:CreateTexture(name.."Border", "BORDER")
	texture:SetPoint("TOPLEFT", sizeGeneral.Border, 0)
	texture:SetPoint("BOTTOMLEFT", 0, sizeGeneral.Border)
	texture:SetPoint("TOPRIGHT", -sizeGeneral.Border, 0)
	texture:SetPoint("BOTTOMRIGHT", -sizeGeneral.Border, -sizeGeneral.Border)
	texture:SetTexture(color.FrameBG.rgba[1], color.FrameBG.rgba[2], color.FrameBG.rgba[3], color.FrameBG.rgba[4])
	texture.Background = frame:CreateTexture(nil, "BACKGROUND")
	texture.Background:SetPoint("TOPLEFT", 0, 0)
	texture.Background:SetPoint("BOTTOMRIGHT", 0, 0)
	texture.Background:SetTexture(color.FrameBorder.rgba[1], color.FrameBorder.rgba[2], color.FrameBorder.rgba[3], color.FrameBorder.rgba[4])
end

local function TemplateHeadlineButton(button)
	local texture = button:CreateTexture(nil, "BORDER")
	texture:SetPoint("TOPLEFT", 1, -1)
	texture:SetPoint("BOTTOMRIGHT", -1, 1)
	texture:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
	local textureBackground = button:CreateTexture(nil, "BACKGROUND")
	textureBackground:SetPoint("TOPLEFT", 0, 0)
	textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
	textureBackground:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])

	local textureNormal = button:CreateTexture(nil, "ARTWORK")
	textureNormal:SetPoint("TOPLEFT", 3, -3)
	textureNormal:SetPoint("BOTTOMRIGHT", -3, 3)
	textureNormal:SetTexture(1, 1, 1, 0)
	button:SetNormalTexture(textureNormal)
	local textureHighlight = button:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetPoint("TOPLEFT", 1, -1)
	textureHighlight:SetPoint("BOTTOMRIGHT", -1, 1)
	textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
	button:SetHighlightTexture(textureHighlight)

	button:SetNormalFontObject(fonts.GameFontNormal.name)
	button:SetHighlightFontObject(fonts.GameFontHighlight.name)
end

local function TemplateSmallCollapsButton(button, which)
	local cut
	if which == 1 then
		cut = "GroupScanAdd"
	elseif which == 2 then
		cut = "History"
	elseif which == 3 then
		cut = "Compare"
	elseif which == 4 then
		cut = "Delete"
	elseif which == 5 then
		cut = "GroupScan"
	elseif which == 6 then
		cut = "Expand"
	elseif which == 7 then
		cut = "SingleScan"
	elseif which == 8 then
		cut = "Options"
	elseif which == 9 then
		cut = "Close"
	end

	button:SetWidth(sizeGeneral.Border + sizeGeneral.Space + Textures[cut].width + sizeGeneral.Space + sizeGeneral.Border)
	button:SetHeight(sizeGeneral.MainButtonHeight)

	local name = button:GetName()
	local texture = button:CreateTexture(name.."Background", "BORDER")
	texture:SetPoint("TOPLEFT", 1, -1)
	texture:SetPoint("BOTTOMRIGHT", -1, 1)
	texture:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
	local textureBackground = button:CreateTexture(name.."BorderColor", "BACKGROUND")
	textureBackground:SetPoint("TOPLEFT", 0, 0)
	textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
	textureBackground:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])

	local textureNormal = button:CreateTexture(name.."Normal", "ARTWORK")
	textureNormal:SetPoint("TOPLEFT", 3, -3)
	textureNormal:SetPoint("BOTTOMRIGHT", -3, 3)
	textureNormal:SetTexture(Textures[cut].path)
	textureNormal:SetTexCoord(Textures[cut].coords[1], Textures[cut].coords[2], Textures[cut].coords[3], Textures[cut].coords[4])
	button:SetNormalTexture(textureNormal)

	local texturePush = button:CreateTexture(name.."Push", "ARTWORK")
	texturePush:SetPoint("TOPLEFT", 4, -4)
	texturePush:SetPoint("BOTTOMRIGHT", -4, 4)
	texturePush:SetTexture(Textures[cut].path)
	texturePush:SetTexCoord(Textures[cut].coords[1], Textures[cut].coords[2], Textures[cut].coords[3], Textures[cut].coords[4])
	button:SetPushedTexture(texturePush)

	local textureDisable = button:CreateTexture(name.."Disable", "ARTWORK")
	textureDisable:SetPoint("TOPLEFT", 3, -3)
	textureDisable:SetPoint("BOTTOMRIGHT", -3, 3)
	textureDisable:SetTexture(Textures[cut].path)
	textureDisable:SetTexCoord(Textures[cut].coords[1], Textures[cut].coords[2], Textures[cut].coords[3], Textures[cut].coords[4])
	button:SetDisabledTexture(textureDisable)

	local shaderSupported = textureDisable:SetDesaturated(true)
	if not shaderSupported then
		textureDisable:SetVertexColor(0.5, 0.5, 0.5)
	end

	if which == 4 then
		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 1, -1)
		textureHighlight:SetPoint("BOTTOMRIGHT", -1, 1)
		textureHighlight:SetTexture(color.ButtonHighlightRed.rgba[1], color.ButtonHighlightRed.rgba[2], color.ButtonHighlightRed.rgba[3], color.ButtonHighlightRed.rgba[4])
		button:SetHighlightTexture(textureHighlight)
	else
		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 1, -1)
		textureHighlight:SetPoint("BOTTOMRIGHT", -1, 1)
		textureHighlight:SetTexture(color.ButtonHighlightLow.rgba[1], color.ButtonHighlightLow.rgba[2], color.ButtonHighlightLow.rgba[3], color.ButtonHighlightLow.rgba[4])
		button:SetHighlightTexture(textureHighlight)
	end

	button:SetNormalFontObject(fonts.GameFontNormal.name)
	button:SetHighlightFontObject(fonts.GameFontHighlight.name)
end

local function TemplateGroupSelectionButton(button)
	local name = button:GetName()
	local texture = button:CreateTexture(name.."Border", "BORDER")
	texture:SetPoint("TOPLEFT", sizeGeneral.Border, -sizeGeneral.Border)
	texture:SetPoint("BOTTOMRIGHT", -sizeGeneral.Border, sizeGeneral.Border)
	texture:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
	local textureBackground = button:CreateTexture(nil, "BACKGROUND")
	textureBackground:SetPoint("TOPLEFT", 0, 0)
	textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
	textureBackground:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])
	local textureHighlight = button:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetPoint("TOPLEFT", 3, -3)
	textureHighlight:SetPoint("BOTTOMRIGHT", -3, 3)
	textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
	button:SetHighlightTexture(textureHighlight)

	button:SetNormalFontObject(fonts.GameFontNormalSmall.name)
	button:SetHighlightFontObject(fonts.GameFontWhiteSmall.name)
end

local function TemplateStats(button, collapse, level, l1, l2, l3)
	local buttonname = button:GetName()
	local collapseWidth = 0
	local nameWidth = sizeStatFrame.NameWidth
	local nameLeft = sizeStatFrame.NameLeft
	local achievementWidth = sizeStatFrame.AchievementWidth
	local col = "White"

	if level == "title" then
		col = "Yellow"
		nameWidth = sizeStatFrame.NameWidthTitle
		achievementWidth = sizeStatFrame.AchievementWidthTitle
	elseif level == "level1" then
		col = "Level1"
		if collapse then
			nameWidth = sizeStatFrame.NameWidth_L1c
			nameLeft = sizeStatFrame.NameLeft_L1c
		end
	elseif level == "level2" then
		col = "Level2"
		nameWidth = sizeStatFrame.NameWidth_L2
		nameLeft = sizeStatFrame.NameLeft_L2
		if collapse then
			nameWidth = sizeStatFrame.NameWidth_L2c
			nameLeft = sizeStatFrame.NameLeft_L2c
			collapseWidth = sizeStatFrame.LevelModifier
		end
	elseif level == "level3" then
		col = "Level3"
		nameWidth = sizeStatFrame.NameWidth_L3
		nameLeft = sizeStatFrame.NameLeft_L3
		if collapse then
			nameWidth = sizeStatFrame.NameWidth_L3c
			nameLeft = sizeStatFrame.NameLeft_L3c
			collapseWidth = sizeStatFrame.LevelModifier*2
		end
	elseif level == "level4" then
		nameWidth = sizeStatFrame.NameWidth_L4
		nameLeft = sizeStatFrame.NameLeft_L4
	end

	button:SetFrameLevel( CharScanner_MainFrame_Stats:GetFrameLevel()+1 )
	button:SetWidth(sizeStatFrame.ButtonWidth)
	button:SetHeight(sizeGeneral.BarHeight)

	local nameButton = button:CreateFontString(buttonname.."Name", "OVERLAY", fonts.GameFontNormalSmall.name)
	nameButton:SetWidth(nameWidth)
	nameButton:SetHeight(fonts.GameFontNormalSmall.height)
	nameButton:SetPoint("LEFT", nameLeft, 0)
	nameButton:SetJustifyH("LEFT")
	nameButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])

	local bossestotalButton = button:CreateFontString(buttonname.."BossesTotal", "OVERLAY", fonts.GameFontNormalSmall.name)
	bossestotalButton:SetWidth(sizeStatFrame.BossesTotalWidth)
	bossestotalButton:SetHeight(fonts.GameFontNormalSmall.height)
	bossestotalButton:SetPoint("RIGHT", -sizeStatFrame.BossesTotalRight, 0)
	bossestotalButton:SetJustifyH("RIGHT")
	bossestotalButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])

	if collapse then
		local collapseButton = button:CreateTexture(buttonname.."Collapse", "ARTWORK")
		collapseButton:SetWidth(Textures.Plus.width)
		collapseButton:SetHeight(Textures.Plus.height)
		collapseButton:SetPoint("LEFT", collapseWidth, 0)
		collapseButton:SetTexture(Textures.Plus.path)
		collapseButton:SetTexCoord(Textures.Plus.coords[1], Textures.Plus.coords[2], Textures.Plus.coords[3], Textures.Plus.coords[4])

		local achievementButton = CreateFrame("Button", buttonname.."AchievementDetail", button)
		achievementButton:SetWidth(achievementWidth + sizeStatFrame.AchievementDetailWidth + sizeStatFrame.Space)
		achievementButton:SetHeight(sizeGeneral.BarHeight)
		achievementButton:SetPoint("RIGHT", -sizeStatFrame.AchievementDetailRight, 0)
		achievementButton:SetFrameLevel( button:GetFrameLevel()+1 )

		local achievementButtonHighlight = achievementButton:CreateTexture(buttonname.."AchievementDetailHighlight", "ARTWORK")
		achievementButtonHighlight:SetPoint("TOPLEFT", button, "TOPRIGHT", -(sizeStatFrame.AchievementRight+achievementWidth), 0)
		achievementButtonHighlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", -sizeStatFrame.AchievementRight, 0)
		achievementButtonHighlight:SetTexture(0, 0, 0, 0)

		local achievementButtonIcon = achievementButton:CreateTexture(buttonname.."AchievementDetailIcon", "ARTWORK")
		achievementButtonIcon:SetWidth(Textures.AchievementDetail.width)
		achievementButtonIcon:SetHeight(Textures.AchievementDetail.height)
		achievementButtonIcon:SetPoint("RIGHT", 0, 0)
		achievementButtonIcon:SetTexture(Textures.AchievementDetail.path)
		achievementButtonIcon:SetTexCoord(Textures.AchievementDetail.coords[1], Textures.AchievementDetail.coords[2], Textures.AchievementDetail.coords[3], Textures.AchievementDetail.coords[4])
		achievementButtonIcon:SetAlpha(0.25) 

		local achievementButtonText = achievementButton:CreateFontString(buttonname.."Achievement", "OVERLAY", fonts.GameFontNormalSmall.name)
		achievementButtonText:SetWidth(achievementWidth)
		achievementButtonText:SetHeight(fonts.GameFontNormalSmall.height)
		achievementButtonText:SetPoint("RIGHT", achievementButtonIcon, "LEFT", -sizeStatFrame.Space, 0)
		achievementButtonText:SetJustifyH("RIGHT")
		achievementButtonText:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])

		local function OnEnter()
			if LatestScanIsComplete then
				if CharScanner_MainFrame_Stats_Title.AchievementClicked ~= buttonname then
					CharScanner:AchievementDetailClickedShow(buttonname, true)
					CharScanner:UpdateAchievementDetail(l1, l2, l3)
				end
			end
		end

		local function OnLeave()
			if LatestScanIsComplete then
				if CharScanner_MainFrame_Stats_Title.AchievementClicked ~= buttonname then
					CharScanner:AchievementDetailClickedShow(buttonname, false)
					CharScanner_AchievementDetailFrame:Hide()
					AchievementDetailFrameIsShown = false
					table.wipe(GUI_achievementdetail_data) -- TEST TODO
				end
				if CharScanner_MainFrame_Stats_Title.AchievementClicked ~= "" then
					CharScanner:UpdateAchievementDetail(CharScanner_MainFrame_Stats_Title.AchievementL1, CharScanner_MainFrame_Stats_Title.AchievementL2, CharScanner_MainFrame_Stats_Title.AchievementL3)
				end
			end
		end

		local function OnClick()
			if LatestScanIsComplete then
				if CharScanner_MainFrame_Stats_Title.AchievementClicked == "" then
					CharScanner_MainFrame_Stats_Title.AchievementClicked = buttonname
					CharScanner_MainFrame_Stats_Title.AchievementL1 = l1
					CharScanner_MainFrame_Stats_Title.AchievementL2 = l2
					CharScanner_MainFrame_Stats_Title.AchievementL3 = l3
					CharScanner:AchievementDetailClickedShow(buttonname, true)
					CharScanner:UpdateAchievementDetail(l1, l2, l3)
				elseif CharScanner_MainFrame_Stats_Title.AchievementClicked == buttonname then
					CharScanner_MainFrame_Stats_Title.AchievementClicked = ""
					CharScanner_MainFrame_Stats_Title.AchievementL1 = nil
					CharScanner_MainFrame_Stats_Title.AchievementL2 = nil
					CharScanner_MainFrame_Stats_Title.AchievementL3 = nil
					CharScanner:AchievementDetailClickedShow(buttonname, false)
					CharScanner_AchievementDetailFrame:Hide()
					AchievementDetailFrameIsShown = false
				elseif  CharScanner_MainFrame_Stats_Title.AchievementClicked ~= buttonname then
					CharScanner:AchievementDetailClickedShow(CharScanner_MainFrame_Stats_Title.AchievementClicked, false)
					CharScanner_MainFrame_Stats_Title.AchievementClicked = buttonname
					CharScanner_MainFrame_Stats_Title.AchievementL1 = l1
					CharScanner_MainFrame_Stats_Title.AchievementL2 = l2
					CharScanner_MainFrame_Stats_Title.AchievementL3 = l3
					CharScanner:AchievementDetailClickedShow(buttonname, true)
					CharScanner:UpdateAchievementDetail(l1, l2, l3)
				end
			end
		end

		achievementButton:SetScript("OnEnter", OnEnter)
		achievementButton:SetScript("OnLeave", OnLeave)
		achievementButton:SetScript("OnClick", OnClick)

		local RaidSelectButton = CreateFrame("Button", buttonname.."RaidSelect", button)
		RaidSelectButton:SetWidth(Textures.RaidSelect.width)
		RaidSelectButton:SetHeight(Textures.RaidSelect.height)
		RaidSelectButton:SetPoint("RIGHT", button, "RIGHT", 0, 0)
		RaidSelectButton:SetFrameLevel( button:GetFrameLevel()+1 )
		local RaidSelectButtonTexture = RaidSelectButton:CreateTexture(nil, "ARTWORK")--buttonname.."RaidSelectTexture", "ARTWORK")
		RaidSelectButtonTexture:SetWidth(Textures.RaidSelect.width)
		RaidSelectButtonTexture:SetHeight(Textures.RaidSelect.height)
		RaidSelectButtonTexture:SetPoint("LEFT", 0, 0)
		RaidSelectButtonTexture:SetTexture(Textures.RaidSelect.path)
		RaidSelectButtonTexture:SetTexCoord(Textures.RaidSelect.coords[1], Textures.RaidSelect.coords[2], Textures.RaidSelect.coords[3], Textures.RaidSelect.coords[4])
		RaidSelectButtonTexture:SetAlpha(0.25) 

		RaidSelectButton:SetScript("OnEnter", function() RaidSelectButtonTexture:SetAlpha(1) end)
		RaidSelectButton:SetScript("OnLeave", function() RaidSelectButtonTexture:SetAlpha(0.25) end)
		RaidSelectButton:SetScript("OnClick", function() CharScanner:CompareFrameShow() CharScanner:GUIsort(GUI_history_SortBy, l1, l2, l3) end)
	end

	if level == "title" or level == "summary" then
		local achievementButton = button:CreateFontString(buttonname.."Achievement", "OVERLAY", fonts.GameFontNormalSmall.name)
		achievementButton:SetWidth(achievementWidth)
		achievementButton:SetHeight(fonts.GameFontNormalSmall.height)
		achievementButton:SetPoint("RIGHT", -sizeStatFrame.AchievementRight, 0)
		achievementButton:SetJustifyH("RIGHT")
		achievementButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])
	end
	if collapse or level == "title" or level == "summary" then
		local statisticButton = button:CreateFontString(buttonname.."Statistic", "OVERLAY", fonts.GameFontNormalSmall.name)
		statisticButton:SetWidth(sizeStatFrame.StatisticWidth)
		statisticButton:SetHeight(fonts.GameFontNormalSmall.height)
		statisticButton:SetPoint("RIGHT", -sizeStatFrame.StatisticRight, 0)
		statisticButton:SetJustifyH("RIGHT")
		statisticButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])
	end

	if level ~= "title" and level ~= "summary" then
		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 0, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
		textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
		button:SetHighlightTexture(textureHighlight)
		local texture = button:CreateTexture(nil, "BACKGROUND")
		texture:SetPoint("TOPLEFT", nameLeft, -5)
		texture:SetPoint("BOTTOMRIGHT", -sizeStatFrame.BossesTotalRight, 5)
		texture:SetTexture(color.ButtonLineMarker.rgba[1], color.ButtonLineMarker.rgba[2], color.ButtonLineMarker.rgba[3], color.ButtonLineMarker.rgba[4])
		button:SetNormalTexture(texture)
	end
--[[ TEST
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "LEFT", nameLeft, 0)
		texture:SetPoint("RIGHT", button, "LEFT", nameWidth+nameLeft, 0)
		texture:SetTexture(0.22, 0.22, 0.22, 1)
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.AchievementRight+achievementWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.AchievementRight, 0)
		texture:SetTexture(0.22, 0.22, 0.22, 1)
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.AchievementDetailRight+sizeStatFrame.AchievementDetailWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.AchievementDetailRight, 0)
		texture:SetTexture(0.22, 0.22, 0.22, 1)
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.StatisticRight+sizeStatFrame.StatisticWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.StatisticRight, 0)
		texture:SetTexture(0.22, 0.22, 0.22, 1)
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.BossesTotalRight+sizeStatFrame.BossesTotalWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.BossesTotalRight, 0)
		texture:SetTexture(0.22, 0.22, 0.22, 1)
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.RaidSelectRight+sizeStatFrame.RaidSelectWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.RaidSelectRight, 0)
		texture:SetTexture(0.22, 0.22, 0.22, 1)
--]]
end

local function TemplateCompareSelect(button, level, height)
	local name = button:GetName()
	local col = "White"
	local buttonLeft = 0

	if level == "title" then
		buttonLeft = sizeGeneral.Offset
		col = "Yellow"
	elseif level == "level1" then
		buttonLeft = 0
		col = "Level1"
	elseif level == "level2" then
		buttonLeft = sizeStatFrame.LevelModifier
		col = "Level2"
	elseif level == "level3" then
		buttonLeft = sizeStatFrame.LevelModifier * 2
		col = "Level3"
	end
	button:SetHeight(height)
	
	local nameButton = button:CreateFontString(name.."Name", "OVERLAY", fonts.GameFontNormalSmall.name)
	nameButton:SetHeight(fonts.GameFontNormalSmall.height)
	nameButton:SetPoint("LEFT", buttonLeft, 0)
	nameButton:SetJustifyH("LEFT")
	nameButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])

	if level == "title" then
		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 0, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
		textureHighlight:SetTexture(color.ButtonHighlightLow.rgba[1], color.ButtonHighlightLow.rgba[2], color.ButtonHighlightLow.rgba[3], color.ButtonHighlightLow.rgba[4])
		button:SetHighlightTexture(textureHighlight)
	else
		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 0, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
		textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
		button:SetHighlightTexture(textureHighlight)
	end
end

local function TemplateHistoryListSortButton(button, text, pos, width)
	local name = button:GetName()

	local textureBackground = button:CreateTexture(nil, "BACKGROUND")
	textureBackground:SetPoint("TOPLEFT", 0, 0)
	textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
	textureBackground:SetTexture(color.ButtonSortBG.rgba[1], color.ButtonSortBG.rgba[2], color.ButtonSortBG.rgba[3], color.ButtonSortBG.rgba[4])

	local buttontext
	if text then
		if pos == "right" then
			buttontext = button:CreateFontString(name.."Text", "OVERLAY", fonts.GameFontNormalSmall.name)
			buttontext:SetText(text)
			local fw = buttontext:GetStringWidth()
			if fw >= width then
				buttontext:SetWidth( width-10 )
			end
			buttontext:SetHeight(fonts.GameFontNormalSmall.height)
			buttontext:SetPoint("RIGHT", button, "RIGHT", 0, 0)
			buttontext:SetJustifyH("RIGHT")
		else
			buttontext = button:CreateFontString(name.."Text", "OVERLAY", fonts.GameFontNormalSmall.name)
			buttontext:SetText(text)
			local fw = buttontext:GetStringWidth()
			if fw >= width then
				buttontext:SetWidth( width-10 )
			end
			buttontext:SetHeight(fonts.GameFontNormalSmall.height)
			buttontext:SetPoint("LEFT", button, "LEFT", 0, 0)
			buttontext:SetJustifyH("LEFT")
		end
	end

	local texture = button:CreateTexture(name.."Arrow", "BACKGROUND")
	texture:SetWidth(Textures.SortArrow.width)
	texture:SetHeight(Textures.SortArrow.height)
	texture:SetTexture(Textures.SortArrow.path)
	if text then
		if pos == "right" then
			texture:SetPoint("RIGHT", buttontext, "LEFT", -1, 0)
		else
			texture:SetPoint("LEFT", buttontext, "RIGHT", 1, 0)
		end
	else
		texture:SetPoint("CENTER", button, "CENTER", 0, 0)
	end

	local textureHighlight = button:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetPoint("TOPLEFT", 0, 0)
	textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
	textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
	button:SetHighlightTexture(textureHighlight)
end

local function TemplateHistoryList(button, buttonid, isSort)
	local name = button:GetName()

	if isSort then
		local sort_icon = CreateFrame("Button", name.."Button_Icon", button)
		TemplateHistoryListSortButton(sort_icon, nil, nil)
		sort_icon:SetWidth(sizeGeneral.BarHeight)
		sort_icon:SetHeight(sizeGeneral.BarHeight)
		sort_icon:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.ClassiconLeft, 0)
		sort_icon:SetScript("OnClick", function() CharScanner:GUIsort(1) end)

		local sort_name = CreateFrame("Button", name.."Button_Name", button)
		TemplateHistoryListSortButton(sort_name, NAME, "left", sizeCompareFrame.NameWidth)
		sort_name:SetWidth(sizeCompareFrame.NameWidth)
		sort_name:SetHeight(sizeGeneral.BarHeight)
		sort_name:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.NameLeft, 0)
		sort_name:SetScript("OnClick", function() CharScanner:GUIsort(2) end)

		local sort_time = CreateFrame("Button", name.."Button_Time", button)
		TemplateHistoryListSortButton(sort_time, L["Time"], "left", sizeCompareFrame.TimeWidth)
		sort_time:SetWidth(sizeCompareFrame.TimeWidth)
		sort_time:SetHeight(sizeGeneral.BarHeight)
		sort_time:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.TimeLeft, 0)
		sort_time:SetScript("OnClick", function() CharScanner:GUIsort(3) end)

		local sort_mainrole = CreateFrame("Button", name.."Button_MainRole", button)
		TemplateHistoryListSortButton(sort_mainrole, nil, nil)--, sizeCompareFrame.MainroleWidth)
		sort_mainrole:SetWidth(sizeCompareFrame.MainroleWidth)
		sort_mainrole:SetHeight(sizeGeneral.BarHeight)
		sort_mainrole:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.MainroleLeft, 0)
		sort_mainrole:SetScript("OnClick", function() CharScanner:GUIsort(4) end)

		local frame = CreateFrame("Frame", name.."_CompareFrame", button)
		frame:SetToplevel(true)

		local sort_talents1 = CreateFrame("Button", name.."Button_Talents1", frame)
		TemplateHistoryListSortButton(sort_talents1, TALENT_SPEC_PRIMARY, "left", sizeCompareFrame.Talents1Width)
		sort_talents1:SetWidth(sizeCompareFrame.Talents1Width)
		sort_talents1:SetHeight(sizeGeneral.BarHeight)
		sort_talents1:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.Talents1Left, 0)
		sort_talents1:SetScript("OnClick", function() CharScanner:GUIsort(5) end)

		local sort_talents2 = CreateFrame("Button", name.."Button_Talents2", frame)
		TemplateHistoryListSortButton(sort_talents2, TALENT_SPEC_SECONDARY, "left", sizeCompareFrame.Talents2Width)
		sort_talents2:SetWidth(sizeCompareFrame.Talents2Width)
		sort_talents2:SetHeight(sizeGeneral.BarHeight)
		sort_talents2:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.Talents2Left, 0)
		sort_talents2:SetScript("OnClick", function() CharScanner:GUIsort(6) end)

		local sort_itemlevel = CreateFrame("Button", name.."Button_ItemLevel", frame)
		TemplateHistoryListSortButton(sort_itemlevel, L["iLvl"], "right", sizeCompareFrame.EquipWidth)
		sort_itemlevel:SetWidth(sizeCompareFrame.EquipWidth)
		sort_itemlevel:SetHeight(sizeGeneral.BarHeight)
		sort_itemlevel:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.EquipLeft, 0)
		sort_itemlevel:SetScript("OnClick", function() CharScanner:GUIsort(7) end)

		local sort_achievement = CreateFrame("Button", name.."Button_Achievement", frame)
		TemplateHistoryListSortButton(sort_achievement, L["Achievements"], "right", sizeCompareFrame.AchievementWidth)
		sort_achievement:SetWidth(sizeCompareFrame.AchievementWidth)
		sort_achievement:SetHeight(sizeGeneral.BarHeight)
		sort_achievement:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.AchievementLeft, 0)
		sort_achievement:SetScript("OnClick", function() CharScanner:GUIsort(8) end)

		local sort_boss = CreateFrame("Button", name.."Button_Boss", frame)
		TemplateHistoryListSortButton(sort_boss, L["Bosses"], "right", sizeCompareFrame.BossWidth)
		sort_boss:SetWidth(sizeCompareFrame.BossWidth)
		sort_boss:SetHeight(sizeGeneral.BarHeight)
		sort_boss:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.BossLeft, 0)
		sort_boss:SetScript("OnClick", function() CharScanner:GUIsort(9) end)

		local sort_kills = CreateFrame("Button", name.."Button_Kills", frame)
		TemplateHistoryListSortButton(sort_kills, L["Kills"], "right", sizeCompareFrame.KillsWidth)
		sort_kills:SetWidth(sizeCompareFrame.KillsWidth)
		sort_kills:SetHeight(sizeGeneral.BarHeight)
		sort_kills:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.KillsLeft, 0)
		sort_kills:SetScript("OnClick", function() CharScanner:GUIsort(10) end)

		-- Raid Selection START
		local button = CreateFrame("Button", name.."_RaidSelectMainButton", frame)
		TemplateBorderTRBL(button)
		TemplateCompareSelect(button, "title", sizeCompareFrame.SortLine1Height)
		button:SetPoint("TOPRIGHT", CharScanner_HistoryFrame, "TOPRIGHT", -sizeCompareFrame.HighlightLeft, -sizeCompareFrame.Offset)
		_G[name.."_RaidSelectMainButtonName"]:SetText(L["Total"])
		button:SetScript("OnClick",
			function()
				if _G[name.."_RaidSelectFrame"]:IsShown() then
					_G[name.."_RaidSelectFrame"]:Hide()
					_G[name.."_RaidSelectMainButton_Expand"]:SetTexture(Textures.Expand.path)
					_G[name.."_RaidSelectMainButton_Expand"]:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])
				else
					_G[name.."_RaidSelectFrame"]:Show()
					_G[name.."_RaidSelectMainButton_Expand"]:SetTexture(Textures.Collapse.path)
					_G[name.."_RaidSelectMainButton_Expand"]:SetTexCoord(Textures.Collapse.coords[1], Textures.Collapse.coords[2], Textures.Collapse.coords[3], Textures.Collapse.coords[4])
				end
			end)

		local buttontexture = button:CreateTexture(name.."_RaidSelectMainButton_Expand", "ARTWORK")
		buttontexture:SetHeight(16)
		buttontexture:SetWidth(16)
		buttontexture:SetPoint("RIGHT", button, "RIGHT", -sizeGeneral.Offset, 0)
		buttontexture:SetTexture(Textures.Expand.path)
		buttontexture:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])

		local historyraidselect = CreateFrame("Frame", name.."_RaidSelectFrame", frame)
		TemplateBorderTRBL(historyraidselect)
		historyraidselect:SetToplevel(true)
		historyraidselect:EnableMouse(true)
		historyraidselect:SetPoint("TOPLEFT", button, "BOTTOMLEFT", 0, sizeGeneral.Border)
		historyraidselect:Hide()

		local bnumber = 1
		local textwidth = 0
		local strw = 0

		local buttonframe = CreateFrame("Button", name.."_RaidSelect_Button"..bnumber, historyraidselect)
		TemplateCompareSelect(buttonframe, "level1", sizeGeneral.BarHeight)
		buttonframe:SetPoint("TOPLEFT", name.."_RaidSelectFrame", "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
		buttonframe:SetScript("OnClick", function() CharScanner:GUIsort(GUI_history_SortBy, 0) end)
		_G[name.."_RaidSelect_Button"..bnumber.."Name"]:SetText(L["Total"])
		strw = _G[name.."_RaidSelect_Button"..bnumber.."Name"]:GetStringWidth()
		if strw > textwidth then textwidth = strw end
		bnumber = bnumber + 1

		for i = 1, #DATA do
			if DATA[i].name then
				local buttonframe = CreateFrame("Button", name.."_RaidSelect_Button"..bnumber, historyraidselect)
				TemplateCompareSelect(buttonframe, "level1", sizeGeneral.BarHeight)
				buttonframe:SetPoint("TOPLEFT", name.."_RaidSelect_Button"..bnumber-1, "BOTTOMLEFT", 0, 0)
				buttonframe:SetScript("OnClick", function() CharScanner:GUIsort(GUI_history_SortBy, i) end)
				_G[name.."_RaidSelect_Button"..bnumber.."Name"]:SetText(L[DATA[i].name])
				strw = _G[name.."_RaidSelect_Button"..bnumber.."Name"]:GetStringWidth()
				if strw > textwidth then textwidth = strw end
				bnumber = bnumber + 1
			end

			for j = 1, #DATA[i] do

				if DATA[i][j].name then
					local buttonframe = CreateFrame("Button", name.."_RaidSelect_Button"..bnumber, historyraidselect)
					TemplateCompareSelect(buttonframe, "level2", sizeGeneral.BarHeight)
					buttonframe:SetPoint("TOPLEFT", name.."_RaidSelect_Button"..bnumber-1, "BOTTOMLEFT", 0, 0)
					buttonframe:SetScript("OnClick", function() CharScanner:GUIsort(GUI_history_SortBy, i, j) end)
					_G[name.."_RaidSelect_Button"..bnumber.."Name"]:SetText(L[DATA[i][j].name])
					strw = _G[name.."_RaidSelect_Button"..bnumber.."Name"]:GetStringWidth()
					if strw > textwidth then textwidth = strw end
					bnumber = bnumber + 1
				end

				for k = 1, #DATA[i][j] do
					if DATA[i][j][k].name then
						local buttonframe = CreateFrame("Button", name.."_RaidSelect_Button"..bnumber, historyraidselect)
						TemplateCompareSelect(buttonframe, "level3", sizeGeneral.BarHeight)
						buttonframe:SetPoint("TOPLEFT", name.."_RaidSelect_Button"..bnumber-1, "BOTTOMLEFT", 0, 0)
						buttonframe:SetScript("OnClick", function() CharScanner:GUIsort(GUI_history_SortBy, i, j, k) end)
						_G[name.."_RaidSelect_Button"..bnumber.."Name"]:SetText(L[DATA[i][j][k].name].." ("..L[DATA[i][j].name]..")")
						strw = _G[name.."_RaidSelect_Button"..bnumber.."Name"]:GetStringWidth()
						if strw > textwidth then textwidth = strw end
						bnumber = bnumber + 1
					end
				end

			end

		end

		historyraidselect:SetHeight(((bnumber-1)*sizeGeneral.BarHeight) + (2*sizeGeneral.Offset))
		
		-- set width
		textwidth = textwidth + 20

		CharScanner_HistoryFrame_Sort_RaidSelectMainButton:SetWidth(textwidth + (2*sizeGeneral.Offset))
		CharScanner_HistoryFrame_Sort_RaidSelectMainButtonName:SetWidth(textwidth + (2*sizeGeneral.Offset))
		CharScanner_HistoryFrame_Sort_RaidSelectFrame:SetWidth(textwidth + (2*sizeGeneral.Offset))

		for i = 1, bnumber-1 do
			_G[name.."_RaidSelect_Button"..i]:SetWidth(textwidth)
			_G[name.."_RaidSelect_Button"..i.."Name"]:SetWidth(textwidth)
		end
		-- set width
		-- Raid Selection END
	else
		local deleteButton = CreateFrame("Button", nil, button)
		deleteButton:SetWidth(Textures.Trashcan.width)
		deleteButton:SetHeight(Textures.Trashcan.height)
		deleteButton:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.TrashcanLeft, 0)
		local deleteButtonTexture = deleteButton:CreateTexture(nil, "ARTWORK")
		deleteButtonTexture:SetPoint("LEFT", 0, 0)
		deleteButtonTexture:SetWidth(Textures.Trashcan.width)
		deleteButtonTexture:SetHeight(Textures.Trashcan.height)
		deleteButtonTexture:SetTexture(Textures.Trashcan.path)
		deleteButtonTexture:SetTexCoord(Textures.Trashcan.coords[1], Textures.Trashcan.coords[2], Textures.Trashcan.coords[3], Textures.Trashcan.coords[4])
		deleteButtonTexture:SetAlpha(0.5)
		deleteButton:SetScript("OnEnter", function() deleteButtonTexture:SetAlpha(1) end)
		deleteButton:SetScript("OnLeave", function() deleteButtonTexture:SetAlpha(0.5) end)
		deleteButton:SetScript("OnClick", function() CharScanner:DeleteCharacter(button.guid, buttonid) end)

		local texture = button:CreateTexture(name.."Arrow", "BACKGROUND")
		texture:SetWidth(Textures.Selection.width)
		texture:SetHeight(Textures.Selection.height)
		texture:SetTexture(Textures.Selection.path)
		texture:SetTexCoord(Textures.Selection.coords[1], Textures.Selection.coords[2], Textures.Selection.coords[3], Textures.Selection.coords[4])
		texture:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.ArrowLeft, 0)
		texture:Hide()

		local iconButton = button:CreateTexture(name.."Icon", "ARTWORK")
		iconButton:SetWidth(sizeGeneral.BarHeight)
		iconButton:SetHeight(sizeGeneral.BarHeight)
		iconButton:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.ClassiconLeft, 0)
		iconButton:SetTexture(Textures.ClassIcon.path)

		local nameButton = button:CreateFontString(name.."Name", "ARTWORK", fonts.GameFontNormalSmall.name)
		nameButton:SetWidth(sizeCompareFrame.NameWidth)
		nameButton:SetHeight(fonts.GameFontNormalSmall.height)
		nameButton:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.NameLeft, 0)
		nameButton:SetJustifyH("LEFT")

		local nameTime = button:CreateFontString(name.."Time", "ARTWORK", fonts.GameFontWhiteSmall.name)
		nameTime:SetWidth(sizeCompareFrame.TimeWidth)
		nameTime:SetHeight(fonts.GameFontWhiteSmall.height)
		nameTime:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.TimeLeft, 0)
		nameTime:SetJustifyH("LEFT")

		local nameMainRole = button:CreateFontString(name.."MainRole", "ARTWORK", fonts.GameFontWhiteSmall.name)
		nameMainRole:SetWidth(sizeCompareFrame.MainroleWidth)
		nameMainRole:SetHeight(fonts.GameFontWhiteSmall.height)
		nameMainRole:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.MainroleLeft, 0)
		nameMainRole:SetJustifyH("CENTER")

		local frame = CreateFrame("Frame", name.."_CompareFrame", button)
		frame:SetToplevel(true)

		local nameTalents1 = frame:CreateFontString(name.."Talents1", "ARTWORK", fonts.GameFontWhiteSmall.name)
		nameTalents1:SetWidth(sizeCompareFrame.Talents1Width)
		nameTalents1:SetHeight(fonts.GameFontWhiteSmall.height)
		nameTalents1:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.Talents1Left, 0)
		nameTalents1:SetJustifyH("LEFT")

		local nameTalents2 = frame:CreateFontString(name.."Talents2", "ARTWORK", fonts.GameFontWhiteSmall.name)
		nameTalents2:SetWidth(sizeCompareFrame.Talents2Width)
		nameTalents2:SetHeight(fonts.GameFontWhiteSmall.height)
		nameTalents2:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.Talents2Left, 0)
		nameTalents2:SetJustifyH("LEFT")

		local nameItemLevel = frame:CreateFontString(name.."ItemLevel", "ARTWORK", fonts.GameFontWhiteSmall.name)
		nameItemLevel:SetWidth(sizeCompareFrame.EquipWidth)
		nameItemLevel:SetHeight(fonts.GameFontWhiteSmall.height)
		nameItemLevel:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.EquipLeft, 0)
		nameItemLevel:SetJustifyH("RIGHT")

		local nameAchievement = frame:CreateFontString(name.."Achievement", "ARTWORK", fonts.GameFontWhiteSmall.name)
		nameAchievement:SetWidth(sizeCompareFrame.AchievementWidth)
		nameAchievement:SetHeight(fonts.GameFontWhiteSmall.height)
		nameAchievement:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.AchievementLeft, 0)
		nameAchievement:SetJustifyH("RIGHT")

		local nameBoss = frame:CreateFontString(name.."Boss", "ARTWORK", fonts.GameFontWhiteSmall.name)
		nameBoss:SetWidth(sizeCompareFrame.BossWidth)
		nameBoss:SetHeight(fonts.GameFontWhiteSmall.height)
		nameBoss:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.BossLeft, 0)
		nameBoss:SetJustifyH("RIGHT")

		local nameKills = frame:CreateFontString(name.."Kills", "ARTWORK", fonts.GameFontWhiteSmall.name)
		nameKills:SetWidth(sizeCompareFrame.KillsWidth)
		nameKills:SetHeight(fonts.GameFontWhiteSmall.height)
		nameKills:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.KillsLeft, 0)
		nameKills:SetJustifyH("RIGHT")

		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", button, "TOPLEFT", sizeCompareFrame.Offset, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", -sizeCompareFrame.HighlightLeft, 0)
		textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
		button:SetHighlightTexture(textureHighlight)
		local texture = button:CreateTexture(nil, "BACKGROUND")
		texture:SetPoint("TOPLEFT", button, "TOPLEFT", sizeCompareFrame.NameLeft, -5)
		texture:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", -sizeCompareFrame.HighlightLeft, 5)
		texture:SetTexture(color.ButtonLineMarker.rgba[1], color.ButtonLineMarker.rgba[2], color.ButtonLineMarker.rgba[3], color.ButtonLineMarker.rgba[4])
		button:SetNormalTexture(texture)

		button:SetScript("OnClick", function() CharScanner:PreShowSavedData(button.guid, true) end)
	end
end

local function TemplateAchievementDetailList(button, isSort)
	local buttonname = button:GetName()

	if isSort then
		-- AchievementDetailSelect START
		local selection = CreateFrame("Button", buttonname.."_AchievementDetailSelectMainButton", button)
		TemplateBorderTRBL(selection)
		TemplateCompareSelect(selection, "title", sizeAchievementDetail.SortLine1Height)
		selection:SetPoint("TOPLEFT", CharScanner_AchievementDetailFrame_Sort, "TOPLEFT", 0, sizeAchievementDetail.SortLine1Height+sizeAchievementDetail.Space)
		_G[buttonname.."_AchievementDetailSelectMainButtonName"]:SetText(ALL)
		_G[buttonname.."_AchievementDetailSelectMainButtonName"]:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
		selection:SetScript("OnClick",
			function()
				if _G[buttonname.."_AchievementDetailSelectFrame"]:IsShown() then
					_G[buttonname.."_AchievementDetailSelectFrame"]:Hide()
					_G[buttonname.."_AchievementDetailSelectMainButton_Expand"]:SetTexture(Textures.Expand.path)
					_G[buttonname.."_AchievementDetailSelectMainButton_Expand"]:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])
				else
					_G[buttonname.."_AchievementDetailSelectFrame"]:Show()
					_G[buttonname.."_AchievementDetailSelectMainButton_Expand"]:SetTexture(Textures.Collapse.path)
					_G[buttonname.."_AchievementDetailSelectMainButton_Expand"]:SetTexCoord(Textures.Collapse.coords[1], Textures.Collapse.coords[2], Textures.Collapse.coords[3], Textures.Collapse.coords[4])
				end
			end)

		local buttontexture = selection:CreateTexture(buttonname.."_AchievementDetailSelectMainButton_Expand", "ARTWORK")
		buttontexture:SetHeight(16)
		buttontexture:SetWidth(16)
		buttontexture:SetPoint("RIGHT", selection, "RIGHT", -sizeGeneral.Offset, 0)
		buttontexture:SetTexture(Textures.Expand.path)
		buttontexture:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])

		local selectionFrame = CreateFrame("Frame", buttonname.."_AchievementDetailSelectFrame", selection)
		TemplateBorderTRBL(selectionFrame)
		selectionFrame:SetToplevel(true)
		selectionFrame:EnableMouse(true)
		selectionFrame:SetPoint("TOPLEFT", selection, "BOTTOMLEFT", 0, sizeGeneral.Border)
		selectionFrame:Hide()

		local textwidth = 0
		local strw = 0

		local buttonAll = CreateFrame("Button", buttonname.."_AchievementDetailSelect_Button1", selectionFrame)
		TemplateCompareSelect(buttonAll, "level1", sizeGeneral.BarHeight)
		buttonAll:SetPoint("TOPLEFT", buttonname.."_AchievementDetailSelectFrame", "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
		buttonAll:SetScript("OnClick", function() CharScanner:UpdateAchievementDetail(nil, nil, nil, 1) end)
		_G[buttonname.."_AchievementDetailSelect_Button1Name"]:SetText(ALL)
		_G[buttonname.."_AchievementDetailSelect_Button1Name"]:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
		strw = _G[buttonname.."_AchievementDetailSelect_Button1Name"]:GetStringWidth()
		if strw > textwidth then textwidth = strw end

		local buttonComplete = CreateFrame("Button", buttonname.."_AchievementDetailSelect_Button2", selectionFrame)
		TemplateCompareSelect(buttonComplete, "level1", sizeGeneral.BarHeight)
		buttonComplete:SetPoint("TOPLEFT", buttonAll, "BOTTOMLEFT", 0, 0)
		buttonComplete:SetScript("OnClick", function() CharScanner:UpdateAchievementDetail(nil, nil, nil, 2) end)
		_G[buttonname.."_AchievementDetailSelect_Button2Name"]:SetText(COMPLETE)
		_G[buttonname.."_AchievementDetailSelect_Button2Name"]:SetTextColor(color.Green.rgba[1], color.Green.rgba[2], color.Green.rgba[3], color.Green.rgba[4])
		strw = _G[buttonname.."_AchievementDetailSelect_Button2Name"]:GetStringWidth()
		if strw > textwidth then textwidth = strw end

		local buttonIncomplete = CreateFrame("Button", buttonname.."_AchievementDetailSelect_Button3", selectionFrame)
		TemplateCompareSelect(buttonIncomplete, "level1", sizeGeneral.BarHeight)
		buttonIncomplete:SetPoint("TOPLEFT", buttonComplete, "BOTTOMLEFT", 0, 0)
		buttonIncomplete:SetScript("OnClick", function() CharScanner:UpdateAchievementDetail(nil, nil, nil, 3) end)
		_G[buttonname.."_AchievementDetailSelect_Button3Name"]:SetText(INCOMPLETE)
		_G[buttonname.."_AchievementDetailSelect_Button3Name"]:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		strw = _G[buttonname.."_AchievementDetailSelect_Button3Name"]:GetStringWidth()
		if strw > textwidth then textwidth = strw end

		selectionFrame:SetHeight((3*sizeGeneral.BarHeight) + (2*sizeGeneral.Offset))
		
		-- set width
		textwidth = textwidth + 20

		_G[buttonname.."_AchievementDetailSelectMainButton"]:SetWidth(textwidth + (2*sizeGeneral.Offset))
		_G[buttonname.."_AchievementDetailSelectMainButtonName"]:SetWidth(textwidth + (2*sizeGeneral.Offset))
		_G[buttonname.."_AchievementDetailSelectFrame"]:SetWidth(textwidth + (2*sizeGeneral.Offset))

		for i = 1, 3 do
			_G[buttonname.."_AchievementDetailSelect_Button"..i]:SetWidth(textwidth)
			_G[buttonname.."_AchievementDetailSelect_Button"..i.."Name"]:SetWidth(textwidth)
		end
		-- set width
		-- AchievementDetailSelect END

		local sort_id = CreateFrame("Button", buttonname.."Button_Id", button)
		TemplateHistoryListSortButton(sort_id, "ID", "right", sizeAchievementDetail.IdWidth)
		sort_id:SetWidth(sizeAchievementDetail.IdWidth)
		sort_id:SetHeight(sizeGeneral.BarHeight)
		sort_id:SetPoint("LEFT", "CharScanner_AchievementDetailFrame_Sort", "LEFT", sizeAchievementDetail.IdRight, 0)
		sort_id:SetScript("OnClick", function() CharScanner:GUIachievementdetailsort(1) end)

		local sort_name = CreateFrame("Button", buttonname.."Button_Name", button)
		TemplateHistoryListSortButton(sort_name, NAME, "left", sizeAchievementDetail.NameWidth)
		sort_name:SetWidth(sizeAchievementDetail.NameWidth)
		sort_name:SetHeight(sizeGeneral.BarHeight)
		sort_name:SetPoint("LEFT", "CharScanner_AchievementDetailFrame_Sort", "LEFT", sizeAchievementDetail.NameRight, 0)
		sort_name:SetScript("OnClick", function() CharScanner:GUIachievementdetailsort(2) end)
	else
		local id = button:CreateFontString(buttonname.."Id", "ARTWORK", fonts.GameFontWhiteSmall.name)
		id:SetWidth(sizeAchievementDetail.IdWidth)
		id:SetHeight(fonts.GameFontNormalSmall.height)
		id:SetPoint("LEFT", button, "LEFT", sizeAchievementDetail.IdRight, 0)
		id:SetJustifyH("RIGHT")

		local name = button:CreateFontString(buttonname.."Name", "ARTWORK", fonts.GameFontNormalSmall.name)
		name:SetWidth(sizeAchievementDetail.NameWidth)
		name:SetHeight(fonts.GameFontNormalSmall.height)
		name:SetPoint("LEFT", button, "LEFT", sizeAchievementDetail.NameRight, 0)
		name:SetJustifyH("LEFT")

		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", button, "TOPLEFT", 0, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", -sizeAchievementDetail.HighlightLeft, 0)
		textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
		button:SetHighlightTexture(textureHighlight)

		local info = CreateFrame("Button", buttonname.."Info", button)
		TemplateBorderTRBL(info)
		info:SetWidth(sizeAchievementDetail.Width)
		info:SetPoint("TOPLEFT", CharScanner_AchievementDetailFrame, "BOTTOMLEFT", 0, -(sizeGeneral.StatusHeight-(2*sizeGeneral.Border)))
		info:Hide()
		info.texture = info:CreateTexture(nil, "ARTWORK")
		info.texture:SetWidth(sizeAchievementDetail.InfoTexture)
		info.texture:SetHeight(sizeAchievementDetail.InfoTexture)
		info.texture:SetPoint("TOPLEFT", info, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
		info.text = info:CreateFontString(buttonname.."InfoText", "OVERLAY", fonts.GameFontNormalSmall.name)
		info.text:SetWidth(sizeAchievementDetail.InfoTextWidth)
		info.text:SetPoint("TOPLEFT", info.texture, "TOPRIGHT", sizeGeneral.Offset, 0)
		info.text:SetJustifyH("LEFT")

		button:SetScript("OnEnter", function()
			local _, _, _, _, _, _, _, description, _, icon, _ = GetAchievementInfo(button.achievementID)
			info.text:SetText(description)
			info.texture:SetTexture(icon)
			local height = info.text:GetStringHeight()+(2*sizeGeneral.Offset)
			if height < sizeAchievementDetail.InfoTextminHeight then
				height = sizeAchievementDetail.InfoTextminHeight
			end
			info:SetHeight(height)
			info:Show()
		end)
		button:SetScript("OnLeave", function() info:Hide() end)
	end
end

local function TemplateTalentButton(button)
	button:SetWidth(Textures.TalentIconTexture.width)
	button:SetHeight(Textures.TalentIconTexture.height)
	local name = button:GetName()

	local texture = button:CreateTexture(name.."IconTexture", "BORDER")
	texture:SetWidth(Textures.TalentIconTexture.width)
	texture:SetHeight(Textures.TalentIconTexture.height)
	texture:SetPoint("TOPLEFT", 0, 0)

	local texture = button:CreateTexture(name.."RankBorder", "OVERLAY")
	texture:SetWidth(Textures.TalentRankBorder.width)
	texture:SetHeight(Textures.TalentRankBorder.height)
	texture:SetPoint("BOTTOMRIGHT", name.."IconTexture", "BOTTOMRIGHT", 1, -1)
	texture:SetTexture(Textures.TalentRankBorder.path)
	texture:SetTexCoord(Textures.TalentRankBorder.coords[1], Textures.TalentRankBorder.coords[2], Textures.TalentRankBorder.coords[3], Textures.TalentRankBorder.coords[4])
	texture:SetAlpha(0.8)

	local text = button:CreateFontString(name.."Rank", "OVERLAY", fonts.GameFontNormalSmall.name)
	text:SetPoint("CENTER", name.."RankBorder", "CENTER", 0, 0)
	text:SetJustifyH("RIGHT")

	local textureHighlight = button:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetWidth(Textures.Highlight.width)
	textureHighlight:SetHeight(Textures.Highlight.height)
	textureHighlight:SetPoint("TOPLEFT", 0, 0)
	textureHighlight:SetTexture(Textures.Highlight.path)
	button:SetHighlightTexture(textureHighlight)
end

local function TemplateGearButton(button, pos)
	button:SetWidth(Textures.GearIconTexture.width)
	button:SetHeight(Textures.GearIconTexture.height)
	local name = button:GetName()

	local texture = button:CreateTexture(name.."IconTexture", "BORDER")
	texture:SetWidth(Textures.GearIconTexture.width)
	texture:SetHeight(Textures.GearIconTexture.height)
	texture:SetPoint("TOPLEFT", 0, 0)

	local text = button:CreateFontString(name.."Count", "OVERLAY", fonts.GameFontNormalSmall.name)
	if pos == 1 then -- LEFT
		text:SetPoint("RIGHT", texture, "LEFT", -sizeEquipmentFrame.Offset, 0)
	elseif pos == 2 then -- RIGHT
		text:SetPoint("LEFT", texture, "RIGHT", sizeEquipmentFrame.Offset, 0)
	elseif pos == 3 then -- BOTTOM
		text:SetPoint("TOP", texture, "BOTTOM", 0, -sizeEquipmentFrame.Offset)
	end

	local enchant = button:CreateTexture(name.."EnchTexture", "BORDER")
	enchant:SetWidth(Textures.Enchantment.width)
	enchant:SetHeight(Textures.Enchantment.height)
	if pos == 1 then -- LEFT
		enchant:SetPoint("RIGHT", text, "LEFT", -sizeEquipmentFrame.Offset, 0)
	elseif pos == 2 then -- RIGHT
		enchant:SetPoint("LEFT", text, "RIGHT", sizeEquipmentFrame.Offset, 0)
	elseif pos == 3 then -- BOTTOM
		enchant:SetPoint("TOP", text, "BOTTOM", 0, -sizeEquipmentFrame.Offset)
	end
	enchant:SetTexture(Textures.Enchantment.path)

	local textureHighlight = button:CreateTexture(name.."Highlight", "ARTWORK")
	textureHighlight:SetWidth(Textures.Highlight.width)
	textureHighlight:SetHeight(Textures.Highlight.height)
	textureHighlight:SetPoint("TOPLEFT", 0, 0)
	textureHighlight:SetTexture(Textures.Highlight.path)
	button:SetHighlightTexture(textureHighlight)
end
-- frame/button templates END

function CharScanner:CreateMainFrame()
	local mainframe = CreateFrame("Frame", "CharScanner_MainFrame", UIParent)
	mainframe:EnableMouse(true)
	mainframe:SetMovable(true)
	mainframe:SetToplevel(true)
	mainframe:SetClampedToScreen(true)
	mainframe:Hide()
	mainframe:SetWidth(sizeGeneral.MainWidth)
	mainframe:SetHeight(sizeGeneral.MainHeight)
	mainframe:SetScript("OnShow", function() CharScanner:MainFrameShow() end)
	mainframe:SetScript("OnHide", function() CharScanner:MainFrameHide() end)

	-- Headline Frame START
	local headline = CreateFrame("Frame", "CharScanner_MainFrame_Headline", CharScanner_MainFrame)
	TemplateBorderTRBL(headline)
	headline:SetWidth(sizeGeneral.MainWidth)
	headline:SetHeight(sizeGeneral.MainHeight)
	headline:SetPoint("TOPLEFT", CharScanner_MainFrame, "TOPLEFT", 0, 0)
	headline:EnableMouse(true)

	local headlineButtonScan = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonScan", headline)
	TemplateSmallCollapsButton(headlineButtonScan, 7)
	headlineButtonScan:SetPoint("LEFT", headline, "LEFT", sizeGeneral.MainOffset, 0)
	headlineButtonScan:SetScript("OnClick", function() CharScanner:ScanUnit("target") end)

	local headlineButtonGroupScan = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonGroupScan", headline)
	TemplateSmallCollapsButton(headlineButtonGroupScan, 1)
	headlineButtonGroupScan:SetPoint("TOPLEFT", headlineButtonScan, "TOPRIGHT", sizeGeneral.MainOffset, 0)
	headlineButtonGroupScan:Disable()
	CharScanner_MainFrame_Headline_ButtonGroupScanBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])

	local headlineButtonGroupScan2 = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonGroupScan2", headline)
	TemplateSmallCollapsButton(headlineButtonGroupScan2, 5)
	headlineButtonGroupScan2:SetPoint("BOTTOM", headlineButtonGroupScan, "TOP", 0, 0)
	CharScanner_MainFrame_Headline_ButtonGroupScan2BorderColor:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])
	headlineButtonGroupScan2:Hide()

	headlineButtonGroupScan:SetScript("OnClick", function() headlineButtonGroupScan2:Hide() CharScanner:ScanGroup(false) end)
	headlineButtonGroupScan:SetScript("OnEnter",
		function()
			headlineButtonGroupScan2:Show()
		end)
	headlineButtonGroupScan:SetScript("OnLeave",
		function()
			if not headlineButtonGroupScan2:IsMouseOver() then
				headlineButtonGroupScan2:Hide()
			end
		end)

	headlineButtonGroupScan2:SetScript("OnClick", function() headlineButtonGroupScan2:Hide() CharScanner:ScanGroup(true) end)
	headlineButtonGroupScan2:SetScript("OnLeave",
		function()
			if not headlineButtonGroupScan:IsMouseOver() then
				headlineButtonGroupScan2:Hide()
			end
		end)

	local headlineButtonCollapse = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonCollapse", headline)
	TemplateSmallCollapsButton(headlineButtonCollapse, 6)
	headlineButtonCollapse:SetPoint("TOPLEFT", headlineButtonGroupScan, "TOPRIGHT", sizeGeneral.MainOffset, 0)
	headlineButtonCollapse:Disable()
	CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	headlineButtonCollapse:SetScript("OnClick", function() CharScanner:ScanFrameToggle() end)

	local headlineButtonHistory = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonHistory", headline)
	TemplateSmallCollapsButton(headlineButtonHistory, 2)
	headlineButtonHistory:SetPoint("LEFT", headlineButtonCollapse, "RIGHT", sizeGeneral.MainOffset, 0)
	headlineButtonHistory:Disable()
	CharScanner_MainFrame_Headline_ButtonHistoryBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	headlineButtonHistory:SetScript("OnClick", function() CharScanner:HistoryFrameToggle() end)

	local headlineButtonCompare = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonCompare", headline)
	TemplateSmallCollapsButton(headlineButtonCompare, 3)
	headlineButtonCompare:SetPoint("LEFT", headlineButtonHistory, "RIGHT", sizeGeneral.MainOffset, 0)
	headlineButtonCompare:Disable()
	CharScanner_MainFrame_Headline_ButtonCompareBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	headlineButtonCompare:SetScript("OnClick", function() CharScanner:CompareFrameToggle() end)

	local headlineButtonX = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonX", headline)
	TemplateSmallCollapsButton(headlineButtonX, 9)
	headlineButtonX:SetPoint("RIGHT", headline, "RIGHT", -sizeGeneral.MainOffset, 0)
	headlineButtonX:SetScript("OnClick", function() CharScanner:MainFrameToggle() end)

	local headlineButtonOptions = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonOptions", headline)
	TemplateSmallCollapsButton(headlineButtonOptions, 8)
	headlineButtonOptions:SetPoint("RIGHT", headlineButtonX, "LEFT", -sizeGeneral.MainOffset, 0)
	headlineButtonOptions:SetScript("OnClick", function() CharScanner:MainFrameToggle() InterfaceOptionsFrame_OpenToCategory("CharScanner") end)

	local headlineButtonDelete = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonDelete", headline)
	TemplateSmallCollapsButton(headlineButtonDelete, 4)
	headlineButtonDelete:SetPoint("RIGHT", headlineButtonOptions, "LEFT", -sizeGeneral.MainOffset, 0)
	headlineButtonDelete:Disable()
	CharScanner_MainFrame_Headline_ButtonDeleteBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	headlineButtonDelete:SetScript("OnClick", function() CharScanner:DeleteCharacter() end)

	local headlineButtonMover = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonMover", headline)
	headlineButtonMover:SetWidth(22)
	headlineButtonMover:SetHeight(22)
	headlineButtonMover:SetPoint("TOPLEFT", headlineButtonCompare, "TOPRIGHT", sizeGeneral.MainOffset, 0)
	headlineButtonMover:SetPoint("BOTTOMRIGHT", headlineButtonDelete, "BOTTOMLEFT", -sizeGeneral.MainOffset, 0)
	local textureMover = headlineButtonMover:CreateTexture(nil, "BACKGROUND")
	textureMover:SetWidth(1)
	textureMover:SetHeight(1)
	textureMover:SetPoint("TOPLEFT", headlineButtonMover, "TOPLEFT", 0, 0)
	textureMover:SetPoint("BOTTOMRIGHT", headlineButtonMover, "BOTTOMRIGHT", 0, 0)
	textureMover:SetTexture(1, 1, 1, 0)
	headlineButtonMover:SetScript("OnEnter", function() textureMover:SetTexture(1, 1, 1, 0.075) end)
	headlineButtonMover:SetScript("OnLeave", function() textureMover:SetTexture(1, 1, 1, 0) end)
	headlineButtonMover:SetScript("OnMouseDown", function() CharScanner_MainFrame:StartMoving() end)
	headlineButtonMover:SetScript("OnMouseUp", function() CharScanner_MainFrame:StopMovingOrSizing() CharScanner:SaveMainFramePos() end)
	-- Headline Frame END

	-- Base Character Information Frame START
	local base = CreateFrame("Frame", "CharScanner_MainFrame_Base", mainframe)
	TemplateBorderTRBL(base)
	base:SetWidth(sizeGeneral.MainWidth)
	base:SetHeight(sizeBaseCharFrame_BaseHeight)
	base:SetPoint("TOPLEFT", headline, "BOTTOMLEFT", 0, sizeGeneral.Border)
	base:EnableMouse(true)
	base:SetScript("OnShow", function() CharScanner:CheckStatusFramePosition() end)
	base:Hide()

	local baseClassIcon = base:CreateTexture("CharScanner_MainFrame_Base_ClassIcon", "ARTWORK")
	baseClassIcon:SetWidth(28)
	baseClassIcon:SetHeight(28)
	baseClassIcon:SetPoint("LEFT", sizeGeneral.Offset, 0)
	baseClassIcon:SetTexture(Textures.ClassIcon.path)

	local baseFactionIcon = base:CreateTexture("CharScanner_MainFrame_Base_FactionIcon", "ARTWORK")
	baseFactionIcon:SetWidth(18)
	baseFactionIcon:SetHeight(18)
	baseFactionIcon:SetPoint("LEFT", baseClassIcon, "RIGHT", 0, 0)

	local baseName = base:CreateFontString("CharScanner_MainFrame_Base_Name", "ARTWORK", fonts.GameFontNormal.name)
	baseName:SetPoint("TOPLEFT", base, "TOPLEFT", sizeGeneral.Offset+28+18+sizeGeneral.Space, -sizeGeneral.Offset)
	baseName:SetJustifyH("LEFT")
	baseName:SetTextColor(1, 1, 1, 1)
	local baseLevel = base:CreateFontString("CharScanner_MainFrame_Base_Level", "ARTWORK", fonts.GameFontNormalSmall.name)
	baseLevel:SetPoint("TOPLEFT", baseName, "BOTTOMLEFT", 0, 0)
	baseLevel:SetJustifyH("LEFT")
	baseLevel:SetTextColor(1, 1, 1, 1)
	local baseGuild = base:CreateFontString("CharScanner_MainFrame_Base_Guild", "ARTWORK", fonts.GameFontNormalSmall.name)
	baseGuild:SetPoint("TOPLEFT", baseLevel, "BOTTOMLEFT", 0, 0)
	baseGuild:SetJustifyH("LEFT")
	baseGuild:SetTextColor(1, 1, 1, 1)
	-- Base Character Information Frame END

	-- Talent Frame START
	local talent = CreateFrame("Frame", "CharScanner_MainFrame_Talent", mainframe)
	TemplateBorderTRBL(talent)
	talent:SetWidth(sizeGeneral.MainWidth)
	talent:SetHeight(sizeTalentFrame.BaseHeight)
	talent:SetPoint("TOPLEFT", base, "BOTTOMLEFT", 0, sizeGeneral.Border)
	talent:EnableMouse(true)
	talent:Hide()

	local talentTitle = talent:CreateFontString("CharScanner_MainFrame_Talent_Title", "ARTWORK", fonts.GameFontNormalSmall.name)
	talentTitle:SetPoint("TOPLEFT", talent, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
	talentTitle:SetJustifyH("LEFT")
	talentTitle:SetText(TALENTS..":")
	
	local talentGroup1Icon = talent:CreateTexture("CharScanner_MainFrame_Talent_TalentGroup1Icon", "ARTWORK")
	talentGroup1Icon:SetWidth(Textures.TalentMiniIcons.width)
	talentGroup1Icon:SetHeight(Textures.TalentMiniIcons.height)
	talentGroup1Icon:SetPoint("TOPLEFT", talent, "TOPLEFT", sizeGeneral.Offset+talentTitle:GetStringWidth()+sizeGeneral.Offset, -sizeGeneral.Offset)
	talentGroup1Icon:SetTexCoord(Textures.TalentMiniIcons.coords[1], Textures.TalentMiniIcons.coords[2], Textures.TalentMiniIcons.coords[3], Textures.TalentMiniIcons.coords[4])
	local talentGroup1Text = talent:CreateFontString("CharScanner_MainFrame_Talent_TalentGroup1Text", "ARTWORK", fonts.GameFontNormalSmall.name)
	talentGroup1Text:SetPoint("BOTTOM", talentTitle, "BOTTOM", 0, 0)
	talentGroup1Text:SetPoint("LEFT", talentGroup1Icon, "RIGHT", sizeGeneral.Offset, 0)
	talentGroup1Text:SetJustifyH("LEFT")

	local talentGroup2Icon = talent:CreateTexture("CharScanner_MainFrame_Talent_TalentGroup2Icon", "ARTWORK")
	talentGroup2Icon:SetWidth(Textures.TalentMiniIcons.width)
	talentGroup2Icon:SetHeight(Textures.TalentMiniIcons.height)
	talentGroup2Icon:SetPoint("TOPLEFT", talentGroup1Icon, "BOTTOMLEFT", 0, 0)
	talentGroup2Icon:SetTexCoord(Textures.TalentMiniIcons.coords[1], Textures.TalentMiniIcons.coords[2], Textures.TalentMiniIcons.coords[3], Textures.TalentMiniIcons.coords[4])
	local talentGroup2Text = talent:CreateFontString("CharScanner_MainFrame_Talent_TalentGroup2Text", "ARTWORK", fonts.GameFontNormalSmall.name)
	talentGroup2Text:SetPoint("LEFT", talentGroup2Icon, "RIGHT", sizeGeneral.Offset, 0)
	talentGroup2Text:SetJustifyH("LEFT")

	local talentButtonCollapse = CreateFrame("Button", "CharScanner_MainFrame_Talent_ButtonCollapse", talent)
	TemplateSmallCollapsButton(talentButtonCollapse, 6)
	talentButtonCollapse:SetPoint("RIGHT", talent, "RIGHT", 0, 0)
	talentButtonCollapse:SetScript("OnClick", function() CharScanner:TalentTreeFrame_OnClick() end)
	talentButtonCollapse:SetScript("OnEnter", function() CharScanner:TalentTreeFrame_OnEnter() end)
	talentButtonCollapse:SetScript("OnLeave", function() CharScanner:TalentTreeFrame_OnLeave() end)

	-- TalentTree Frame START
	local TalentTree = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree", talent)
	TemplateBorderRBL(TalentTree)
	TalentTree:SetFrameLevel( CharScanner_MainFrame_Talent:GetFrameLevel() )
	TalentTree:SetWidth(sizeGeneral.MainWidth)
	TalentTree:SetHeight(sizeTalentFrame.TreeBaseHeight)
	TalentTree:SetPoint("TOPLEFT", talent, "BOTTOMLEFT", 0, sizeGeneral.Border)
	TalentTree:EnableMouse(true)
	TalentTree:Hide()

	local GroupSelection = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree_GroupSelection", TalentTree)
	TemplateBorderRBL(GroupSelection)
	GroupSelection:SetWidth(sizeGeneral.MainWidth)
	GroupSelection:SetHeight(sizeTalentFrame.GroupSelectionHeight)
	GroupSelection:SetPoint("TOPLEFT", TalentTree, "TOPLEFT", 0, 0)

	local GroupSelection_Button1 = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_GroupSelection_Button1", GroupSelection)
	TemplateGroupSelectionButton(GroupSelection_Button1)
	GroupSelection_Button1:SetWidth(sizeTalentFrame.GroupSelectionWidth)
	GroupSelection_Button1:SetHeight(sizeTalentFrame.GroupSelectionHeight)
	GroupSelection_Button1:SetPoint("BOTTOMLEFT", GroupSelection, "BOTTOMLEFT", sizeTalentFrame.GroupSelection1Left, 0)
	GroupSelection_Button1:SetText(TALENT_SPEC_PRIMARY)
	GroupSelection_Button1:Hide()

	local GroupSelection_Button2 = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_GroupSelection_Button2", GroupSelection)
	TemplateGroupSelectionButton(GroupSelection_Button2)
	GroupSelection_Button2:SetWidth(sizeTalentFrame.GroupSelectionWidth)
	GroupSelection_Button2:SetHeight(sizeTalentFrame.GroupSelectionHeight)
	GroupSelection_Button2:SetPoint("BOTTOMLEFT", GroupSelection, "BOTTOMLEFT", sizeTalentFrame.GroupSelection2Left, 0)
	GroupSelection_Button2:SetText(TALENT_SPEC_SECONDARY)
	GroupSelection_Button2:Hide()

	local TalentTree1 = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree1", TalentTree)
	TalentTree1:SetWidth(sizeTalentFrame.TreeWidth)
	TalentTree1:SetHeight(sizeTalentFrame.TreeHeight)
	TalentTree1:SetPoint("TOPLEFT", GroupSelection, "BOTTOMLEFT", sizeTalentFrame.Tree1Left, -sizeTalentFrame.Offset)
	local TalentTree1Name = TalentTree1:CreateFontString("CharScanner_MainFrame_TalentTree1Name", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree1Name:SetPoint("TOP", TalentTree1, "TOP", 0, 0)
	TalentTree1Name:SetJustifyH("CENTER")
	local TalentTree1Points = TalentTree1:CreateFontString("CharScanner_MainFrame_TalentTree1Points", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree1Points:SetPoint("TOP", TalentTree1Name, "BOTTOM", 0, 0)
	TalentTree1Points:SetJustifyH("CENTER")

	local TalentTree2 = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree2", TalentTree)
	TalentTree2:SetWidth(sizeTalentFrame.TreeWidth)
	TalentTree2:SetHeight(sizeTalentFrame.TreeHeight)
	TalentTree2:SetPoint("TOPLEFT", GroupSelection, "BOTTOMLEFT", sizeTalentFrame.Tree2Left, -sizeTalentFrame.Offset)
	local TalentTree2Name = TalentTree2:CreateFontString("CharScanner_MainFrame_TalentTree2Name", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree2Name:SetPoint("TOP", TalentTree2, "TOP", 0, 0)
	TalentTree2Name:SetJustifyH("CENTER")
	local TalentTree2Points = TalentTree2:CreateFontString("CharScanner_MainFrame_TalentTree2Points", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree2Points:SetPoint("TOP", TalentTree2Name, "BOTTOM", 0, 0)
	TalentTree2Points:SetJustifyH("CENTER")

	local TalentTree3 = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree3", TalentTree)
	TalentTree3:SetWidth(sizeTalentFrame.TreeWidth)
	TalentTree3:SetHeight(sizeTalentFrame.TreeHeight)
	TalentTree3:SetPoint("TOPLEFT", GroupSelection, "BOTTOMLEFT", sizeTalentFrame.Tree3Left, -sizeTalentFrame.Offset)
	local TalentTree3Name = TalentTree3:CreateFontString("CharScanner_MainFrame_TalentTree3Name", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree3Name:SetPoint("TOP", TalentTree3, "TOP", 0, 0)
	TalentTree3Name:SetJustifyH("CENTER")
	local TalentTree3Points = TalentTree3:CreateFontString("CharScanner_MainFrame_TalentTree3Points", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree3Points:SetPoint("TOP", TalentTree3Name, "BOTTOM", 0, 0)
	TalentTree3Points:SetJustifyH("CENTER")

	local tier = 1
	local column = 1
	for tab = 1, myMAX_TALENT_TABS do
		for talent = 1, myMAX_NUM_TALENTS do
			local button
			if talent == 1 then
				tier = 1
				column = 1
				button = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column, TalentTree)
				button:SetPoint("TOPLEFT", "CharScanner_MainFrame_TalentTree"..tab, "TOPLEFT", sizeTalentFrame.Offset, fonts.GameFontWhiteSmall.height*-2) -- -(12*-2) font height
			elseif mod(talent, myMAX_COLUMN) == 1 then
				tier = tier + 1
				column = 1
				button = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column, TalentTree)
				button:SetPoint("TOP", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..(tier-1).."_"..(column), "BOTTOM", 0, -sizeTalentFrame.Offset)
			else
				column = column + 1
				button = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column, TalentTree)
				button:SetPoint("LEFT", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..(column-1), "RIGHT", sizeTalentFrame.Offset, 0)
			end
			TemplateTalentButton(button)
		end
	end
	-- TalentTree Frame END
	-- Talent Frame END

	-- Equipment Frame START
	local equipment = CreateFrame("Frame", "CharScanner_MainFrame_Equipment", mainframe)
	TemplateBorderTRBL(equipment)
	equipment:SetWidth(sizeGeneral.MainWidth)
	equipment:SetHeight(sizeEquipmentFrame.BaseHeight)
	equipment:SetPoint("TOPLEFT", talent, "BOTTOMLEFT", 0, sizeGeneral.Border)
	equipment:EnableMouse(true)
	equipment:Hide()

	local equipmentTitle = equipment:CreateFontString("CharScanner_MainFrame_Equipment_Title", "ARTWORK", fonts.GameFontNormalSmall.name)
	equipmentTitle:SetPoint("TOPLEFT", equipment, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
	equipmentTitle:SetJustifyH("LEFT")
	local equipmentText1 = equipment:CreateFontString("CharScanner_MainFrame_Equipment_Text1", "ARTWORK", fonts.GameFontNormalSmall.name)
	equipmentText1:SetPoint("LEFT", equipmentTitle, "RIGHT", sizeGeneral.Offset, 0)
	equipmentText1:SetJustifyH("LEFT")
	equipmentText1:SetTextColor(1, 1, 1, 1)
	local equipmentText2 = equipment:CreateFontString("CharScanner_MainFrame_Equipment_Text2", "ARTWORK", fonts.GameFontNormalSmall.name)
	equipmentText2:SetPoint("TOPLEFT", equipmentText1, "BOTTOMLEFT", 0, -(sizeGeneral.Space+sizeGeneral.Space))
	equipmentText2:SetJustifyH("LEFT")
	equipmentText2:SetTextColor(1, 1, 1, 1)

	local equipmentButtonCollapse = CreateFrame("Button", "CharScanner_MainFrame_Equipment_ButtonCollapse", equipment)
	TemplateSmallCollapsButton(equipmentButtonCollapse, 6)
	equipmentButtonCollapse:SetPoint("RIGHT", equipment, "RIGHT", 0, 0)
	equipmentButtonCollapse:SetScript("OnClick", function() CharScanner:GearFrame_OnClick() end)
	equipmentButtonCollapse:SetScript("OnEnter", function() CharScanner:GearFrame_OnEnter() end)
	equipmentButtonCollapse:SetScript("OnLeave", function() CharScanner:GearFrame_OnLeave() end)
	-- Equipment Frame END

	-- Gear Frame START
	local gear = CreateFrame("Frame", "CharScanner_MainFrame_Gear", equipment)
	TemplateBorderRBL(gear)
	gear:SetFrameLevel( CharScanner_MainFrame_Equipment:GetFrameLevel() )
	gear:SetWidth(sizeGeneral.MainWidth)
	gear:SetHeight(sizeEquipmentFrame.GearHeight)
	gear:SetPoint("TOPLEFT", equipment, "BOTTOMLEFT", 0, sizeGeneral.Border)
	gear:EnableMouse(true)
	gear:Hide()

	for i = 1, #slotButton do
		local button = CreateFrame("Button", "CharScanner_MainFrame_Gear_Slot"..i, gear)
		if i == 1 then
			TemplateGearButton(button, 1)
			button:SetPoint("TOPLEFT", gear, "TOPLEFT", sizeEquipmentFrame.FirstSlotLeft, -sizeEquipmentFrame.Offset)
		else
			if i >= 2 and i <= 8 then
				TemplateGearButton(button, 1)
				button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "BOTTOMLEFT", 0, 0)
			else
				if i == 9 then
					TemplateGearButton(button, 3)
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "TOPRIGHT", Textures.GearIconTexture.width, -(Textures.GearIconTexture.height/2))
				else
					if i >= 10 and i <= 12 then
						TemplateGearButton(button, 3)
						button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "TOPRIGHT", 0, 0)
						if i == 12 then
							button:SetScale(0.75)
						end
					else
						if i == 13 then
							TemplateGearButton(button, 2)
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-2), "TOPRIGHT", Textures.GearIconTexture.width, (Textures.GearIconTexture.height/2))
						else
							TemplateGearButton(button, 2)
							button:SetPoint("BOTTOMLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "TOPLEFT", 0, 0)
						end
					end
				end
			end
		end
		_G["CharScanner_MainFrame_Gear_Slot"..i.."IconTexture"]:SetTexture(slotButton[i].texture)
		_G["CharScanner_MainFrame_Gear_Slot"..i.."Count"]:Hide()
		_G["CharScanner_MainFrame_Gear_Slot"..i.."EnchTexture"]:Hide()

		if not slotButton[i].IQ then -- disabled slots
			_G["CharScanner_MainFrame_Gear_Slot"..i.."Highlight"]:Hide()
			_G["CharScanner_MainFrame_Gear_Slot"..i.."IconTexture"]:SetAlpha(0.4)
		end
	end

	--local statsButton = CreateFrame("Button", "CharScanner_MainFrame_Equipment_StatsButton", gear)
	--TemplateHeadlineButton(statsButton)
	--statsButton:SetText(L["Base Stats"])
	--statsButton:SetWidth( ceil(CharScanner_MainFrame_Equipment_StatsButton:GetTextWidth()) + 20 )
	--statsButton:SetHeight(22)
	--statsButton:SetPoint("TOP", gear, "TOP", 0, -20)
	--statsButton:SetScript("OnEnter", function() CharScanner_MainFrame_Equipment_StatsFrame:Show() end)
	--statsButton:SetScript("OnLeave", function() CharScanner_MainFrame_Equipment_StatsFrame:Hide() end)
	--local statsframe = CreateFrame("Frame", "CharScanner_MainFrame_Equipment_StatsFrame", gear)
	--TemplateBorderTRBL(statsframe)
	--statsframe:SetToplevel(true)
	--statsframe:SetPoint("TOP", statsButton, "BOTTOM", 0, 0)
	--statsframe:EnableMouse(true)
	--statsframe:Hide()
	--local statsText = statsframe:CreateFontString("CharScanner_MainFrame_Equipment_StatsText", "ARTWORK", fonts.GameFontWhiteSmall.name)
	--statsText:SetPoint("TOP", statsframe, "TOP", 0, -10)
	--statsText:SetJustifyH("LEFT")
	local nogemText = gear:CreateFontString("CharScanner_MainFrame_Equipment_nogemText", "ARTWORK", fonts.GameFontNormalSmall.name)
	nogemText:SetPoint("TOP", gear, "TOP", 0, -18)
	nogemText:SetText("GEM SCAN: TODO!")
	nogemText:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
	-- Gear Frame END
	-- Equipment Frame END

	-- Achievement & Statistics Frame START
	local stats = CreateFrame("Frame", "CharScanner_MainFrame_Stats", mainframe)
	TemplateBorderTRBL(stats)
	stats:SetWidth(sizeGeneral.MainWidth)
	stats:SetHeight(sizeStatFrame.BaseHeight)
	stats:SetPoint("TOPLEFT", equipment, "BOTTOMLEFT", 0, sizeGeneral.Border)
	stats:EnableMouse(true)
	stats:Hide()

	-- title START
	local button = CreateFrame("Button", "CharScanner_MainFrame_Stats_Title", stats)
	TemplateStats(button, false, "title")
	button:SetPoint("TOPLEFT", stats, "TOPLEFT", sizeStatFrame.Offset, -sizeStatFrame.Offset)
	button.AchievementClicked = ""
	-- title END

	local TEMPstats = {}
	-- level 1 START
	for i = 1, #DATA do
		local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel1_"..DATA[i].name, stats)
		TemplateStats(button, true, "level1", i)
		if i == 1 then
			button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Stats_Title", "BOTTOMLEFT", 0, 0)
		else
			button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel1_"..DATA[i-1].name, "BOTTOMLEFT", 0, 0)
		end
		button:SetScript("OnClick", function() CharScanner:NodeClick(1, DATA[i].name, i) end)
		--_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Name"]:SetText("StatsLevel1_"..DATA[i].name) -- TEST

		local n = "CharScanner_MainFrame_StatsLevel1_"..DATA[i].name
		TEMPstats[n] = {}
		TEMPstats[n].level = 1
		TEMPstats[n].index1 = i
		TEMPstats[n].index2 = 0
		TEMPstats[n].index3 = 0
		TEMPstats[n].height = sizeGeneral.BarHeight
		TEMPstats[n].collapse = false
	end
	-- level 1 END

	-- level 2 START
	for i = 1, #DATA do
		if DATA[i].statistic then
			local height = 0
			local statsLvL2 = CreateFrame("Frame", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name, _G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name])
			statsLvL2:SetWidth(sizeGeneral.MainWidth)
			statsLvL2:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel1_"..DATA[i].name, "BOTTOMLEFT", 0, 0)
			for j = 1, #DATA[i].statistic do
				local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j, statsLvL2)
				TemplateStats(button, false, "level2")
				if j == 1 then
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name, "TOPLEFT", 0, 0)
				else
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..(j-1), "BOTTOMLEFT", 0, 0)
				end
				--_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j.."Name"]:SetText("StatsLevel2_"..DATA[i].name..j) -- TEST
				height = height + 1
			end
			statsLvL2:SetHeight(height*sizeGeneral.BarHeight)
			statsLvL2:Hide()

			local n = "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name
			TEMPstats[n] = {}
			TEMPstats[n].level = 2
			TEMPstats[n].index1 = i
			TEMPstats[n].index2 = 0
			TEMPstats[n].index3 = 0
			TEMPstats[n].height = height*sizeGeneral.BarHeight
			TEMPstats[n].collapse = false
		else
			local height = 0
			local statsLvL2 = CreateFrame("Frame", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name, _G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name])
			statsLvL2:SetWidth(sizeGeneral.MainWidth)
			statsLvL2:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel1_"..DATA[i].name, "BOTTOMLEFT", 0, 0)
			for j = 1, #DATA[i] do
				local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name, statsLvL2)
				TemplateStats(button, true, "level2", i, j)
				if j == 1 then
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name, "TOPLEFT", 0, 0)
				else
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j-1].name, "BOTTOMLEFT", 0, 0)
				end
				button:SetScript("OnClick", function() CharScanner:NodeClick(2, DATA[i].name, i, DATA[i][j].name, j) end)
				--_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Name"]:SetText("StatsLevel2_"..DATA[i].name..DATA[i][j].name) -- TEST
				height = height + 1
			end
			statsLvL2:SetHeight(height*sizeGeneral.BarHeight)
			statsLvL2:Hide()

			local n = "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name
			TEMPstats[n] = {}
			TEMPstats[n].level = 2
			TEMPstats[n].index1 = i
			TEMPstats[n].index2 = 0
			TEMPstats[n].index3 = 0
			TEMPstats[n].height = height*sizeGeneral.BarHeight
			TEMPstats[n].collapse = false
		end
	end
	-- level 2 END

	-- level 3 START
	for i = 1, #DATA do
		if not DATA[i].statistic then
			for j = 1, #DATA[i] do

				if DATA[i][j].statistic then
					local height = 0
					local statsLvL3 = CreateFrame("Frame", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name, _G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name])
					statsLvL3:SetWidth(sizeGeneral.MainWidth)
					statsLvL3:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name, "BOTTOMLEFT", 0, 0)
					for k = 1, #DATA[i][j].statistic do
						local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k, statsLvL3)
						TemplateStats(button, false, "level3")
						if k == 1 then
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name, "TOPLEFT", 0, 0)
						else
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..(k-1), "BOTTOMLEFT", 0, 0)
						end
						--_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText("StatsLevel3_"..DATA[i].name..DATA[i][j].name..k) -- TEST
						height = height + 1
					end
					statsLvL3:SetHeight(height*sizeGeneral.BarHeight)
					statsLvL3:Hide()

					local n = "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name
					TEMPstats[n] = {}
					TEMPstats[n].level = 3
					TEMPstats[n].index1 = i
					TEMPstats[n].index2 = j
					TEMPstats[n].index3 = 0
					TEMPstats[n].height = height*sizeGeneral.BarHeight
					TEMPstats[n].collapse = false
				else
					local height = 0
					local statsLvL3 = CreateFrame("Frame", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name, _G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name])
					statsLvL3:SetWidth(sizeGeneral.MainWidth)
					statsLvL3:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name, "BOTTOMLEFT", 0, 0)
					for k = 1, #DATA[i][j] do
						local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k, statsLvL3)
						TemplateStats(button, true, "level3", i, j, k)
						if k == 1 then
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name, "TOPLEFT", 0, 0)
						else
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..(k-1), "BOTTOMLEFT", 0, 0)
						end
						button:SetScript("OnClick", function() CharScanner:NodeClick(3, DATA[i].name, i, DATA[i][j].name, j, DATA[i][j][k].name, k) end)
						--_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText("StatsLevel3_"..DATA[i].name..DATA[i][j].name..k) -- TEST
						height = height + 1
					end
					statsLvL3:SetHeight(height*sizeGeneral.BarHeight)
					statsLvL3:Hide()

					local n = "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name
					TEMPstats[n] = {}
					TEMPstats[n].level = 3
					TEMPstats[n].index1 = i
					TEMPstats[n].index2 = j
					TEMPstats[n].index3 = 0
					TEMPstats[n].height = height*sizeGeneral.BarHeight
					TEMPstats[n].collapse = false
				end

			end
		end
	end
	-- level 3 END

	-- level 4 START
	for i = 1, #DATA do
		if not DATA[i].statistic then
			for j = 1, #DATA[i] do

				if not DATA[i][j].statistic then
					for k = 1, #DATA[i][j] do

						if DATA[i][j][k].statistic then
							local height = 0
							local statsLvL4 = CreateFrame("Frame", "CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name, _G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name])
							statsLvL4:SetWidth(sizeGeneral.MainWidth)
							statsLvL4:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k, "BOTTOMLEFT", 0, 0)
							for l = 1, #DATA[i][j][k].statistic do
								local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l, statsLvL4)
								TemplateStats(button, false, "level4")
								if l == 1 then
									button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name, "TOPLEFT", 0, 0)
								else
									button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..(l-1), "BOTTOMLEFT", 0, 0)
								end
								--_G["CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l.."Name"]:SetText("StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l) -- TEST
								height = height + 1
							end
							statsLvL4:SetHeight(height*sizeGeneral.BarHeight)
							statsLvL4:Hide()

							local n = "CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name
							TEMPstats[n] = {}
							TEMPstats[n].level = 4
							TEMPstats[n].index1 = i
							TEMPstats[n].index2 = j
							TEMPstats[n].index3 = k
							TEMPstats[n].height = height*sizeGeneral.BarHeight
							TEMPstats[n].collapse = false
						end

					end
				end

			end
		end
	end
	-- level 4 END

	-- stats sort START
	local function SortedPairs(t)
		local ntab = {}
		for k,v in pairs(t) do
			table.insert(ntab,k)
		end
		table.sort(ntab,
			function(a,b)
				if t[a].level == t[b].level then
					if t[a].index1 == t[b].index1 then
						if t[a].index2 == t[b].index2 then
							if t[a].index3 == t[b].index3 then
								if a < b then
									return true
						  	end
							elseif t[a].index3 < t[b].index3 then
								return true
							end
						elseif t[a].index2 < t[b].index2 then
							return true
						end
					elseif t[a].index1 < t[b].index1 then
						return true
					end
				elseif t[a].level < t[b].level then
					return true
				end
			end
		)
		local i = 0
		local function _f(_s,_v)
			i = i + 1
			local k = ntab[i]
			if (k) then
				return k,t[k]
			end
		end
		return _f,nil,nil
	end

	local ni = 1
	for key, value in SortedPairs(TEMPstats) do
		GUI_menu_values[ni] = {}
		GUI_menu_values[ni].name     = key
		GUI_menu_values[ni].level    = value.level
		GUI_menu_values[ni].index1   = value.index1
		GUI_menu_values[ni].index2   = value.index2
		GUI_menu_values[ni].index3   = value.index3
		GUI_menu_values[ni].height   = value.height
		GUI_menu_values[ni].collapse = value.collapse
		ni = ni + 1
	end
	table.wipe(TEMPstats)
	-- stats sort END

	-- AchievementDetailFrame START
	local achievementDetail = CreateFrame("Frame", "CharScanner_AchievementDetailFrame", stats)
	TemplateBorderTRBL(achievementDetail)
	achievementDetail:SetWidth(sizeAchievementDetail.Width)
	achievementDetail:SetHeight(sizeAchievementDetail.Height)
	achievementDetail:SetPoint("TOPLEFT", base, "TOPRIGHT", -sizeGeneral.Border, 0)
	achievementDetail:EnableMouse(true)
	achievementDetail:SetFrameLevel( CharScanner_MainFrame_Stats:GetFrameLevel()+3 )
	achievementDetail:Hide()

	-- AchievementDetail Sort START
	local achievementDetailsort = CreateFrame("Frame", "CharScanner_AchievementDetailFrame_Sort", achievementDetail)
	achievementDetailsort:SetWidth(sizeAchievementDetail.ButtonWidth)
	achievementDetailsort:SetHeight(sizeGeneral.BarHeight)
	achievementDetailsort:SetPoint("TOPLEFT", achievementDetail, "TOPLEFT", sizeAchievementDetail.Offset, -sizeAchievementDetail.SortLine2Top)
	TemplateAchievementDetailList(achievementDetailsort, true)
	-- AchievementDetail Sort END

	-- AchievementDetail ScrollFrame START
	local achievementDetailScrollFrame = CreateFrame("ScrollFrame", "CharScanner_AchievementDetailFrameScrollFrame", achievementDetail, "FauxScrollFrameTemplate")
	achievementDetailScrollFrame:SetPoint("TOPLEFT", achievementDetail, "TOPLEFT", 0, -sizeAchievementDetail.SortLineXTop)
	achievementDetailScrollFrame:SetPoint("BOTTOMRIGHT", achievementDetail, "BOTTOMRIGHT", -27, 4)
	achievementDetailScrollFrame:SetScript("OnVerticalScroll", function() CharScanner:OnVerticalScroll_AchievementDetail() end)
	-- scroll knob tweak START
	CharScanner_AchievementDetailFrameScrollFrameScrollBarThumbTexture:SetWidth(sizeGeneral.SliderWidth)
	CharScanner_AchievementDetailFrameScrollFrameScrollBarThumbTexture:SetHeight(sizeGeneral.SliderWidth)
	CharScanner_AchievementDetailFrameScrollFrameScrollBarThumbTexture:SetTexCoord(0.21875, 0.75, 0.25, 0.71875) -- (7/32, 24/32, 8/32, 23/32)
	-- scroll knob tweak END
	for i = 1, GUI_achievementdetail_maxNum do
		local button = CreateFrame("Button", "CharScanner_AchievementDetailFrame_Button"..i, achievementDetail)
		TemplateAchievementDetailList(button)
		button:SetWidth(sizeAchievementDetail.ButtonWidth)
		button:SetHeight(sizeGeneral.BarHeight)
		button:SetID(i)
		if i == 1 then
			button:SetPoint("TOPLEFT", "CharScanner_AchievementDetailFrame", "TOPLEFT", sizeAchievementDetail.Offset, -sizeAchievementDetail.SortLineXTop)
		else
			button:SetPoint("TOPLEFT", "CharScanner_AchievementDetailFrame_Button"..(i-1), "BOTTOMLEFT", 0, 0)
		end
	end
	-- AchievementDetail ScrollFrame END

	-- AchievementDetail StatusFrame START
	local achievementDetailStatus = CreateFrame("Frame", "CharScanner_AchievementDetailFrame_Status", achievementDetail)
	TemplateBorderTRBL(achievementDetailStatus)
	achievementDetailStatus:SetWidth(sizeAchievementDetail.Width)
	achievementDetailStatus:SetHeight(sizeGeneral.StatusHeight)
	achievementDetailStatus:SetPoint("TOPLEFT", achievementDetail, "BOTTOMLEFT", 0, sizeGeneral.Border)
	achievementDetailStatus:EnableMouse(true)

	local achievementDetailStatusText = achievementDetailStatus:CreateFontString("CharScanner_AchievementDetailFrame_Status_Text", "ARTWORK", fonts.GameFontNormalSmall.name)
	achievementDetailStatusText:SetPoint("CENTER", achievementDetailStatus, "CENTER", 0, 0)
	achievementDetailStatusText:SetTextColor(1, 1, 1, 1)
	-- AchievementDetail StatusFrame END
	-- AchievementDetailFrame END
	-- Achievement & Statistics Frame END

	-- Summary Frame START
	local summary = CreateFrame("Frame", "CharScanner_MainFrame_Summary", mainframe)
	TemplateBorderTRBL(summary)
	summary:SetWidth(sizeGeneral.MainWidth)
	summary:SetHeight(sizeSummaryFrame_BaseHeight)
	summary:SetPoint("TOPLEFT", stats, "BOTTOMLEFT", 0, sizeGeneral.Border)
	summary:Hide()

	local button1 = CreateFrame("Button", "CharScanner_MainFrame_SummaryButton1", summary)
	TemplateStats(button1, false, "summary")
	button1:SetPoint("TOPLEFT", summary, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
	local button2 = CreateFrame("Button", "CharScanner_MainFrame_SummaryButton2", summary)
	TemplateStats(button2, false, "summary")
	button2:SetPoint("TOPLEFT", button1, "BOTTOMLEFT", 0, 0)
	-- Summary Frame END

	-- Status Frame START
	local status = CreateFrame("Frame", "CharScanner_MainFrame_Status", mainframe)
	TemplateBorderTRBL(status)
	status:SetWidth(sizeGeneral.MainWidth)
	status:SetHeight(sizeGeneral.StatusHeight)
	status:SetPoint("TOPLEFT", summary, "BOTTOMLEFT", 0, sizeGeneral.Border)
	status:EnableMouse(true)
	status:SetScript("OnShow", function() CharScanner:CheckStatusFramePosition() end)
	status:Hide()

	local statusText = status:CreateFontString("CharScanner_MainFrame_Status_Text", "ARTWORK", fonts.GameFontNormalSmall.name)
	statusText:SetPoint("CENTER", status, "CENTER", 0, 0)
	statusText:SetTextColor(1, 1, 1, 1)
	-- Status Frame END



	-- History Frame START
	local history = CreateFrame("Frame", "CharScanner_HistoryFrame", mainframe)
	TemplateBorderTRBL(history)
	history:SetWidth(sizeCompareFrame.HistoryWidth)
	history:SetHeight(sizeCompareFrame.HistoryHeight)
	history:SetPoint("TOPLEFT", base, "TOPLEFT", -(sizeCompareFrame.HistoryWidth-sizeGeneral.Border), 0)
	history:EnableMouse(true)
	history:SetToplevel(true)
	history:Hide()

	-- History Sort START
	local historysort = CreateFrame("Frame", "CharScanner_HistoryFrame_Sort", history)
	historysort:SetWidth(sizeCompareFrame.HistoryWidth)
	historysort:SetHeight(sizeGeneral.BarHeight)
	historysort:SetPoint("TOPLEFT", history, "TOPLEFT", 0, -sizeCompareFrame.SortLine2Top)
	TemplateHistoryList(historysort, nil, true)
	-- History Sort END

	-- History ScrollFrame START
	local historyScrollFrame = CreateFrame("ScrollFrame", "CharScanner_HistoryFrameScrollFrame", history, "FauxScrollFrameTemplate")
	historyScrollFrame:SetPoint("TOPLEFT", history, "TOPLEFT", 0, -sizeCompareFrame.SortLineXTop)
	historyScrollFrame:SetPoint("BOTTOMRIGHT", history, "BOTTOMRIGHT", -27, 4)
	historyScrollFrame:SetScript("OnVerticalScroll", function() CharScanner:OnVerticalScroll() end)
	-- scroll knob tweak START
	CharScanner_HistoryFrameScrollFrameScrollBarThumbTexture:SetWidth(sizeGeneral.SliderWidth)
	CharScanner_HistoryFrameScrollFrameScrollBarThumbTexture:SetHeight(sizeGeneral.SliderWidth)
	CharScanner_HistoryFrameScrollFrameScrollBarThumbTexture:SetTexCoord(0.21875, 0.75, 0.25, 0.71875) -- (7/32, 24/32, 8/32, 23/32)
	-- scroll knob tweak END
	for i = 1, GUI_history_maxNum do
		local button = CreateFrame("Button", "CharScanner_HistoryFrame_Button"..i, history)
		TemplateHistoryList(button, i)
		button:SetWidth(sizeCompareFrame.HistoryWidth)
		button:SetHeight(sizeGeneral.BarHeight)
		button:SetID(i)
		if i == 1 then
			button:SetPoint("TOPLEFT", "CharScanner_HistoryFrame", "TOPLEFT", 0, -sizeCompareFrame.SortLineXTop)
		else
			button:SetPoint("TOPLEFT", "CharScanner_HistoryFrame_Button"..(i-1), "BOTTOMLEFT", 0, 0)
		end
	end
	-- History ScrollFrame END

	-- History StatusFrame START
	local historyStatus = CreateFrame("Frame", "CharScanner_HistoryFrame_Status", history)
	TemplateBorderTRBL(historyStatus)
	historyStatus:SetWidth(sizeCompareFrame.HistoryWidth)
	historyStatus:SetHeight(sizeGeneral.StatusHeight)
	historyStatus:SetPoint("TOPLEFT", history, "BOTTOMLEFT", 0, sizeGeneral.Border)
	historyStatus:EnableMouse(true)

	local faketexture = historyStatus:CreateTexture(nil, "ARTWORK")
	faketexture:SetWidth(100)
	faketexture:SetHeight(sizeGeneral.StatusHeight)
	faketexture:SetPoint("TOPLEFT", historyStatus, "TOPLEFT", (sizeCompareFrame.HistoryWidth-100)/2, 0)
	faketexture:SetTexture(0, 0, 0, 0)
	local historyStatusText = historyStatus:CreateFontString("CharScanner_HistoryFrame_Status_Text", "ARTWORK", fonts.GameFontNormalSmall.name)
	historyStatusText:SetPoint("CENTER", faketexture, "CENTER", 0, 0)
	historyStatusText:SetTextColor(1, 1, 1, 1)

	local texture = CharScanner_HistoryFrame_Sort_CompareFrame:CreateTexture(nil, "ARTWORK")
	texture:SetPoint("TOPLEFT", historyStatus, "TOPLEFT", sizeCompareFrame.Talents1Left, -sizeGeneral.Space)
	texture:SetPoint("BOTTOMRIGHT", historyStatus, "TOPLEFT", sizeCompareFrame.Talents2Left+sizeCompareFrame.Talents2Width, -(sizeGeneral.StatusHeight-sizeGeneral.Space))
	texture:SetTexture(1, 1, 1, 0.075)
	local historyStatusRoleText = CharScanner_HistoryFrame_Sort_CompareFrame:CreateFontString("CharScanner_HistoryFrame_Status_RoleText", "ARTWORK", fonts.GameFontNormal.name)
	historyStatusRoleText:SetPoint("CENTER", texture, "CENTER", 0, 0)
	-- History StatusFrame END
	-- History Frame END



	-- SCAN-Button Frame START
	local scanbutton = CreateFrame("Frame", "CharScanner_ScanButtonFrame", UIParent)
	scanbutton:EnableMouse(true)
	scanbutton:SetMovable(true)
	scanbutton:SetToplevel(true)
	scanbutton:SetClampedToScreen(true)
	scanbutton:Hide()
	scanbutton:SetWidth(CharScanner_Options.scanButtonSize)
	scanbutton:SetHeight(CharScanner_Options.scanButtonSize)
	scanbutton:SetScript("OnShow", function() CharScanner:ScanButtonPos() end)
	local scanbuttonButton = CreateFrame("Button", "CharScanner_ScanButtonFrame_Button", scanbutton)
	scanbuttonButton:SetPoint("TOPLEFT", 0, 0)
	scanbuttonButton:SetPoint("BOTTOMRIGHT", 0, 0)
	scanbuttonButton:SetScript("OnClick", function() CharScanner_MainFrame:Show() CharScanner:ScanUnit("target") end)
	local texture = scanbuttonButton:CreateTexture("CharScanner_ScanButtonFrame_ButtonNormalTexture", "ARTWORK")
	texture:SetPoint("TOPLEFT", 0, 0)
	texture:SetPoint("BOTTOMRIGHT", 0, 0)
	texture:SetTexture(Textures.CharScanner.path)
	scanbuttonButton:SetNormalTexture(texture)
	local texturePush = scanbuttonButton:CreateTexture("CharScanner_ScanButtonFrame_ButtonPushTexture", "ARTWORK")
	texturePush:SetPoint("TOPLEFT", 1, -1)
	texturePush:SetPoint("BOTTOMRIGHT", -1, 1)
	texturePush:SetTexture(Textures.CharScanner.path)
	scanbuttonButton:SetPushedTexture(texturePush)
	local textureHighlight = scanbuttonButton:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetPoint("TOPLEFT", 0, 0)
	textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
	textureHighlight:SetTexture(Textures.CharScannerHL.path)
	scanbuttonButton:SetHighlightTexture(textureHighlight)

	local mover = CreateFrame("Frame", "CharScanner_ScanButtonFrame_Mover", scanbutton)
	mover:EnableMouse(true)
	mover:SetToplevel(true)
	mover:SetWidth(300)
	mover:SetHeight(80)
	mover:SetPoint("TOP", scanbutton, "TOP", 0, 32)
	local moverBGtop = mover:CreateTexture(nil, "ARTWORK")
	moverBGtop:SetPoint("TOPLEFT", mover, "TOPLEFT", 0, 0)
	moverBGtop:SetPoint("BOTTOMRIGHT", mover, "TOPRIGHT", 0, -28)
	moverBGtop:SetTexture(0, 0, 0, 0.7)
	local moverBGbottom = mover:CreateTexture(nil, "ARTWORK")
	moverBGbottom:SetPoint("TOPLEFT", mover, "TOPLEFT", 0, -84)
	moverBGbottom:SetPoint("BOTTOMRIGHT", 0, 0)
	moverBGbottom:SetTexture(0, 0, 0, 0.7)
	mover:SetScript("OnEnter", function() moverBGtop:SetTexture(0, 0, 0, 0.4) moverBGbottom:SetTexture(0, 0, 0, 0.4) end)
	mover:SetScript("OnLeave", function() moverBGtop:SetTexture(0, 0, 0, 0.7) moverBGbottom:SetTexture(0, 0, 0, 0.7) end)
	mover:SetScript("OnMouseDown", function() CharScanner_ScanButtonFrame:StartMoving() end)
	mover:SetScript("OnMouseUp", function() CharScanner_ScanButtonFrame:StopMovingOrSizing() CharScanner:SaveScanButtonPos() end)

	local scanbuttonScale = CreateFrame("Slider", "CharScanner_ScanButtonFrame_ButtonScale", mover, "OptionsSliderTemplate")
	CharScanner_ScanButtonFrame_ButtonScale:SetWidth(200)
	CharScanner_ScanButtonFrame_ButtonScale:SetHeight(16)
	CharScanner_ScanButtonFrame_ButtonScale:SetPoint("TOP", mover, "TOP", 0, -88)
	CharScanner_ScanButtonFrame_ButtonScaleLow:SetText("10px")
	CharScanner_ScanButtonFrame_ButtonScaleHigh:SetText("48px")
	CharScanner_ScanButtonFrame_ButtonScale:SetMinMaxValues(10, 48)
	CharScanner_ScanButtonFrame_ButtonScale:SetValueStep(2)
	CharScanner_ScanButtonFrame_ButtonScale:SetValue(CharScanner_Options.scanButtonSize)
	local valuetext = scanbuttonScale:CreateFontString("CharScanner_ScanButtonFrame_ButtonScale_valuetext", "OVERLAY", fonts.GameFontNormalSmall.name)
	valuetext:SetPoint("TOP", scanbuttonScale, "BOTTOM", 0, 0)
	valuetext:SetJustifyH("CENTER")
	valuetext:SetText(CharScanner_Options.scanButtonSize.."x"..CharScanner_Options.scanButtonSize.."px")
	CharScanner_ScanButtonFrame_ButtonScale:SetScript("OnValueChanged", function(self, value)
		CharScanner_ScanButtonFrame_ButtonScale_valuetext:SetText(value.."x"..value.."px")
		CharScanner_Options.scanButtonSize = value
		CharScanner_ScanButtonFrame:SetWidth(value)
		CharScanner_ScanButtonFrame:SetHeight(value)
	end)

	local text = mover:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	text:SetPoint("TOP", valuetext, "BOTTOM", 0, -4)
	text:SetJustifyH("CENTER")
	text:SetWidth(300)
	text:SetText(L["mover_desc"])
	text:SetNonSpaceWrap(true)
	text:SetTextColor(1, 1, 1, 0.5)
	local scanbuttonLock = CreateFrame("Button", "CharScanner_ScanButtonFrame_ButtonLock", mover)
	TemplateHeadlineButton(scanbuttonLock)
	scanbuttonLock:SetText(L["LOCK"])
	scanbuttonLock:SetWidth( ceil(CharScanner_ScanButtonFrame_ButtonLock:GetTextWidth()) + 20 )
	scanbuttonLock:SetHeight(22)
	scanbuttonLock:SetPoint("TOP", text, "BOTTOM", 0, -4)
	scanbuttonLock:SetScript("OnClick", function()
		CharScanner_Options.scanButtonLock = true
		CharScanner_ScanButtonFrame_Mover:Hide()
		CharScanner_InterfaceOptions_ScanButtonLock:SetChecked(CharScanner_Options.scanButtonLock)
	end)
	
	-- +inc
	mover:SetHeight( 28 + 56 + 4 +   16 + ceil(valuetext:GetStringHeight())   + 4 + ceil(text:GetStringHeight()) + 4 + ceil(scanbuttonLock:GetHeight()) + 4 )

	if CharScanner_Options.scanButtonLock then
		mover:Hide()
	end
	-- SCAN-Button Frame END



	-- Protection Frame START
	local ProtectionFrame = CreateFrame("Frame", "CharScanner_ProtectionFrame", mainframe)
	ProtectionFrame:SetToplevel(true)
	ProtectionFrame:SetWidth(sizeGeneral.MainWidth)
	ProtectionFrame:SetHeight(50)
	ProtectionFrame:SetPoint("TOPLEFT", mainframe, "TOPLEFT", 0, 0)
	ProtectionFrame:SetPoint("BOTTOMLEFT", summary, "BOTTOMLEFT", 0, -sizeGeneral.StatusHeight) -- for status
	ProtectionFrame:EnableMouse(true)
	ProtectionFrame:Hide()
	local texture = ProtectionFrame:CreateTexture(nil, "BACKGROUND")
	texture:SetPoint("TOPLEFT", 0, 0)
	texture:SetPoint("BOTTOMRIGHT", 0, 0)
	texture:SetTexture(0, 0, 0, 0.85)
	local ProtectionButton = CreateFrame("Button", "CharScanner_ProtectionFrame_Button", ProtectionFrame)
	TemplateHeadlineButton(ProtectionButton)
	ProtectionButton:SetText(L["STOP SCAN"])
	ProtectionButton:SetWidth( ceil(CharScanner_ProtectionFrame_Button:GetTextWidth()) + 20 )
	ProtectionButton:SetHeight(22)
	ProtectionButton:SetPoint("TOP", ProtectionFrame, "TOP", 0, -(sizeGeneral.MainHeight+sizeBaseCharFrame_BaseHeight))
	ProtectionButton:SetScript("OnClick", function() CharScanner:StopGroupScan() end)
	local progressText = ProtectionFrame:CreateFontString("CharScanner_ProtectionFrame_ProgressText", "ARTWORK", fonts.GameFontNormalSmall.name)
	progressText:SetPoint("TOP", ProtectionButton, "BOTTOM", 0, -20)
	progressText:SetJustifyH("CENTER")
	progressText:SetWidth(100)
	progressText:SetText("-/-")
	progressText:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
	local progressBackground = ProtectionFrame:CreateTexture(nil, "BORDER")
	progressBackground:SetPoint("TOP", progressText, "BOTTOM", 0, -sizeGeneral.Offset)
	progressBackground:SetWidth(sizeProtectionFrame_BarWidth)
	progressBackground:SetHeight(12)
	progressBackground:SetTexture(color.DarkGrey.rgba[1], color.DarkGrey.rgba[2], color.DarkGrey.rgba[3], color.DarkGrey.rgba[4])
	local progressDone = ProtectionFrame:CreateTexture("CharScanner_ProtectionFrame_ProgressDone", "ARTWORK")
	progressDone:SetPoint("LEFT", progressBackground, "LEFT", 0, 0)
	progressDone:SetWidth(1)
	progressDone:SetHeight(12)
	progressDone:SetBlendMode("BLEND")
	progressDone:SetTexture(color.Green.rgba[1], color.Green.rgba[2], color.Green.rgba[3], color.Green.rgba[4])

	local ProtectionHistoryFrame = CreateFrame("Frame", "CharScanner_ProtectionHistoryFrame", history)
	ProtectionHistoryFrame:SetToplevel(true)
	ProtectionHistoryFrame:SetWidth(sizeCompareFrame.HistoryWidth)
	ProtectionHistoryFrame:SetHeight(50)
	ProtectionHistoryFrame:SetPoint("TOPLEFT", history, "TOPLEFT", 0, 0)
	ProtectionHistoryFrame:SetPoint("BOTTOMLEFT", historyStatus, "BOTTOMLEFT", 0, 0)
	ProtectionHistoryFrame:EnableMouse(true)
	ProtectionHistoryFrame:Hide()
	local texture = ProtectionHistoryFrame:CreateTexture(nil, "ARTWORK")
	texture:SetPoint("TOPLEFT", 0, 0)
	texture:SetPoint("BOTTOMRIGHT", 0, 0)
	texture:SetTexture(0, 0, 0, 0.85)

	local ProtectionAchievementDetailFrame = CreateFrame("Frame", "CharScanner_ProtectionAchievementDetailFrame", achievementDetail)
	ProtectionAchievementDetailFrame:SetToplevel(true)
	ProtectionAchievementDetailFrame:SetWidth(sizeAchievementDetail.Width)
	ProtectionAchievementDetailFrame:SetHeight(50)
	ProtectionAchievementDetailFrame:SetPoint("TOPLEFT", achievementDetail, "TOPLEFT", 0, 0)
	ProtectionAchievementDetailFrame:SetPoint("BOTTOMLEFT", achievementDetailStatus, "BOTTOMLEFT", 0, 0)
	ProtectionAchievementDetailFrame:EnableMouse(true)
	ProtectionAchievementDetailFrame:Hide()
	local texture = ProtectionAchievementDetailFrame:CreateTexture(nil, "ARTWORK")
	texture:SetPoint("TOPLEFT", 0, 0)
	texture:SetPoint("BOTTOMRIGHT", 0, 0)
	texture:SetTexture(0, 0, 0, 0.85)
	-- Protection Frame END
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:CreateMinimapButton()
	if not CharScanner_Options.MinimapButton then
		if CharScanner_MinimapButton then
			CharScanner_MinimapButton:Hide()
		end
		return
	else
		if CharScanner_MinimapButton then
			CharScanner_MinimapButton:Show()
			return
		end
	end

	if CharScanner_MinimapButton then return end -- hmm I am currently infected with brain damage x)

	local function MoveMinimapButton()
		local xpos
		local ypos
		local minimapShape = GetMinimapShape and GetMinimapShape() or "ROUND"
		if minimapShape == "SQUARE" then
			xpos = 110 * cos(CharScanner_Options.MinimapButtonPos or 0)
			ypos = 110 * sin(CharScanner_Options.MinimapButtonPos or 0)
			xpos = math.max(-82, math.min(xpos, 84))
			ypos = math.max(-86, math.min(ypos, 82))
		else
			xpos = 80 * cos(CharScanner_Options.MinimapButtonPos or 0)
			ypos = 80 * sin(CharScanner_Options.MinimapButtonPos or 0)
		end
		CharScanner_MinimapButton:SetPoint("TOPLEFT", Minimap, "TOPLEFT", 52-xpos, ypos-52)
	end

	local function DragMinimapButton()
		local xpos, ypos = GetCursorPosition()
		local xmin, ymin = Minimap:GetLeft() or 400, Minimap:GetBottom() or 400
		local scale = Minimap:GetEffectiveScale()
		xpos = xmin-xpos/scale+70
		ypos = ypos/scale-ymin-70
		CharScanner_Options.MinimapButtonPos = math.deg(math.atan2(ypos, xpos))
		MoveMinimapButton()
	end

	local MinimapButton = CreateFrame("Button", "CharScanner_MinimapButton", Minimap)
	MinimapButton:EnableMouse(true)
	MinimapButton:SetMovable(true)
	MinimapButton:SetToplevel(true)
	MinimapButton:SetWidth(Textures.MinimapButton.width)
	MinimapButton:SetHeight(Textures.MinimapButton.height)
	MinimapButton:SetPoint("TOPLEFT")
	MinimapButton:SetFrameStrata("LOW")
	MinimapButton:RegisterForClicks("AnyUp")
	MinimapButton:RegisterForDrag("LeftButton")

	MinimapButton.NormalTexture = MinimapButton:CreateTexture(nil, "ARTWORK")
	MinimapButton.NormalTexture:SetPoint("TOPLEFT", 0, 0)
	MinimapButton.NormalTexture:SetPoint("BOTTOMRIGHT", 0, 0)
	MinimapButton.NormalTexture:SetTexture(Textures.MinimapButton.path)
	MinimapButton.NormalTexture:SetTexCoord(Textures.MinimapButton.coords[1], Textures.MinimapButton.coords[2], Textures.MinimapButton.coords[3], Textures.MinimapButton.coords[4])
	MinimapButton:SetNormalTexture(MinimapButton.NormalTexture)

	MinimapButton.PushedTexture = MinimapButton:CreateTexture(nil, "ARTWORK")
	MinimapButton.PushedTexture:SetPoint("TOPLEFT", 1, -1)
	MinimapButton.PushedTexture:SetPoint("BOTTOMRIGHT", -1, 1)
	MinimapButton.PushedTexture:SetTexture(Textures.MinimapButton.path)
	MinimapButton.PushedTexture:SetTexCoord(Textures.MinimapButton.coords[1], Textures.MinimapButton.coords[2], Textures.MinimapButton.coords[3], Textures.MinimapButton.coords[4])
	MinimapButton:SetPushedTexture(MinimapButton.PushedTexture)

	MinimapButton.HighlightTexture = MinimapButton:CreateTexture(nil, "ARTWORK")
	MinimapButton.HighlightTexture:SetPoint("TOPLEFT", 0, 0)
	MinimapButton.HighlightTexture:SetPoint("BOTTOMRIGHT", 0, 0)
	MinimapButton.HighlightTexture:SetTexture("Interface\\Minimap\\UI-Minimap-ZoomButton-Highlight")
	MinimapButton:SetHighlightTexture(MinimapButton.HighlightTexture)

	MinimapButton:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		GameTooltip:AddLine("CharScanner", color.Yellow.rgba[1], color.Yellow.rgba[2], color.Yellow.rgba[3], color.Yellow.rgba[4])
		GameTooltip:Show()
	end)
	MinimapButton:SetScript("OnLeave", function(self) GameTooltip:Hide() end)
	MinimapButton:SetScript("OnClick", function(self, button) CharScanner:MiniOptionsToggle(button, 2) end)
	MinimapButton:SetScript("OnDragStart", function(self) self:LockHighlight() self:SetScript("OnUpdate", DragMinimapButton) end)
	MinimapButton:SetScript("OnDragStop", function(self) self:SetScript("OnUpdate", nil) self:UnlockHighlight() end)

	MoveMinimapButton()
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:CreateMiniOptionsFrame()
	if CharScanner_MiniOptionsFrame then return end

	local buttondata = {}
	buttondata[1] = "SingleScan"
	buttondata[2] = "GroupScanAdd"
	buttondata[3] = "GroupScan"
	buttondata[4] = "Options"

	local maxiconwidth = 0
	local maxtextwidth = 0

	local iw = 0
	for i = 1, #buttondata do
		iw = Textures[ buttondata[i] ].width
		if iw > maxiconwidth then
			maxiconwidth = iw
		end
	end
	
	local buttonheight = Textures.SingleScan.height
	local frameheight = sizeGeneral.Offset + (#buttondata*buttonheight) + sizeGeneral.Offset

	local miniOptionsFrame = CreateFrame("Frame", "CharScanner_MiniOptionsFrame", UIParent)
	TemplateBorderTRBL(miniOptionsFrame)
	miniOptionsFrame:EnableMouse(true)
	miniOptionsFrame:SetMovable(true)
	miniOptionsFrame:SetToplevel(true)
	miniOptionsFrame:SetClampedToScreen(true)
	miniOptionsFrame:SetFrameStrata("DIALOG")
	miniOptionsFrame:SetHeight(frameheight)
	miniOptionsFrame:Hide()

	for i = 1, #buttondata do
		local buttonname = "CharScanner_MiniOptionsFrame_"..buttondata[i]

		local button = CreateFrame("Button", buttonname, miniOptionsFrame)
		button:SetHeight(buttonheight)
		if i == 1 then
			button:SetPoint("TOPLEFT", miniOptionsFrame, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
		else
			button:SetPoint("TOPLEFT", "CharScanner_MiniOptionsFrame_"..buttondata[(i-1)], "BOTTOMLEFT", 0, 0)
		end

		button.text = button:CreateFontString(buttonname.."Text", "OVERLAY", fonts.GameFontNormalSmall.name)
		button.text:SetHeight(fonts.GameFontNormalSmall.height)
		button.text:SetPoint("LEFT", button, "LEFT", maxiconwidth+sizeGeneral.Space, 0)
		button.text:SetJustifyH("LEFT")
		button.text:SetText(L[buttondata[i]])
		local tw = button.text:GetStringWidth()
		if tw > maxtextwidth then
			maxtextwidth = tw
		end
		--button.text:SetWidth(tw) -- TEST
	end

	local framewidth = sizeGeneral.Offset + maxiconwidth + sizeGeneral.Offset + maxtextwidth + sizeGeneral.Offset
	miniOptionsFrame:SetWidth(framewidth)

	for i = 1, #buttondata do
		local button = _G["CharScanner_MiniOptionsFrame_"..buttondata[i]]

		button:SetWidth(framewidth-(2*sizeGeneral.Offset))
		
		local bottomright = (framewidth-(2*sizeGeneral.Offset)) - Textures[ buttondata[i] ].width

		button.textureNormal = button:CreateTexture(nil, "ARTWORK")
		button.textureNormal:SetPoint("TOPLEFT", 2, -2)
		button.textureNormal:SetPoint("BOTTOMRIGHT", -(bottomright + 2), 2)
		button.textureNormal:SetTexture(Textures[ buttondata[i] ].path)
		button.textureNormal:SetTexCoord(Textures[ buttondata[i] ].coords[1], Textures[ buttondata[i] ].coords[2], Textures[ buttondata[i] ].coords[3], Textures[ buttondata[i] ].coords[4])
		button:SetNormalTexture(button.textureNormal)

		button.texturePush = button:CreateTexture(nil, "ARTWORK")
		button.texturePush:SetPoint("TOPLEFT", 3, -3)
		button.texturePush:SetPoint("BOTTOMRIGHT", -(bottomright + 3), 3)
		button.texturePush:SetTexture(Textures[ buttondata[i] ].path)
		button.texturePush:SetTexCoord(Textures[ buttondata[i] ].coords[1], Textures[ buttondata[i] ].coords[2], Textures[ buttondata[i] ].coords[3], Textures[ buttondata[i] ].coords[4])
		button:SetPushedTexture(button.texturePush)

		button.textureDisable = button:CreateTexture(nil, "ARTWORK")
		button.textureDisable:SetPoint("TOPLEFT", 2, -2)
		button.textureDisable:SetPoint("BOTTOMRIGHT", -(bottomright + 2), 2)
		button.textureDisable:SetTexture(Textures[ buttondata[i] ].path)
		button.textureDisable:SetTexCoord(Textures[ buttondata[i] ].coords[1], Textures[ buttondata[i] ].coords[2], Textures[ buttondata[i] ].coords[3], Textures[ buttondata[i] ].coords[4])
		button:SetDisabledTexture(button.textureDisable)

		local shaderSupported = button.textureDisable:SetDesaturated(true)
		if not shaderSupported then
			button.textureDisable:SetVertexColor(0.5, 0.5, 0.5)
		end

		button.Highlight = button:CreateTexture(nil, "ARTWORK")
		button.Highlight:SetPoint("TOPLEFT", button, "TOPLEFT", 0, 0)
		button.Highlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 0, 0)
		button.Highlight:SetTexture(0.2, 0.2, 0.2, 1)
		button:SetHighlightTexture(button.Highlight)

		if (buttondata[i] == "GroupScanAdd" or buttondata[i] == "GroupScan") and (not inPartyOrRaid or GroupScanIsActive) then
			button:Disable()
			button.text:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
		end
		if buttondata[i] == "SingleScan" and SingleScanIsActive then
			button:Disable()
			button.text:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
		end

		if i == 1 then
			button:SetScript("OnClick", function()
				CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
				CharScanner_MainFrame:Show()
				CharScanner:ScanUnit("target")
				CharScanner_MiniOptionsFrame:Hide()
			end)
		elseif i == 2 then
			button:SetScript("OnClick", function()
				CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
				CharScanner_MainFrame:Show()
				CharScanner:ScanGroup(false)
				CharScanner_MiniOptionsFrame:Hide()
			end)
		elseif i == 3 then
			button:SetScript("OnClick", function()
				CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
				CharScanner_MainFrame:Show()
				CharScanner:ScanGroup(true)
				CharScanner_MiniOptionsFrame:Hide()
			end)
		elseif i == 4 then
			button:SetScript("OnClick", function()
				CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
				InterfaceOptionsFrame_OpenToCategory("CharScanner")
				CharScanner_MiniOptionsFrame:Hide()
			end)
		end

		button:SetScript("OnEnter", function() MiniOptions_TimerStart = GetTime() end)
		button:SetScript("OnLeave", function() MiniOptions_TimerStart = GetTime() end)
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:NodeClick(lvl, name1, index1, name2, index2, name3, index3)
	if lvl == 1 then
	-- level 1 START
		if _G["CharScanner_MainFrame_StatsLevel2_"..name1]:IsVisible() then
			_G["CharScanner_MainFrame_StatsLevel1_"..name1.."Collapse"]:SetTexture(Textures.Plus.path)
			_G["CharScanner_MainFrame_StatsLevel1_"..name1.."Collapse"]:SetTexCoord(Textures.Plus.coords[1], Textures.Plus.coords[2], Textures.Plus.coords[3], Textures.Plus.coords[4])

			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel2_"..name1, false)
			CharScanner_MainFrame_Stats:SetHeight(sizeStatFrame.BaseHeight + h)

			if index1 == 1 then
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[2].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel1_"..DATA[1].name, "BOTTOMLEFT", 0, 0)
			elseif index1 > 1 and index1 < #DATA then
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[index1+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel1_"..DATA[index1].name, "BOTTOMLEFT", 0, 0)
			end
		else
			_G["CharScanner_MainFrame_StatsLevel1_"..name1.."Collapse"]:SetTexture(Textures.Minus.path)
			_G["CharScanner_MainFrame_StatsLevel1_"..name1.."Collapse"]:SetTexCoord(Textures.Minus.coords[1], Textures.Minus.coords[2], Textures.Minus.coords[3], Textures.Minus.coords[4])
			
			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel2_"..name1, true)
			CharScanner_MainFrame_Stats:SetHeight(sizeStatFrame.BaseHeight + h)

			if index1 == 1 then
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[2].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..name1, "BOTTOMLEFT", 0, 0)
			elseif index1 > 1 and index1 < #DATA then
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[index1+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..name1, "BOTTOMLEFT", 0, 0)
			end
		end
	-- level 1 END
	elseif lvl == 2 then
	-- level 2 START
		if _G["CharScanner_MainFrame_StatsLevel3_"..name1..name2]:IsVisible() then
			_G["CharScanner_MainFrame_StatsLevel2_"..name1..name2.."Collapse"]:SetTexture(Textures.Plus.path)
			_G["CharScanner_MainFrame_StatsLevel2_"..name1..name2.."Collapse"]:SetTexCoord(Textures.Plus.coords[1], Textures.Plus.coords[2], Textures.Plus.coords[3], Textures.Plus.coords[4])

			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel3_"..name1..name2, false)
			CharScanner_MainFrame_Stats:SetHeight(sizeStatFrame.BaseHeight + h)

			if index2 == 1 then
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][2].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][1].name, "BOTTOMLEFT", 0, 0)
			elseif index2 > 1 and index2 < #DATA[index1] then
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2].name, "BOTTOMLEFT", 0, 0)
			end
		else
			_G["CharScanner_MainFrame_StatsLevel2_"..name1..name2.."Collapse"]:SetTexture(Textures.Minus.path)
			_G["CharScanner_MainFrame_StatsLevel2_"..name1..name2.."Collapse"]:SetTexCoord(Textures.Minus.coords[1], Textures.Minus.coords[2], Textures.Minus.coords[3], Textures.Minus.coords[4])

			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel3_"..name1..name2, true)
			CharScanner_MainFrame_Stats:SetHeight(sizeStatFrame.BaseHeight + h)

			if index2 == 1 then
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][2].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..name2, "BOTTOMLEFT", 0, 0)
			elseif index2 > 1 and index2 < #DATA[index1] then
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..name2..#DATA[index1][index2], "BOTTOMLEFT", 0, 0)
				local c1 = #DATA[index1][index2]
				local c2 = #DATA[index1][index2][c1].statistic
				if _G["CharScanner_MainFrame_StatsLevel4_"..DATA[index1].name..DATA[index1][index2].name..DATA[index1][index2][c1].name..c2]:IsVisible() then
					_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[index1].name..DATA[index1][index2].name..DATA[index1][index2][c1].name..c2, "BOTTOMLEFT", 0, 0)
				end
			end
		end
	-- level 2 END
	elseif lvl == 3 then
	-- level 3 START
		if _G["CharScanner_MainFrame_StatsLevel4_"..name1..name2..name3]:IsVisible() then
			_G["CharScanner_MainFrame_StatsLevel3_"..name1..name2..index3.."Collapse"]:SetTexture(Textures.Plus.path)
			_G["CharScanner_MainFrame_StatsLevel3_"..name1..name2..index3.."Collapse"]:SetTexCoord(Textures.Plus.coords[1], Textures.Plus.coords[2], Textures.Plus.coords[3], Textures.Plus.coords[4])

			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel4_"..name1..name2..name3, false)
			CharScanner_MainFrame_Stats:SetHeight(sizeStatFrame.BaseHeight + h)

			if index3 == 1 then
				_G["CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name.."2"]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name.."1", "BOTTOMLEFT", 0, 0)
			elseif index3 > 1 and index3 < #DATA[index1][index2] then
				_G["CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name..(index3+1)]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name..index3, "BOTTOMLEFT", 0, 0)
			elseif index3 == #DATA[index1][index2] and index2 < #DATA[index1] then
					_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name..#DATA[index1][index2], "BOTTOMLEFT", 0, 0)
			end
		else
			_G["CharScanner_MainFrame_StatsLevel3_"..name1..name2..index3.."Collapse"]:SetTexture(Textures.Minus.path)
			_G["CharScanner_MainFrame_StatsLevel3_"..name1..name2..index3.."Collapse"]:SetTexCoord(Textures.Minus.coords[1], Textures.Minus.coords[2], Textures.Minus.coords[3], Textures.Minus.coords[4])

			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel4_"..name1..name2..name3, true)
			CharScanner_MainFrame_Stats:SetHeight(sizeStatFrame.BaseHeight + h)

			if index3 == 1 then
				_G["CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name.."2"]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[index1].name..DATA[index1][index2].name..DATA[index1][index2][1].name, "BOTTOMLEFT", 0, 0)
			elseif index3 > 1 and index3 < #DATA[index1][index2] then
				_G["CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name..(index3+1)]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[index1].name..DATA[index1][index2].name..DATA[index1][index2][index3].name..#DATA[index1][index2][index3].statistic, "BOTTOMLEFT", 0, 0)
			elseif index3 == #DATA[index1][index2] and index2 < #DATA[index1] then
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[index1].name..DATA[index1][index2].name..DATA[index1][index2][index3].name..#DATA[index1][index2][index3].statistic, "BOTTOMLEFT", 0, 0)
			end
		end
	-- level 3 END
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:UpdateHistory()
	CharScanner:CreateGUIhistory()
	CharScanner:FillHistoryList()
	CharScanner:UpdateHistoryStatus()
end

function CharScanner:CreateGUIhistory()
	for k, v in pairs(SCAN) do -- clean
		if not v.time then
			SCAN[k] = nil
		end
	end

	table.wipe(GUI_history_data)
	table.wipe(GUI_history_Role)

	local x = 1
	for k, v in pairs(SCAN) do
		GUI_history_data[x] = {}
		GUI_history_data[x].guid = k
		GUI_history_data[x].classEN = v.classEN
		GUI_history_data[x].name = v.name
		GUI_history_data[x].time = v.time

		local firstRole
		if v.talentGroup1name then
			local col = color.Grey.hex
			local isActive = 0
			if v.activeTalentGroup == 1 then
				col = color.White.hex
				isActive = 1
			end
			local role = classes[v.classEN].role[v.talentGroup1role]
			local rolesort = classes[v.classEN].role[v.talentGroup1role]

			if v.classEN == "DEATHKNIGHT" then
				if SCAN[k].talents[1][1][1][3]+SCAN[k].talents[1][2][1][3]+SCAN[k].talents[1][3][1][3] >= 10 then -- Blade Barrier+Toughness+Anticipation
					-- TANK
					role = "T"
					rolesort = "T"
				else
					-- DAMAGE
					role = "D"
					rolesort = "D"
				end
			elseif v.classEN == "DRUID" then
				if role == "T" and SCAN[k].talents[1][2][2][3]+SCAN[k].talents[1][2][8][1] < 4 then -- Thick Hide+Protector of the Pack
					-- DAMAGE
					role = "D"
					rolesort = "D"
				end
			end

			CreateGUIhistoryRole(role, isActive)
			firstRole = role
			if role == "D" then
				role = "|cff"..color.Firebrick.hex..L["D"].."|r"
				rolesort = "3"..rolesort
			elseif role == "T" then
				role = "|cff"..color.Brown.hex..L["T"].."|r"
				rolesort = "1"..rolesort
			elseif role == "H" then
				role = "|cff"..color.Green.hex..L["H"].."|r"
				rolesort = "2"..rolesort
			end
			if isActive > 0 then
				GUI_history_data[x].mainrole = role
				GUI_history_data[x].mainroleNum = rolesort
			end
			local unspent = ""
			if v.talentGroup1pointsUnspent > 0 then
				unspent = " |cff"..color.Red.hex.."("..v.talentGroup1pointsUnspent..")|r"
			end
			GUI_history_data[x].talents1 = (role or "").." |cff"..col..(v.talentGroup1name or "").." "..(v.talentGroup1pointsString or "").."|r"..unspent
			GUI_history_data[x].talents1Num = (rolesort or "").." "..(v.talentGroup1name or "").." "..(v.talentGroup1pointsString or "")
		else
			GUI_history_data[x].mainrole = ""
			GUI_history_data[x].mainroleNum = ""
			GUI_history_data[x].talents1 = ""
			GUI_history_data[x].talents1Num = ""
		end
		if v.talentGroup2name then
			local col = color.Grey.hex
			local isActive = 0
			if v.activeTalentGroup == 2 then
				col = color.White.hex
				isActive = 1
			end
			local role = classes[v.classEN].role[v.talentGroup2role]
			local rolesort = classes[v.classEN].role[v.talentGroup2role]

			if v.classEN == "DEATHKNIGHT" then
				if SCAN[k].talents[2][1][1][3]+SCAN[k].talents[2][2][1][3]+SCAN[k].talents[2][3][1][3] >= 10 then -- Blade Barrier+Toughness+Anticipation
					-- TANK
					role = "T"
					rolesort = "T"
				else
					-- DAMAGE
					role = "D"
					rolesort = "D"
				end
			elseif v.classEN == "DRUID" then
				if role == "T" and SCAN[k].talents[2][2][2][3]+SCAN[k].talents[2][2][8][1] < 4 then -- Thick Hide+Protector of the Pack
					-- DAMAGE
					role = "D"
					rolesort = "D"
				end
			end

			if isActive == 1 then
				if firstRole == role then
					CreateGUIhistoryRole(role, isActive, true)
				else
					CreateGUIhistoryRole(role, isActive)
				end
			else
				if firstRole ~= role then
					CreateGUIhistoryRole(role, isActive)
				end
			end
			if role == "D" then
				role = "|cff"..color.Firebrick.hex..L["D"].."|r"
				rolesort = "3"..rolesort
			elseif role == "T" then
				role = "|cff"..color.Brown.hex..L["T"].."|r"
				rolesort = "1"..rolesort
			elseif role == "H" then
				role = "|cff"..color.Green.hex..L["H"].."|r"
				rolesort = "2"..rolesort
			end
			if isActive > 0 then
				GUI_history_data[x].mainrole = role
				GUI_history_data[x].mainroleNum = rolesort
			end
			local unspent = ""
			if v.talentGroup2pointsUnspent > 0 then
				unspent = " |cff"..color.Red.hex.."("..v.talentGroup2pointsUnspent..")|r"
			end
			GUI_history_data[x].talents2 = (role or "").." |cff"..col..(v.talentGroup2name or "").." "..(v.talentGroup2pointsString or "").."|r"..unspent
			GUI_history_data[x].talents2Num = (rolesort or "").." "..(v.talentGroup2name or "").." "..(v.talentGroup2pointsString or "")
		else
			GUI_history_data[x].talents2 = ""
			GUI_history_data[x].talents2Num = ""
		end

		GUI_history_data[x].itemlevel = cnum(v.itemlevelAvg)
		GUI_history_data[x].itemlevelNum = v.itemlevelAvg

		-- raid selection START
		local col
		if GUI_history_RaidL3 then
			local achievement = math_floor((v.DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementCount/DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementTotal*100) * mult2 + 0.5) / mult2
			if DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementTotal == v.DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementCount then
				col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
			end
			GUI_history_data[x].achievement = col..v.DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementCount.."|r / |cff"..color.Level3.hex..DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementTotal.."|r"
			GUI_history_data[x].achievementNum = achievement
			local boss = math_floor((v.DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticBoss/DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticTotal*100) * mult2 + 0.5) / mult2
			if DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticTotal == v.DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticBoss then
				col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
			end
			GUI_history_data[x].boss = col..v.DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticBoss.."|r / |cff"..color.Level3.hex..DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticTotal.."|r"
			GUI_history_data[x].bossNum = boss
			GUI_history_data[x].kills = v.DATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticBossAll
		elseif GUI_history_RaidL2 then
			local achievement = math_floor((v.DATA[GUI_history_RaidL1][GUI_history_RaidL2].achievementCount/DATA[GUI_history_RaidL1][GUI_history_RaidL2].achievementTotal*100) * mult2 + 0.5) / mult2
			if DATA[GUI_history_RaidL1][GUI_history_RaidL2].achievementTotal == v.DATA[GUI_history_RaidL1][GUI_history_RaidL2].achievementCount then
				col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
			end
			GUI_history_data[x].achievement = col..v.DATA[GUI_history_RaidL1][GUI_history_RaidL2].achievementCount.."|r / |cff"..color.Level2.hex..DATA[GUI_history_RaidL1][GUI_history_RaidL2].achievementTotal.."|r"
			GUI_history_data[x].achievementNum = achievement
			local boss = math_floor((v.DATA[GUI_history_RaidL1][GUI_history_RaidL2].statisticBoss/DATA[GUI_history_RaidL1][GUI_history_RaidL2].statisticTotal*100) * mult2 + 0.5) / mult2
			if DATA[GUI_history_RaidL1][GUI_history_RaidL2].statisticTotal == v.DATA[GUI_history_RaidL1][GUI_history_RaidL2].statisticBoss then
				col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
			end
			GUI_history_data[x].boss = col..v.DATA[GUI_history_RaidL1][GUI_history_RaidL2].statisticBoss.."|r / |cff"..color.Level2.hex..DATA[GUI_history_RaidL1][GUI_history_RaidL2].statisticTotal.."|r"
			GUI_history_data[x].bossNum = boss
			GUI_history_data[x].kills = v.DATA[GUI_history_RaidL1][GUI_history_RaidL2].statisticBossAll
		else
			if GUI_history_RaidL1 == 0 then -- Total
				local achievement = math_floor((v.statsAchievementCount/v.statsAchievementTotal*100) * mult2 + 0.5) / mult2
				GUI_history_data[x].achievement = "|cff"..color.LightYellow.hex..cnum(achievement).."%|r"
				GUI_history_data[x].achievementNum = achievement
				local boss = math_floor((v.statsStatisticCount/v.statsStatisticTotal*100) * mult2 + 0.5) / mult2
				GUI_history_data[x].boss = "|cff"..color.LightYellow.hex..cnum(boss).."%|r"
				GUI_history_data[x].bossNum = boss
				GUI_history_data[x].kills = v.statsStatisticBossKills
			else
				local achievement = math_floor((v.DATA[GUI_history_RaidL1].achievementCount/DATA[GUI_history_RaidL1].achievementTotal*100) * mult2 + 0.5) / mult2
				if DATA[GUI_history_RaidL1].achievementTotal == v.DATA[GUI_history_RaidL1].achievementCount then
					col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
				end
				GUI_history_data[x].achievement = col..v.DATA[GUI_history_RaidL1].achievementCount.."|r / |cff"..color.Level1.hex..DATA[GUI_history_RaidL1].achievementTotal.."|r"
				GUI_history_data[x].achievementNum = achievement
				local boss = math_floor((v.DATA[GUI_history_RaidL1].statisticBoss/DATA[GUI_history_RaidL1].statisticTotal*100) * mult2 + 0.5) / mult2
				if DATA[GUI_history_RaidL1].statisticTotal == v.DATA[GUI_history_RaidL1].statisticBoss then
					col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
				end
				GUI_history_data[x].boss = col..v.DATA[GUI_history_RaidL1].statisticBoss.."|r / |cff"..color.Level1.hex..DATA[GUI_history_RaidL1].statisticTotal.."|r"
				GUI_history_data[x].bossNum = boss
				GUI_history_data[x].kills = v.DATA[GUI_history_RaidL1].statisticBossAll
			end
		end
		-- raid selection END

		x = x + 1
	end
	GUI_history_Count = #GUI_history_data
	NumScannedChars = #GUI_history_data

	local sortfunc
	if GUI_history_SortBy == 1 then -- 1 = class
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.classEN == b.classEN then
					if a.name < b.name then
						return true
					end
				elseif a.classEN < b.classEN then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.classEN == b.classEN then
					if a.name < b.name then
						return true
					end
				elseif a.classEN > b.classEN then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 2 then -- 2 = name
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.name < b.name then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.name > b.name then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 3 then -- 3 = time
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.time == b.time then
					if a.name < b.name then
						return true
					end
				elseif a.time < b.time then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.time == b.time then
					if a.name < b.name then
						return true
					end
				elseif a.time > b.time then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 4 then -- 4 = mainrole
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.mainroleNum == b.mainroleNum then
					if a.classEN == b.classEN then
						if a.name < b.name then
							return true
						end
					elseif a.classEN < b.classEN then
						return true
					end
				elseif a.mainroleNum < b.mainroleNum then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.mainroleNum == b.mainroleNum then
					if a.classEN == b.classEN then
						if a.name < b.name then
							return true
						end
					elseif a.classEN < b.classEN then
						return true
					end
				elseif a.mainroleNum > b.mainroleNum then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 5 then -- 5 = talents1
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.talents1Num == b.talents1Num then
					if a.name < b.name then
						return true
					end
				elseif a.talents1Num < b.talents1Num then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.talents1Num == b.talents1Num then
					if a.name < b.name then
						return true
					end
				elseif a.talents1Num > b.talents1Num then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 6 then -- 6 = talents2
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.talents2Num == b.talents2Num then
					if a.name < b.name then
						return true
					end
				elseif a.talents2Num < b.talents2Num then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.talents2Num == b.talents2Num then
					if a.name < b.name then
						return true
					end
				elseif a.talents2Num > b.talents2Num then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 7 then -- 7 = itemlevel (itemlevelNum)
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.itemlevelNum == b.itemlevelNum then
					if a.classEN == b.classEN then
						if a.name < b.name then
							return true
						end
					elseif a.classEN < b.classEN then
						return true
					end
				elseif a.itemlevelNum > b.itemlevelNum then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.itemlevelNum == b.itemlevelNum then
					if a.classEN == b.classEN then
						if a.name < b.name then
							return true
						end
					elseif a.classEN < b.classEN then
						return true
					end
				elseif a.itemlevelNum < b.itemlevelNum then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 8 then -- 8 = achievement (achievementNum)
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.achievementNum == b.achievementNum then
					if a.bossNum == b.bossNum then
						if a.kills == b.kills then
							if a.name < b.name then
								return true
							end
						elseif a.kills > b.kills then
							return true
						end
					elseif a.bossNum > b.bossNum then
						return true
					end
				elseif a.achievementNum > b.achievementNum then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.achievementNum == b.achievementNum then
					if a.bossNum == b.bossNum then
						if a.kills == b.kills then
							if a.name < b.name then
								return true
							end
						elseif a.kills < b.kills then
							return true
						end
					elseif a.bossNum < b.bossNum then
						return true
					end
				elseif a.achievementNum < b.achievementNum then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 9 then -- 9 = boss (bossNum)
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.bossNum == b.bossNum then
					if a.achievementNum == b.achievementNum then
						if a.kills == b.kills then
							if a.name < b.name then
								return true
							end
						elseif a.kills > b.kills then
							return true
						end
					elseif a.achievementNum > b.achievementNum then
						return true
					end
				elseif a.bossNum > b.bossNum then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.bossNum == b.bossNum then
					if a.achievementNum == b.achievementNum then
						if a.kills == b.kills then
							if a.name < b.name then
								return true
							end
						elseif a.kills < b.kills then
							return true
						end
					elseif a.achievementNum < b.achievementNum then
						return true
					end
				elseif a.bossNum < b.bossNum then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 10 then -- 10 = kills (kills)
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.kills == b.kills then
					if a.bossNum == b.bossNum then
						if a.achievementNum == b.achievementNum then
							if a.name < b.name then
								return true
							end
						elseif a.achievementNum > b.achievementNum then
							return true
						end
					elseif a.bossNum > b.bossNum then
						return true
					end
				elseif a.kills > b.kills then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.kills == b.kills then
					if a.bossNum == b.bossNum then
						if a.achievementNum == b.achievementNum then
							if a.name < b.name then
								return true
							end
						elseif a.achievementNum < b.achievementNum then
							return true
						end
					elseif a.bossNum < b.bossNum then
						return true
					end
				elseif a.kills < b.kills then
					return true
				end
			end
		end
	end

	table.sort(GUI_history_data, sortfunc)
end

function CharScanner:GUIsort(sortBy, l1, l2, l3)
	if l1 then
		GUI_history_RaidL1 = l1
		GUI_history_RaidL2 = l2
		GUI_history_RaidL3 = l3
		local name
		if GUI_history_RaidL3 then
			name = L[DATA[l1][l2][l3].name].." ("..L[DATA[l1][l2].name]..")"
		elseif GUI_history_RaidL2 then
			name = L[DATA[l1][l2].name]
		else
			if l1 == 0 then
				name = L["Total"]
			else
				name = L[DATA[l1].name]
			end
		end
		CharScanner_HistoryFrame_Sort_RaidSelectMainButtonName:SetText(name)
		CharScanner_HistoryFrame_Sort_RaidSelectMainButton_Expand:SetTexture(Textures.Expand.path)
		CharScanner_HistoryFrame_Sort_RaidSelectMainButton_Expand:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])
		CharScanner_HistoryFrame_Sort_RaidSelectFrame:Hide()
		GUI_history_SortOrder = 1
		GUI_history_SortBy = 9 -- Boss
	else
		if GUI_history_SortBy == sortBy then
			if GUI_history_SortOrder == 1 then
				GUI_history_SortOrder = 0
			else
				GUI_history_SortOrder = 1
			end
		else
			GUI_history_SortOrder = 1
		end
		GUI_history_SortBy = sortBy
	end
	CharScanner:UpdateSortButtons()
	CharScanner:CreateGUIhistory()
	CharScanner:FillHistoryList()
end

function CharScanner:UpdateSortButtons()
	local arrow
	for i=1, 10 do
		if i == 1 then     -- 1 = class
			arrow = "CharScanner_HistoryFrame_SortButton_IconArrow"
		elseif i == 2 then -- 2 = name
			arrow = "CharScanner_HistoryFrame_SortButton_NameArrow"
		elseif i == 3 then -- 3 = time
			arrow = "CharScanner_HistoryFrame_SortButton_TimeArrow"
		elseif i == 4 then -- 4 = mainrole
			arrow = "CharScanner_HistoryFrame_SortButton_MainRoleArrow"
		elseif i == 5 then -- 5 = talents1
			arrow = "CharScanner_HistoryFrame_SortButton_Talents1Arrow"
		elseif i == 6 then -- 6 = talents2
			arrow = "CharScanner_HistoryFrame_SortButton_Talents2Arrow"
		elseif i == 7 then -- 7 = ItemLevel
			arrow = "CharScanner_HistoryFrame_SortButton_ItemLevelArrow"
		elseif i == 8 then -- 8 = Achievement
			arrow = "CharScanner_HistoryFrame_SortButton_AchievementArrow"
		elseif i == 9 then -- 9 = Boss
			arrow = "CharScanner_HistoryFrame_SortButton_BossArrow"
		elseif i == 10 then -- 10 = Kills
			arrow = "CharScanner_HistoryFrame_SortButton_KillsArrow"
		end
		if GUI_history_SortBy == i then
			if GUI_history_SortOrder == 1 then _G[arrow]:SetTexCoord(Textures.SortArrow.coords1[1], Textures.SortArrow.coords1[2], Textures.SortArrow.coords1[3], Textures.SortArrow.coords1[4]) end
			if GUI_history_SortOrder == 0 then _G[arrow]:SetTexCoord(Textures.SortArrow.coords2[1], Textures.SortArrow.coords2[2], Textures.SortArrow.coords2[3], Textures.SortArrow.coords2[4]) end
		else
			_G[arrow]:SetTexCoord(0, 0, 0, 0)
		end
	end
end

function CharScanner:OnVerticalScroll()
	CharScanner_HistoryFrameScrollFrameScrollBar:SetValue(arg1)
	CharScanner_HistoryFrameScrollFrame.offset = math_floor((arg1 / sizeGeneral.BarHeight) + 0.5)
	CharScanner:FillHistoryList()
end

function CharScanner:FillHistoryList()
	local showtableNumSize = GUI_history_maxNum

	if GUI_history_Count < showtableNumSize then
		for i=1, showtableNumSize do
			local button = _G["CharScanner_HistoryFrame_Button"..i]
			button:Hide()
			button.guid = nil
		end
		showtableNumSize = GUI_history_Count
	end

	FauxScrollFrame_Update(CharScanner_HistoryFrameScrollFrame, GUI_history_Count, showtableNumSize, sizeGeneral.BarHeight)

	CharScanner_HistoryFrameScrollFrame:Show()
	if GUI_history_Count <= GUI_history_maxNum then
		CharScanner_HistoryFrameScrollFrame:SetAlpha(0.4)
		CharScanner_HistoryFrameScrollFrameScrollBarScrollUpButton:Disable()   -- cosmetic fix
		CharScanner_HistoryFrameScrollFrameScrollBarScrollDownButton:Disable() -- cosmetic fix
		--CharScanner_HistoryFrameScrollFrame:Hide()
	else
		CharScanner_HistoryFrameScrollFrame:SetAlpha(1)
		--CharScanner_HistoryFrameScrollFrameScrollBarScrollUpButton:Enable()   -- set back automatically in FauxScrollFrameTemplate
		--CharScanner_HistoryFrameScrollFrameScrollBarScrollDownButton:Enable() -- set back automatically in FauxScrollFrameTemplate
		--CharScanner_HistoryFrameScrollFrame:Show()
	end

	local j = 1
	local sliderPos = CharScanner_HistoryFrameScrollFrame.offset
	for i=sliderPos+1, sliderPos+showtableNumSize do
		if sliderPos < GUI_history_Count then
			_G["CharScanner_HistoryFrame_Button"..j.."Icon"]:SetTexCoord(classes[GUI_history_data[i].classEN].icon[1], classes[GUI_history_data[i].classEN].icon[2], classes[GUI_history_data[i].classEN].icon[3], classes[GUI_history_data[i].classEN].icon[4])
			_G["CharScanner_HistoryFrame_Button"..j.."Name"]:SetText(GUI_history_data[i].name)
			_G["CharScanner_HistoryFrame_Button"..j.."Name"]:SetTextColor(classcolors[GUI_history_data[i].classEN].r, classcolors[GUI_history_data[i].classEN].g, classcolors[GUI_history_data[i].classEN].b, 1)
			_G["CharScanner_HistoryFrame_Button"..j.."Time"]:SetText(GUI_history_data[i].time)

			_G["CharScanner_HistoryFrame_Button"..j.."MainRole"]:SetText(GUI_history_data[i].mainrole)
			_G["CharScanner_HistoryFrame_Button"..j.."Talents1"]:SetText(GUI_history_data[i].talents1)
			_G["CharScanner_HistoryFrame_Button"..j.."Talents2"]:SetText(GUI_history_data[i].talents2)

			_G["CharScanner_HistoryFrame_Button"..j.."ItemLevel"]:SetText(GUI_history_data[i].itemlevel)

			_G["CharScanner_HistoryFrame_Button"..j.."Achievement"]:SetText(GUI_history_data[i].achievement)
			_G["CharScanner_HistoryFrame_Button"..j.."Boss"]:SetText(GUI_history_data[i].boss)
			_G["CharScanner_HistoryFrame_Button"..j.."Kills"]:SetText(GUI_history_data[i].kills)

			_G["CharScanner_HistoryFrame_Button"..j]:Show()
			_G["CharScanner_HistoryFrame_Button"..j].guid = GUI_history_data[i].guid

			_G["CharScanner_HistoryFrame_Button"..j.."Arrow"]:Hide()
			if GUIselectedGUID == GUI_history_data[i].guid then
				_G["CharScanner_HistoryFrame_Button"..j.."Arrow"]:Show()
			end

			j = j + 1
		end
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:UpdateAchievementDetail(l1, l2, l3, sel)
	-- do not show AchievementDetail if last scan is incomplete
	if not LatestScanIsComplete then
		CharScanner:AchievementDetailHide()
		return
	end

	if sel then
		GUI_achievementdetail_Select = sel
		local name
		local col
		if sel == 1 then
			name = ALL
			col = "White"
		elseif sel == 2 then
			name = COMPLETE
			col = "Green"
		elseif sel == 3 then
			name = INCOMPLETE
			col = "Red"
		end
		CharScanner_AchievementDetailFrame_Sort_AchievementDetailSelectMainButtonName:SetText(name)
		CharScanner_AchievementDetailFrame_Sort_AchievementDetailSelectMainButtonName:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])
		CharScanner_AchievementDetailFrame_Sort_AchievementDetailSelectMainButton_Expand:SetTexture(Textures.Expand.path)
		CharScanner_AchievementDetailFrame_Sort_AchievementDetailSelectMainButton_Expand:SetTexCoord(Textures.Expand.coords[1], Textures.Expand.coords[2], Textures.Expand.coords[3], Textures.Expand.coords[4])
		CharScanner_AchievementDetailFrame_Sort_AchievementDetailSelectFrame:Hide()
	else
		GUI_achievementdetail_L1 = l1
		GUI_achievementdetail_L2 = l2
		GUI_achievementdetail_L3 = l3
	end

	CharScanner_AchievementDetailFrame:Show()
	AchievementDetailFrameIsShown = true
	CharScanner:CreateGUIachievementDetail()
	CharScanner:FillAchievementDetailList()
	CharScanner:UpdateAchievementDetailStatus()
end

function CharScanner:UpdateAchievementDetailStatus()
	local all = 0
	local complete = 0
	local incomplete = 0

	if GUI_achievementdetail_L3 then
		all        = DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievementTotal
		complete   = SCAN[GUIselectedGUID].DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievementCount
		incomplete = all-complete
	elseif GUI_achievementdetail_L2 then
		all        = DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievementTotal
		complete   = SCAN[GUIselectedGUID].DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievementCount
		incomplete = all-complete
	else
		all        = DATA[GUI_achievementdetail_L1].achievementTotal
		complete   = SCAN[GUIselectedGUID].DATA[GUI_achievementdetail_L1].achievementCount
		incomplete = all-complete
	end

	CharScanner_AchievementDetailFrame_Status_Text:SetText("|cff"..color.Green.hex..complete.."|r | |cff"..color.Red.hex..incomplete.."|r | |cff"..color.White.hex..all.."|r")
end

function CharScanner:CreateGUIachievementDetail()
	-- --------------------------------------------
	-- DUMMY_FactionFix START
	-- --------------------------------------------
	if SCAN[GUIselectedGUID].faction == "Horde" then
		DATA[3][1].achievement[54] = 4297
	else
		DATA[3][1].achievement[54] = 4298
	end
	-- --------------------------------------------
	-- DUMMY_FactionFix END
	-- --------------------------------------------

	table.wipe(GUI_achievementdetail_data)
	local x = 1
	if GUI_achievementdetail_L3 then

		if GUI_achievementdetail_Select == 1 then
			for i = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement do
				GUI_achievementdetail_data[x] = {}
				GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i]
				GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ]
				if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ] then
					GUI_achievementdetail_data[x].color = 1
				end
				x = x + 1
			end
			GUI_achievementdetail_Count = x - 1
		elseif GUI_achievementdetail_Select == 2 then
			for i = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement do
				if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ] then
					GUI_achievementdetail_data[x] = {}
					GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i]
					GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ]
					GUI_achievementdetail_data[x].color = 1
					x = x + 1
				end
			end
			GUI_achievementdetail_Count = x - 1
		elseif GUI_achievementdetail_Select == 3 then
			for i = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement do
				if not SCAN[GUIselectedGUID].DATA.achievementID or not SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ] then
					GUI_achievementdetail_data[x] = {}
					GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i]
					GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ]
					x = x + 1
				end
			end
			GUI_achievementdetail_Count = x - 1
		end

	elseif GUI_achievementdetail_L2 then

		if GUI_achievementdetail_Select == 1 then
			if not DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement then
				for k = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2] do
					if DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement then
						for i = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement do
							GUI_achievementdetail_data[x] = {}
							GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i]
							GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ]
							if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ] then
								GUI_achievementdetail_data[x].color = 1
							end
							x = x + 1
						end
					end
				end
			else
				for i = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement do
					GUI_achievementdetail_data[x] = {}
					GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i]
					GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ]
					if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ] then
						GUI_achievementdetail_data[x].color = 1
					end
					x = x + 1
				end
			end
			GUI_achievementdetail_Count = x - 1
		elseif GUI_achievementdetail_Select == 2 then
			if not DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement then
				for k = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2] do
					if DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement then
						for i = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement do
							if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ] then
								GUI_achievementdetail_data[x] = {}
								GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i]
								GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ]
								GUI_achievementdetail_data[x].color = 1
								x = x + 1
							end
						end
					end
				end
			else
				for i = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement do
					if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ] then
						GUI_achievementdetail_data[x] = {}
						GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i]
						GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ]
						GUI_achievementdetail_data[x].color = 1
						x = x + 1
					end
				end
			end
			GUI_achievementdetail_Count = x - 1
		elseif GUI_achievementdetail_Select == 3 then
			if not DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement then
				for k = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2] do
					if DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement then
						for i = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement do
							if not SCAN[GUIselectedGUID].DATA.achievementID or not SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ] then
								GUI_achievementdetail_data[x] = {}
								GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i]
								GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ]
								x = x + 1
							end
						end
					end
				end
			else
				for i = 1, #DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement do
					if not SCAN[GUIselectedGUID].DATA.achievementID or not SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ] then
						GUI_achievementdetail_data[x] = {}
						GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i]
						GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ]
						x = x + 1
					end
				end
			end
			GUI_achievementdetail_Count = x - 1
		end

	else

		if GUI_achievementdetail_Select == 1 then
			if not DATA[GUI_achievementdetail_L1].achievement then
				for j = 1, #DATA[GUI_achievementdetail_L1] do
					if DATA[GUI_achievementdetail_L1][j].achievement then
						for i = 1, #DATA[GUI_achievementdetail_L1][j].achievement do
							GUI_achievementdetail_data[x] = {}
							GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][j].achievement[i]
							GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][j].achievement[i] ]
							if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][j].achievement[i] ] then
								GUI_achievementdetail_data[x].color = 1
							end
							x = x + 1
						end
					end
					for k = 1, #DATA[GUI_achievementdetail_L1][j] do
						if DATA[GUI_achievementdetail_L1][j][k].achievement then
							for i = 1, #DATA[GUI_achievementdetail_L1][j][k].achievement do
								GUI_achievementdetail_data[x] = {}
								GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][j][k].achievement[i]
								GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][j][k].achievement[i] ]
								if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][j][k].achievement[i] ] then
									GUI_achievementdetail_data[x].color = 1
								end
								x = x + 1
							end
						end
					end
				end
			else
				for i = 1, #DATA[GUI_achievementdetail_L1].achievement do
					GUI_achievementdetail_data[x] = {}
					GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1].achievement[i]
					GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1].achievement[i] ]
					if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1].achievement[i] ] then
						GUI_achievementdetail_data[x].color = 1
					end
					x = x + 1
				end
			end
			GUI_achievementdetail_Count = x - 1
		elseif GUI_achievementdetail_Select == 2 then
			if not DATA[GUI_achievementdetail_L1].achievement then
				for j = 1, #DATA[GUI_achievementdetail_L1] do
					if DATA[GUI_achievementdetail_L1][j].achievement then
						for i = 1, #DATA[GUI_achievementdetail_L1][j].achievement do
							if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][j].achievement[i] ] then
								GUI_achievementdetail_data[x] = {}
								GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][j].achievement[i]
								GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][j].achievement[i] ]
								GUI_achievementdetail_data[x].color = 1
								x = x + 1
							end
						end
					end
					for k = 1, #DATA[GUI_achievementdetail_L1][j] do
						if DATA[GUI_achievementdetail_L1][j][k].achievement then
							for i = 1, #DATA[GUI_achievementdetail_L1][j][k].achievement do
								if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][j][k].achievement[i] ] then
									GUI_achievementdetail_data[x] = {}
									GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][j][k].achievement[i]
									GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][j][k].achievement[i] ]
									GUI_achievementdetail_data[x].color = 1
									x = x + 1
								end
							end
						end
					end
				end
			else
				for i = 1, #DATA[GUI_achievementdetail_L1].achievement do
					if SCAN[GUIselectedGUID].DATA.achievementID and SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1].achievement[i] ] then
						GUI_achievementdetail_data[x] = {}
						GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1].achievement[i]
						GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1].achievement[i] ]
						GUI_achievementdetail_data[x].color = 1
						x = x + 1
					end
				end
			end
			GUI_achievementdetail_Count = x - 1
		elseif GUI_achievementdetail_Select == 3 then
			if not DATA[GUI_achievementdetail_L1].achievement then
				for j = 1, #DATA[GUI_achievementdetail_L1] do
					if DATA[GUI_achievementdetail_L1][j].achievement then
						for i = 1, #DATA[GUI_achievementdetail_L1][j].achievement do
							if not SCAN[GUIselectedGUID].DATA.achievementID or not SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][j].achievement[i] ] then
								GUI_achievementdetail_data[x] = {}
								GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][j].achievement[i]
								GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][j].achievement[i] ]
								x = x + 1
							end
						end
					end
					for k = 1, #DATA[GUI_achievementdetail_L1][j] do
						if DATA[GUI_achievementdetail_L1][j][k].achievement then
							for i = 1, #DATA[GUI_achievementdetail_L1][j][k].achievement do
								if not SCAN[GUIselectedGUID].DATA.achievementID or not SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1][j][k].achievement[i] ] then
									GUI_achievementdetail_data[x] = {}
									GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1][j][k].achievement[i]
									GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1][j][k].achievement[i] ]
									x = x + 1
								end
							end
						end
					end
				end
			else
				for i = 1, #DATA[GUI_achievementdetail_L1].achievement do
					if not SCAN[GUIselectedGUID].DATA.achievementID or not SCAN[GUIselectedGUID].DATA.achievementID[ DATA[GUI_achievementdetail_L1].achievement[i] ] then
						GUI_achievementdetail_data[x] = {}
						GUI_achievementdetail_data[x].id = DATA[GUI_achievementdetail_L1].achievement[i]
						GUI_achievementdetail_data[x].name = AchievementIDTable[ DATA[GUI_achievementdetail_L1].achievement[i] ]
						x = x + 1
					end
				end
			end
			GUI_achievementdetail_Count = x - 1
		end

	end

	local sortfunc
	if GUI_achievementdetail_SortBy == 1 then -- 1 = id
		if GUI_achievementdetail_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.id < b.id then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.id > b.id then
					return true
				end
			end
		end
	elseif GUI_achievementdetail_SortBy == 2 then -- 2 = name
		if GUI_achievementdetail_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.name < b.name then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.name > b.name then
					return true
				end
			end
		end
	end

	table.sort(GUI_achievementdetail_data, sortfunc)
end

function CharScanner:GUIachievementdetailsort(sortBy)
	if GUI_achievementdetail_SortBy == sortBy then
		if GUI_achievementdetail_SortOrder == 1 then
			GUI_achievementdetail_SortOrder = 0
		else
			GUI_achievementdetail_SortOrder = 1
		end
	else
		GUI_achievementdetail_SortOrder = 1
	end
	GUI_achievementdetail_SortBy = sortBy
	CharScanner:UpdateachievementdetailSortButtons()
	CharScanner:CreateGUIachievementDetail()
	CharScanner:FillAchievementDetailList()
end

function CharScanner:UpdateachievementdetailSortButtons()
	local arrow
	for i=1, 2 do
		if i == 1 then -- 1 = id
			arrow = "CharScanner_AchievementDetailFrame_SortButton_IdArrow"
		elseif i == 2 then -- 2 = name
			arrow = "CharScanner_AchievementDetailFrame_SortButton_NameArrow"
		end
		if GUI_achievementdetail_SortBy == i then
			if GUI_achievementdetail_SortOrder == 1 then _G[arrow]:SetTexCoord(Textures.SortArrow.coords1[1], Textures.SortArrow.coords1[2], Textures.SortArrow.coords1[3], Textures.SortArrow.coords1[4]) end
			if GUI_achievementdetail_SortOrder == 0 then _G[arrow]:SetTexCoord(Textures.SortArrow.coords2[1], Textures.SortArrow.coords2[2], Textures.SortArrow.coords2[3], Textures.SortArrow.coords2[4]) end
		else
			_G[arrow]:SetTexCoord(0, 0, 0, 0)
		end
	end
end

function CharScanner:OnVerticalScroll_AchievementDetail()
	CharScanner_AchievementDetailFrameScrollFrameScrollBar:SetValue(arg1)
	CharScanner_AchievementDetailFrameScrollFrame.offset = math_floor((arg1 / sizeGeneral.BarHeight) + 0.5)
	CharScanner:FillAchievementDetailList()
end

function CharScanner:FillAchievementDetailList()
	local showtableNumSize = GUI_achievementdetail_maxNum

	if GUI_achievementdetail_Count < showtableNumSize then
		for i=1, showtableNumSize do
			local button = _G["CharScanner_AchievementDetailFrame_Button"..i]
			button:Hide()
		end
		showtableNumSize = GUI_achievementdetail_Count
	end

	FauxScrollFrame_Update(CharScanner_AchievementDetailFrameScrollFrame, GUI_achievementdetail_Count, showtableNumSize, sizeGeneral.BarHeight)

	CharScanner_AchievementDetailFrameScrollFrame:Show()
	if GUI_achievementdetail_Count <= GUI_achievementdetail_maxNum then
		CharScanner_AchievementDetailFrameScrollFrame:SetAlpha(0.4)
		CharScanner_AchievementDetailFrameScrollFrameScrollBarScrollUpButton:Disable()   -- cosmetic fix
		CharScanner_AchievementDetailFrameScrollFrameScrollBarScrollDownButton:Disable() -- cosmetic fix
		--CharScanner_AchievementDetailFrameScrollFrame:Hide()
	else
		CharScanner_AchievementDetailFrameScrollFrame:SetAlpha(1)
		--CharScanner_AchievementDetailFrameScrollFrameScrollBarScrollUpButton:Enable()   -- set back automatically in FauxScrollFrameTemplate
		--CharScanner_AchievementDetailFrameScrollFrameScrollBarScrollDownButton:Enable() -- set back automatically in FauxScrollFrameTemplate
		--CharScanner_AchievementDetailFrameScrollFrame:Show()
	end

	local j = 1
	local sliderPos = CharScanner_AchievementDetailFrameScrollFrame.offset
	for i=sliderPos+1, sliderPos+showtableNumSize do
		if sliderPos < GUI_achievementdetail_Count then
			_G["CharScanner_AchievementDetailFrame_Button"..j.."Id"]:SetText(GUI_achievementdetail_data[i].id)
			_G["CharScanner_AchievementDetailFrame_Button"..j.."Name"]:SetText(GUI_achievementdetail_data[i].name)
			if GUI_achievementdetail_data[i].color then
				_G["CharScanner_AchievementDetailFrame_Button"..j.."Name"]:SetTextColor(color.Green.rgba[1], color.Green.rgba[2], color.Green.rgba[3], color.Green.rgba[4])
			else
				_G["CharScanner_AchievementDetailFrame_Button"..j.."Name"]:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
			end
			_G["CharScanner_AchievementDetailFrame_Button"..j].achievementID = GUI_achievementdetail_data[i].id

			_G["CharScanner_AchievementDetailFrame_Button"..j]:Show()
			j = j + 1
		end
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:PreShowSavedData(guid, switch)
	CharScanner:ClearData()
	CharScanner:ScanFrameToggle(true)
	if CompareFrameIsShown and switch then
		CharScanner:HistoryFrameToggle()
	end
	DataStep = 0
	CharScanner:ShowData(1, guid, false)
	CharScanner:ShowData(2, guid, false)
	CharScanner:ShowData(3, guid, false)
	CharScanner:ShowData(4, guid, false)
end

function CharScanner:DeleteCharacter(guid, id)
	if guid then -- delete one GUID
		NumScannedChars = NumScannedChars - 1
		if NumScannedChars == 0 then
			table.wipe(SCAN)
			GUIselectedGUID = nil
			CharScanner:DeleteButtonToggle(false)
			CharScanner:ScanFrameToggle(nil, true)
			CharScanner:ClearData()
			CharScanner:UpdateHistory()
		elseif NumScannedChars == 1 then
			SCAN[guid] = nil
			if guid == GUIselectedGUID then
				if id == 1 then
					CharScanner:PreShowSavedData(CharScanner_HistoryFrame_Button2.guid)
				else
					CharScanner:PreShowSavedData(CharScanner_HistoryFrame_Button1.guid)
				end
			end
			CharScanner:UpdateHistory()
		elseif NumScannedChars > 1 then
			SCAN[guid] = nil
			if guid == GUIselectedGUID then
				if id == 1 then
					CharScanner:PreShowSavedData(CharScanner_HistoryFrame_Button2.guid)
				else
					CharScanner:PreShowSavedData(_G["CharScanner_HistoryFrame_Button"..(id-1)].guid)
				end
			end
			CharScanner:UpdateHistory()
		end
	else -- delete all GUIDs
		table.wipe(SCAN)
		GUIselectedGUID = nil
		NumScannedChars = 0
		CharScanner:DeleteButtonToggle(false)
		CharScanner:ScanFrameToggle(nil, true)
		CharScanner:ClearData()
		CharScanner:UpdateHistory()
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:UpdateHistoryStatus()
	local text = string.format("%s %s", NumScannedChars, L["|4Char:Chars;"])
	if memusage then	-- FOR TEST/DEBUG ONLY
		collectgarbage("collect") -- uhh Am I evil?
		UpdateAddOnMemoryUsage()
		text = ceil(GetAddOnMemoryUsage("CharScanner")).."KB | "..text
	end
	CharScanner_HistoryFrame_Status_Text:SetText(text)
	
	local rtext = ""
	if GUI_history_Role["T"] then
		rtext = rtext.."|cff"..color.Brown.hex..L["T"]..": "..GUI_history_Role["T"].active.."|r |cff"..color.Grey.hex.."("..GUI_history_Role["T"].count..")|r"
	end
	if GUI_history_Role["H"] then
		if rtext ~= "" then
			rtext = rtext.."    "
		end
		rtext = rtext.."|cff"..color.Green.hex..L["H"]..": "..GUI_history_Role["H"].active.."|r |cff"..color.Grey.hex.."("..GUI_history_Role["H"].count..")|r"
	end
	if GUI_history_Role["D"] then
		if rtext ~= "" then
			rtext = rtext.."    "
		end
		rtext = rtext.."|cff"..color.Firebrick.hex..L["D"]..": "..GUI_history_Role["D"].active.."|r |cff"..color.Grey.hex.."("..GUI_history_Role["D"].count..")|r"
	end
	CharScanner_HistoryFrame_Status_RoleText:SetText(rtext)
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ScanGroup(delData)
	if inCombat then
		CharScanner_MainFrame_Status_Text:SetText(L["Scan is deactivated in combat!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		CharScanner_MainFrame_Status:Show()
		return
	end

	-- ScanGroup()-is-only-available-if-we-are-in-group,-but-a-paranoia-check-is-good! START
	local partymember = GetNumPartyMembers()
	local raidmember  = GetNumRaidMembers()
	local inParty
	local inRaid
	if partymember > 0 then
		inRaid  = false
		inParty = true
	end
	if raidmember > 1 then -- 1, not 0. Think the only situation where a 1-member-raid can exist is at the end of a battleground.
		inRaid  = true
		inParty = false
	end
	if not inRaid and not inParty then return end
	-- ScanGroup()-is-only-available-if-we-are-in-group,-but-a-paranoia-check-is-good! END

	if delData then -- maybe this need some aditional checks...we will see
		CharScanner:DeleteCharacter()
	end

	CharScanner:ProtectionFrameToggle(true)
	GroupScanTimerStart = GetTime()

	-- temp membertable START
	-- we build our own membertable. so we are independent from PARTY_MEMBERS_CHANGED/RAID_ROSTER_UPDATE changes
	-- raid roster changes between button click and end of groupscan are discarded!
	local t = {}
	local x = 1
	if inParty then
		for i = 1, partymember do
			local unit = "party"..i
			local guid = UnitGUID(unit)
			if guid then
				t[x] = {}
				t[x].guid = guid
				t[x].unit = unit
				x = x + 1
			end
		end
		-- add player to temp db
		t[x] = {}
		t[x].guid = playerGUID
		t[x].unit = "player"
	end
	if inRaid then
		for i = 1, raidmember do
			local unit = "raid"..i
			local guid = UnitGUID(unit)
			if guid then
				t[x] = {}
				t[x].guid = guid
				t[x].unit = unit
				x = x + 1
			end
		end
	end
	local numMember = #t
	-- temp membertable END

	GroupScanIsActive = true
	if CharScanner_MiniOptionsFrame then
		CharScanner_MiniOptionsFrame_GroupScanAdd:Disable()
		CharScanner_MiniOptionsFrame_GroupScanAddText:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
		CharScanner_MiniOptionsFrame_GroupScan:Disable()
		CharScanner_MiniOptionsFrame_GroupScanText:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
	end
	GroupScanNext = false
	local gsStart = GetTime()
	local currentScanNumber = 1

	local function scanNextMember()
		local guid = UnitGUID(t[currentScanNumber].unit)
		if guid and guid == t[currentScanNumber].guid then
			CharScanner_ProtectionFrame_ProgressText:SetText(currentScanNumber.."/"..numMember)
			CharScanner_ProtectionFrame_ProgressDone:SetWidth(currentScanNumber*sizeProtectionFrame_BarWidth/numMember)
			CharScanner:ScanUnit(t[currentScanNumber].unit)
			currentScanNumber = currentScanNumber + 1
		else
			CharScanner_ProtectionFrame_ProgressText:SetText(currentScanNumber.."/"..numMember)
			CharScanner_ProtectionFrame_ProgressDone:SetWidth(currentScanNumber*sizeProtectionFrame_BarWidth/numMember)
			currentScanNumber = currentScanNumber + 1
		end
	end

	local function gsTimeCheck()
		if inCombat then
			CharScanner:StopGroupScan()
			return
		end
		if not inPartyOrRaid then
			CharScanner:StopGroupScan()
			return
		end
		if (GetTime() - gsStart) >= CharScanner_Options.scanDuration or GroupScanNext then
			if currentScanNumber > numMember then
				CharScanner:StopGroupScan()
				return
			end
			-- next scan
			gsStart = GetTime()
			GroupScanNext = false
			scanNextMember()
		end
	end

	CharScanner_MainFrame_Headline_ButtonGroupScan:SetScript("OnUpdate", gsTimeCheck)
	scanNextMember()
end

function CharScanner:StopGroupScan()
	CharScanner_MainFrame_Headline_ButtonGroupScan:SetScript("OnUpdate", nil)
	CharScanner:Unregister_INSPECT_TALENT_READY()
	CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()

	GroupScanNext = false
	GroupScanIsActive = false
	if CharScanner_MiniOptionsFrame then
		CharScanner_MiniOptionsFrame_GroupScanAdd:Enable()
		CharScanner_MiniOptionsFrame_GroupScanAddText:SetTextColor(color.Yellow.rgba[1], color.Yellow.rgba[2], color.Yellow.rgba[3], color.Yellow.rgba[4])
		CharScanner_MiniOptionsFrame_GroupScan:Enable()
		CharScanner_MiniOptionsFrame_GroupScanText:SetTextColor(color.Yellow.rgba[1], color.Yellow.rgba[2], color.Yellow.rgba[3], color.Yellow.rgba[4])
	end
	if GUIselectedGUID and SCAN[GUIselectedGUID] then
		CharScanner:PreShowSavedData(GUIselectedGUID)
		CharScanner:UpdateHistory()
		CharScanner:CompareFrameShow() -- TEST
	else
		CharScanner:ClearData()
		CharScanner:UpdateHistory()
		if NumScannedChars == 0 then
			table.wipe(SCAN)
			CharScanner:DeleteButtonToggle(false)
			CharScanner:ScanFrameToggle(nil, true)
		end
	end
	CharScanner:ProtectionFrameToggle()

	if CharScanner_Options.chatoutput then
		local time = 0
		local TimeElapsed = math_floor((GetTime()-GroupScanTimerStart) * mult4 + 0.5) / mult4
		if type(TimeElapsed) == "number" and TimeElapsed > 0 then
			time = TimeElapsed
		end
		if time == 1 then
			time = string.format("%s %s", time, SECONDS)
		else
			time = time.." "..SECONDS
		end
		Print(L["GroupScan finished!"].." |cff"..color.Green.hex.."("..time..")|r")
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ScanUnit(unit)
	if inCombat then
		CharScanner_MainFrame_Status_Text:SetText(L["Scan is deactivated in combat!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end
	if AchievementFrame and AchievementFrame:IsVisible() then
		CharScanner_MainFrame_Status_Text:SetText(L["Please close AchievementFrame!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end
	if not UnitExists(unit) or not UnitIsPlayer(unit) then
		CharScanner_MainFrame_Status_Text:SetText(L["Target is NOT a player!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end
	if not CanInspect(unit) then
		CharScanner_MainFrame_Status_Text:SetText(L["Target is NOT in inspect range!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end

	-- --------------------------------------------
	-- DUMMY_FactionFix START
	-- --------------------------------------------
	local faction
	if UnitFactionGroup(unit) == "Horde" then
		DATA[3][1].achievement[54] = 4297
		faction = "Horde"
	else
		DATA[3][1].achievement[54] = 4298
		faction = "Alliance"
	end
	-- --------------------------------------------
	-- DUMMY_FactionFix END
	-- --------------------------------------------

	SingleScanTimerStart = GetTime()
	DataStep = 0
	LatestScanIsComplete = false

	if not GroupScanIsActive then
		CharScanner:ProtectionFrameToggle(true)
		CharScanner:ClearData()
	end
	CharScanner:ScanFrameToggle(true)
	CharScanner_MainFrame_Status_Text:SetText("..."..L["scan"].."...")
	CharScanner_MainFrame_Status_Text:SetTextColor(color.LightYellow.rgba[1], color.LightYellow.rgba[2], color.LightYellow.rgba[3], color.LightYellow.rgba[4])

	if UnitIsUnit("player", unit) then
		IsUnitInpected = false -- "Character" -> it's me
		unit = "player"
	else
		IsUnitInpected = true -- "Inspect"
	end

	SCANGUID = UnitGUID(unit)
	if not SCANGUID then Print("GUID ERROR!", SCANGUID) return end -- TEST
	SCANGUID = "_"..SCANGUID -- previewGUID -> if a succesfuly scanned unit is scanned again with an uncomplete scan the original (complete) scan is untouched
	CreateScanDataByGUID(SCANGUID)

	if not GroupScanIsActive then
		SingleScanIsActive = true
		if CharScanner_MiniOptionsFrame then
			CharScanner_MiniOptionsFrame_SingleScan:Disable()
			CharScanner_MiniOptionsFrame_SingleScanText:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
		end

		local function ssTimeCheck()
			if inCombat then
				CharScanner:StopSingleScan(true)
				return
			end
			if (GetTime() - SingleScanTimerStart) >= CharScanner_Options.scanDuration then
				CharScanner:StopSingleScan(true)
				return
			end
		end

		CharScanner_MainFrame_Headline_ButtonScan:SetScript("OnUpdate", ssTimeCheck)
	end

	local name, realm = UnitName(unit)
	if realm then
		name = name.."-"..realm
	end
	SCAN[SCANGUID].name = name
	SCAN[SCANGUID].level = UnitLevel(unit)
	SCAN[SCANGUID].race = UnitRace(unit)
	SCAN[SCANGUID].faction = faction
	SCAN[SCANGUID].class, SCAN[SCANGUID].classEN = UnitClass(unit)
	SCAN[SCANGUID].guildName, SCAN[SCANGUID].guildRankName, SCAN[SCANGUID].guildRankIndex = GetGuildInfo(unit)

	CharScanner:ShowData(1, SCANGUID, true)

	-- achievementFunctions workaround START
	-- hack to avoid error messages because of achievementFunctions.selectedCategory
	if achievementFunctions and achievementFunctions.selectedCategory then
		bugfix = achievementFunctions.selectedCategory
		achievementFunctions.selectedCategory = -1
	end
	-- achievementFunctions workaround END
	ClearAchievementComparisonUnit()
	SetAchievementComparisonUnit(unit)
	CharScanner:RegisterEvent("INSPECT_ACHIEVEMENT_READY")

	NotifyInspect(unit)
	if IsUnitInpected then
		CharScanner:ItemLevelCalc(unit, "Inspect")
	else
		CharScanner:ItemLevelCalc(unit, "Character")
	end
	CharScanner:RegisterEvent("INSPECT_TALENT_READY")
end

function CharScanner:StopSingleScan(statusupdate)
	CharScanner_MainFrame_Headline_ButtonScan:SetScript("OnUpdate", nil)
	CharScanner:Unregister_INSPECT_TALENT_READY()
	CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()
	SingleScanIsActive = false
	if CharScanner_MiniOptionsFrame then
		CharScanner_MiniOptionsFrame_SingleScan:Enable()
		CharScanner_MiniOptionsFrame_SingleScanText:SetTextColor(color.Yellow.rgba[1], color.Yellow.rgba[2], color.Yellow.rgba[3], color.Yellow.rgba[4])
	end
	CharScanner:ProtectionFrameToggle()
	if statusupdate then
		CharScanner:AchievementDetailHide()
		CharScanner_MainFrame_Status_Text:SetText(L["Scan incomplete!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ShowData(step, guid, scan)
	DataStep = DataStep + 1

	if not GroupScanIsActive then

		if step == 1 then

			for i = 1, GUI_history_maxNum do
				_G["CharScanner_HistoryFrame_Button"..i.."Arrow"]:Hide()
			end
			CharScanner_MainFrame_Headline_ButtonCollapse:Enable()
			CharScanner_MainFrame_Headline_ButtonCollapseBackground:SetTexture(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
			CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])

			CharScanner:DeleteButtonToggle(true)

			-- ClassIcon
			CharScanner_MainFrame_Base_ClassIcon:SetTexCoord(classes[SCAN[guid].classEN].icon[1], classes[SCAN[guid].classEN].icon[2], classes[SCAN[guid].classEN].icon[3], classes[SCAN[guid].classEN].icon[4])
			-- FactionIcon
			if SCAN[guid].faction == "Horde" then
				CharScanner_MainFrame_Base_FactionIcon:SetTexture(factionimgHorde)
			else
				CharScanner_MainFrame_Base_FactionIcon:SetTexture(factionimgAlliance)
			end
			CharScanner_MainFrame_Base_FactionIcon:SetTexCoord(0.046875, 0.609375, 0.03125, 0.59375) -- (3/64, 39/64, 2/64, 38/64)
			-- Name
			CharScanner_MainFrame_Base_Name:SetText(SCAN[guid].name)
			CharScanner_MainFrame_Base_Name:SetTextColor(classcolors[SCAN[guid].classEN].r, classcolors[SCAN[guid].classEN].g, classcolors[SCAN[guid].classEN].b, 1)
			-- Guild
			local guild = ""
			if SCAN[guid].guildName then
				guild = GUILD..": "..SCAN[guid].guildName
				if SCAN[guid].guildRankIndex and SCAN[guid].guildRankIndex < 5 then
					guild = guild.." | "..RANK..": "..guildRank[SCAN[guid].guildRankIndex+1]
				else
					guild = guild.." | "..RANK..": -"
				end
			else
				guild = GUILD..": - "
			end
			CharScanner_MainFrame_Base_Guild:SetText(guild)
			CharScanner_MainFrame_Base_Guild:SetTextColor(color.Level2.rgba[1], color.Level2.rgba[2], color.Level2.rgba[3], color.Level2.rgba[4])
			-- Level/Race/Class
			CharScanner_MainFrame_Base_Level:SetText(LEVEL.." "..SCAN[guid].level.." - "..SCAN[guid].race..", "..SCAN[guid].class)
			CharScanner_MainFrame_Base_Level:SetTextColor(classcolors[SCAN[guid].classEN].r, classcolors[SCAN[guid].classEN].g, classcolors[SCAN[guid].classEN].b, 1)

		elseif step == 2 then

			-- Equipment
			CharScanner_MainFrame_Equipment_Title:SetText(L["Equipment"]..":")

			local miss = ""
			if SCAN[guid].itemlevelMiss > 0 then
				miss = " - |cff"..color.Red.hex..string.format(L["%s %s not unsed"], SCAN[guid].itemlevelMiss, L["|4Slot:Slots;"]).."|r"
			end
			local txt = L["Avg"].." "..cnum(SCAN[guid].itemlevelAvg).." ("..L["min."].." "..SCAN[guid].itemlevelMin.." - "..L["max."].." "..SCAN[guid].itemlevelMax..")"..miss

			CharScanner_MainFrame_Equipment_Text1:SetText(txt)
			CharScanner_MainFrame_Equipment_Text2:SetText(SCAN[guid].itemlevelQuality)

			local w1 = ceil(CharScanner_MainFrame_Equipment_Title:GetStringWidth())
			local w2 = ceil(CharScanner_MainFrame_Equipment_Text2:GetStringWidth())
			local maxwidth = sizeGeneral.MainWidth - (sizeGeneral.Offset + w1 + sizeGeneral.Offset + sizeGeneral.Offset + 25) -- ~25px for buttonwidth

			if w2 >= maxwidth then
				CharScanner_MainFrame_Equipment_Text2:SetWidth(maxwidth)
				CharScanner_MainFrame_Equipment:SetHeight(sizeEquipmentFrame.BaseHeightXXL)
			else
				CharScanner_MainFrame_Equipment_Text2:SetWidth(maxwidth)
				CharScanner_MainFrame_Equipment:SetHeight(sizeEquipmentFrame.BaseHeight)
			end

			-- Gear
			for i = 1, #SCAN[guid].gear do
				for j = 1, #slotButton do

					if i == slotButton[j].id then
						local linkID = select(3, SCAN[guid].gear[i].iLink:find("item:(%d+)"))
						local itemTexture = GetItemIcon(linkID)
						if itemTexture then
							local iLink = SCAN[guid].gear[i].iLink
							_G["CharScanner_MainFrame_Gear_Slot"..j.."IconTexture"]:SetTexture(itemTexture)
							_G["CharScanner_MainFrame_Gear_Slot"..j.."Count"]:SetText("|cff"..itemQuality[SCAN[guid].gear[i].iRar+1].color..SCAN[guid].gear[i].iLvl.."|r")
							_G["CharScanner_MainFrame_Gear_Slot"..j.."Count"]:Show()
							if SCAN[guid].gear[i].iEnch ~= "0" then
								_G["CharScanner_MainFrame_Gear_Slot"..j.."EnchTexture"]:Show()
							--else
							--	_G["CharScanner_MainFrame_Gear_Slot"..j.."EnchTexture"]:Hide()
							end
							_G["CharScanner_MainFrame_Gear_Slot"..j]:SetScript("OnEnter",
								function(self)
									GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
									GameTooltip:SetHyperlink(iLink)
									self.UpdateTooltip = GearTooltipHandler
								end)
							_G["CharScanner_MainFrame_Gear_Slot"..j]:SetScript("OnLeave",
								function(self)
									GameTooltip:Hide()
									ResetCursor()
									self.comparing = false
								end)
							_G["CharScanner_MainFrame_Gear_Slot"..j]:SetScript("OnClick",
								function()
									HandleModifiedItemClick(iLink)
								end)
						end
						break
					end

				end
			end

			-- GearStats START
			--local statsText = ""
			--for key, value in pairs(SCAN[guid].gearStats) do
			--	statsText = statsText.."|cff"..color.LightYellow.hex.._G[key]..":|r "..(math_floor(value * mult2 + 0.5) / mult2).."\n"
				--TEST code for future GearStats updates START
				--if not CharScanner_Options.statsTest then CharScanner_Options.statsTest = {} end
				--if not CharScanner_Options.statsTest[key] then
				--	CharScanner_Options.statsTest[key] = _G[key]
				--end
				--TEST code for future GearStats updates END
			--end
			--statsText = string_sub(statsText, 0, -2) -- last '\n'
			--CharScanner_MainFrame_Equipment_StatsText:SetText(statsText)
			--local h = ceil(CharScanner_MainFrame_Equipment_StatsText:GetStringHeight())
			--local w = ceil(CharScanner_MainFrame_Equipment_StatsText:GetStringWidth())
			--CharScanner_MainFrame_Equipment_StatsFrame:SetHeight(h+20)
			--CharScanner_MainFrame_Equipment_StatsFrame:SetWidth(w+20)
			-- GearStats END

		elseif step == 3 then

			-- Talent
			local col
			CharScanner_MainFrame_Talent_Title:SetText(TALENTS..":")
			if SCAN[guid].talentGroups then
				if (SCAN[guid].level-9-SCAN[guid].talentGroup1pointsUnspent) > 0 then
					CharScanner_MainFrame_Talent_TalentGroup1Icon:SetTexture(SCAN[guid].talentGroup1icon)
					local text = SCAN[guid].talentGroup1name.." ("..SCAN[guid].talentGroup1pointsString..")"
					if SCAN[guid].talentGroup1pointsUnspent > 0 then
						if SCAN[guid].activeTalentGroup == 1 then
							col = "|cff"..color.Red.hex
						else
							col = "|cff"..color.Grey.hex
						end
						text = text.." - "..col..string.format(UNSPENT_TALENT_POINTS, SCAN[guid].talentGroup1pointsUnspent).."|r"
					end
					CharScanner_MainFrame_Talent_TalentGroup1Text:SetText(text)
				else
					local text = "("..SCAN[guid].talentGroup1pointsString..")"
					if SCAN[guid].talentGroup1pointsUnspent > 0 then
						if SCAN[guid].activeTalentGroup == 1 then
							col = "|cff"..color.Red.hex
						else
							col = "|cff"..color.Grey.hex
						end
						text = text.." - "..col..string.format(UNSPENT_TALENT_POINTS, SCAN[guid].talentGroup1pointsUnspent).."|r"
					end
					CharScanner_MainFrame_Talent_TalentGroup1Text:SetText(text)
				end
				if SCAN[guid].activeTalentGroup == 1 then
					CharScanner_MainFrame_Talent_TalentGroup1Text:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
				else
					CharScanner_MainFrame_Talent_TalentGroup1Text:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
				end
				if SCAN[guid].talentGroups > 1 then
					if (SCAN[guid].level-9-SCAN[guid].talentGroup2pointsUnspent) > 0 then
						CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(SCAN[guid].talentGroup2icon)
						local text = SCAN[guid].talentGroup2name.." ("..SCAN[guid].talentGroup2pointsString..")"
						if SCAN[guid].talentGroup2pointsUnspent > 0 then
							if SCAN[guid].activeTalentGroup == 2 then
								col = "|cff"..color.Red.hex
							else
								col = "|cff"..color.Grey.hex
							end
							text = text.." - "..col..string.format(UNSPENT_TALENT_POINTS, SCAN[guid].talentGroup2pointsUnspent).."|r"
						end
						CharScanner_MainFrame_Talent_TalentGroup2Text:SetText(text)
					else
						local text = "("..SCAN[guid].talentGroup2pointsString..")"
						if SCAN[guid].talentGroup2pointsUnspent > 0 then
							if SCAN[guid].activeTalentGroup == 2 then
								col = "|cff"..color.Red.hex
							else
								col = "|cff"..color.Grey.hex
							end
							text = text.." - "..col..string.format(UNSPENT_TALENT_POINTS, SCAN[guid].talentGroup2pointsUnspent).."|r"
						end
						CharScanner_MainFrame_Talent_TalentGroup2Text:SetText(text)
					end
					if SCAN[guid].activeTalentGroup == 2 then
						CharScanner_MainFrame_Talent_TalentGroup2Text:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
					else
						CharScanner_MainFrame_Talent_TalentGroup2Text:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
					end
				else
					CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(nil)
					CharScanner_MainFrame_Talent_TalentGroup2Text:SetText("")
				end
				
			else
				CharScanner_MainFrame_Talent_TalentGroup1Icon:SetTexture(nil)
				CharScanner_MainFrame_Talent_TalentGroup1Text:SetText(L["No Talents"])
				CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(nil)
				CharScanner_MainFrame_Talent_TalentGroup2Text:SetText("")
			end

			-- TalentTree
			if SCAN[guid].level > 9 then
				CharScanner:SwitchTalentGroup(guid, SCAN[guid].activeTalentGroup)
			else
				CharScanner:SwitchTalentGroup(guid, 0)
			end

		elseif step == 4 then

			local col

			CharScanner_MainFrame_Stats_TitleName:SetText("")
			CharScanner_MainFrame_Stats_TitleAchievement:SetText(L["Achievements"])
			CharScanner_MainFrame_Stats_TitleStatistic:SetText(L["Bosses"])
			CharScanner_MainFrame_Stats_TitleBossesTotal:SetText(L["Kills"])

			-- level 1 START
			for i = 1, #DATA do
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Name"]:SetText(L[DATA[i].name])
				if DATA[i].achievementTotal == SCAN[guid].DATA[i].achievementCount then
					col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
				end
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Achievement"]:SetText(col..SCAN[guid].DATA[i].achievementCount.."|r / "..DATA[i].achievementTotal)

				if DATA[i].statisticTotal == SCAN[guid].DATA[i].statisticBoss then
					col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
				end
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Statistic"]:SetText(col..SCAN[guid].DATA[i].statisticBoss.."|r / "..DATA[i].statisticTotal)
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."BossesTotal"]:SetText(SCAN[guid].DATA[i].statisticBossAll)
			--end
			-- level 1 END

			-- level 2 START
			--for i = 1, #DATA do
				if DATA[i].statistic then
					for j = 1, #DATA[i].statistic do
						if SCAN[guid].DATA[i].statisticBossEach[j] > 0 then
							col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
						end
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j.."Name"]:SetText(col..StatisticIDTable[DATA[i].statistic[j]].."|r")
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j.."BossesTotal"]:SetText(SCAN[guid].DATA[i].statisticBossEach[j])
					end
				else
					for j = 1, #DATA[i] do
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Name"]:SetText(L[DATA[i][j].name])
						if DATA[i][j].achievementTotal == SCAN[guid].DATA[i][j].achievementCount then
							col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
						end
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Achievement"]:SetText(col..SCAN[guid].DATA[i][j].achievementCount.."|r / "..DATA[i][j].achievementTotal)
						if DATA[i][j].statisticTotal == SCAN[guid].DATA[i][j].statisticBoss then
							col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
						end
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Statistic"]:SetText(col..SCAN[guid].DATA[i][j].statisticBoss.."|r / "..DATA[i][j].statisticTotal)
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."BossesTotal"]:SetText(SCAN[guid].DATA[i][j].statisticBossAll)
					--end
				--end
			--end
			-- level 2 END

			-- level 3 START
			--for i = 1, #DATA do
				--if not DATA[i].statistic then
					--for j = 1, #DATA[i] do

						if DATA[i][j].statistic then
							for k = 1, #DATA[i][j].statistic do
								if SCAN[guid].DATA[i][j].statisticBossEach[k] > 0 then
									col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
								end
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText(col..StatisticIDTable[DATA[i][j].statistic[k]].."|r")
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."BossesTotal"]:SetText(SCAN[guid].DATA[i][j].statisticBossEach[k])
							end
						else
							for k = 1, #DATA[i][j] do
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText(L[DATA[i][j][k].name])
								if DATA[i][j][k].achievementTotal == SCAN[guid].DATA[i][j][k].achievementCount then
									col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
								end
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Achievement"]:SetText(col..SCAN[guid].DATA[i][j][k].achievementCount.."|r / "..DATA[i][j][k].achievementTotal)
								if DATA[i][j][k].statisticTotal == SCAN[guid].DATA[i][j][k].statisticBoss then
									col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
								end
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Statistic"]:SetText(col..SCAN[guid].DATA[i][j][k].statisticBoss.."|r / "..DATA[i][j][k].statisticTotal)
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."BossesTotal"]:SetText(SCAN[guid].DATA[i][j][k].statisticBossAll)
							--end
						--end

					--end
				--end
			--end
			-- level 3 END

			-- level 4 START
			--for i = 1, #DATA do
				--if not DATA[i].statistic then
					--for j = 1, #DATA[i] do

						--if not DATA[i][j].statistic then
							--for k = 1, #DATA[i][j] do

								if DATA[i][j][k].statistic then
									for l = 1, #DATA[i][j][k].statistic do
										if SCAN[guid].DATA[i][j][k].statisticBossEach[l] > 0 then
											col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
										end
										_G["CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l.."Name"]:SetText(col..StatisticIDTable[DATA[i][j][k].statistic[l]].."|r")
										_G["CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l.."BossesTotal"]:SetText(SCAN[guid].DATA[i][j][k].statisticBossEach[l])
									end
								end

							end
						end

					end
				end
			end
			-- level 4 END

			-- summary
			CharScanner_MainFrame_SummaryButton1Achievement:SetText(SCAN[guid].statsAchievementCount.." / "..SCAN[guid].statsAchievementTotal)
			CharScanner_MainFrame_SummaryButton1Achievement:SetTextColor(color.Level1.rgba[1], color.Level1.rgba[2], color.Level1.rgba[3], color.Level1.rgba[4])
			local achievement = cnum(math_floor((SCAN[guid].statsAchievementCount/SCAN[guid].statsAchievementTotal*100) * mult2 + 0.5) / mult2)
			CharScanner_MainFrame_SummaryButton2Achievement:SetText(achievement.."%")
			CharScanner_MainFrame_SummaryButton2Achievement:SetTextColor(color.LightYellow.rgba[1], color.LightYellow.rgba[2], color.LightYellow.rgba[3], color.LightYellow.rgba[4])

			CharScanner_MainFrame_SummaryButton1Statistic:SetText(SCAN[guid].statsStatisticCount.." / "..SCAN[guid].statsStatisticTotal)
			CharScanner_MainFrame_SummaryButton1Statistic:SetTextColor(color.Level1.rgba[1], color.Level1.rgba[2], color.Level1.rgba[3], color.Level1.rgba[4])
			local statistic = cnum(math_floor((SCAN[guid].statsStatisticCount/SCAN[guid].statsStatisticTotal*100) * mult2 + 0.5) / mult2)
			CharScanner_MainFrame_SummaryButton2Statistic:SetText(statistic.."%")
			CharScanner_MainFrame_SummaryButton2Statistic:SetTextColor(color.LightYellow.rgba[1], color.LightYellow.rgba[2], color.LightYellow.rgba[3], color.LightYellow.rgba[4])

			CharScanner_MainFrame_SummaryButton1BossesTotal:SetText(SCAN[guid].statsStatisticBossKills)
			CharScanner_MainFrame_SummaryButton1BossesTotal:SetTextColor(color.Level1.rgba[1], color.Level1.rgba[2], color.Level1.rgba[3], color.Level1.rgba[4])

		end

		CharScanner_ProtectionFrame_ProgressText:SetText(L["Step"].." "..DataStep.."/4")
		CharScanner_ProtectionFrame_ProgressDone:SetWidth(DataStep*sizeProtectionFrame_BarWidth/4)

	end

	if DataStep == 4 then
		-- DevTools_Dump(SCAN[SCANGUID].gearStats) -- TEST

		if scan then
			-- scan complete: copy&delete preview scandata START
			local realGUID = string_sub(guid, 2)
			CreateScanDataByGUID(realGUID)
			SCAN[realGUID] = SCAN[guid] -- copy preview
			for k, v in pairs(SCAN) do
				if string_sub(k, 0, 1) == "_" then
					SCAN[k] = nil -- delete preview
				end
			end
			guid = realGUID
			-- scan complete: copy&delete preview scandata END

			-- talent group guid adjustment
			CharScanner_MainFrame_TalentTree_GroupSelection_Button1:SetScript("OnClick", function() CharScanner:SwitchTalentGroup(guid, 1) end)
			CharScanner_MainFrame_TalentTree_GroupSelection_Button2:SetScript("OnClick", function() CharScanner:SwitchTalentGroup(guid, 2) end)

			local time = 0
			local TimeElapsed = math_floor((GetTime()-SingleScanTimerStart) * mult4 + 0.5) / mult4
			if type(TimeElapsed) == "number" and TimeElapsed > 0 then
				time = TimeElapsed
			end
			if time == 1 then
				time = string.format("%s %s", time, SECONDS)
			else
				time = time.." "..SECONDS
			end
			SCAN[guid].scanTime = time
			CharScanner_MainFrame_Status_Text:SetText(time)
			CharScanner_MainFrame_Status_Text:SetTextColor(color.Green.rgba[1], color.Green.rgba[2], color.Green.rgba[3], color.Green.rgba[4])

			local hour, minute = GetGameTime()
			if hour < 10 then
				hour = "0"..hour
			end
			if minute < 10 then
				minute = "0"..minute
			end
			SCAN[guid].time = hour..":"..minute

			if (CharScanner_Options.chatoutput and not GroupScanIsActive) or
			   (CharScanner_Options.chatoutput and GroupScanIsActive and not CharScanner_Options.chatoutputGrp) then
				local col = ClassHexColor(SCAN[guid].classEN)
				local txt= "|cff"..col..SCAN[guid].name.."|r"				
				txt = txt.." |cff"..color.Level2.hex..LEVEL.." "..SCAN[guid].level.." - "..SCAN[guid].race..",|r"
				txt = txt.." |cff"..col..SCAN[guid].class.."|r"
				if SCAN[guid].guildName then
					txt = txt.." |cff"..color.Level2.hex.."("..GUILD..": "..SCAN[guid].guildName..")|r"
				end
				Print(txt)
			end
			GUIselectedGUID = guid
			LatestScanIsComplete = true
			if not GroupScanIsActive then
				CharScanner:UpdateHistory()
				CharScanner:StopSingleScan(false)
			end
		else
			GUIselectedGUID = guid
			LatestScanIsComplete = true
			CharScanner_MainFrame_Status_Text:SetText(L["SCANNED AT"]..": "..SCAN[guid].time.."  |cff"..color.Green.hex.."("..SCAN[guid].scanTime..")|r")
			CharScanner_MainFrame_Status_Text:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
		end

		if not GroupScanIsActive then

			if AchievementDetailFrameIsShown then
				CharScanner:UpdateAchievementDetail(GUI_achievementdetail_L1, GUI_achievementdetail_L2, GUI_achievementdetail_L3)
			end

			for i = 1, GUI_history_maxNum do
				if _G["CharScanner_HistoryFrame_Button"..i].guid == guid then
					_G["CharScanner_HistoryFrame_Button"..i.."Arrow"]:Show()
					break
				end
			end

			CharScanner_MainFrame_Headline_ButtonHistory:Enable()
			CharScanner_MainFrame_Headline_ButtonHistoryBorderColor:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])
			CharScanner_MainFrame_Headline_ButtonCompare:Enable()
			CharScanner_MainFrame_Headline_ButtonCompareBorderColor:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])
		end

		GroupScanNext = true

	end

end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ItemLevelCalc(unit, who)
	local SlotCount = 17
	if TwoHandWeaponCheck(unit, who) then
		SlotCount = 16
	end
	local ItemLevel = 0
	local ItemCount = 0
	local minItem = 0
	local maxItem = 0

	for i = 1, itemQualityNum do
		itemQuality[i].count = 0
	end

	for i = 1, slotsNum do
		local i1, iR = ItemData(unit, i)
		if i1 >= 0 then
			ItemLevel = ItemLevel + i1
			ItemCount = ItemCount + 1
		end
		if i1 > 0 then
			if minItem == 0 then
				minItem = i1
			end
			if maxItem == 0 then
				maxItem = i1
			end
			if i1 < minItem then
				minItem = i1
			end
			if i1 > maxItem then
				maxItem = i1
			end
		end
		if iR >= 0 then
			itemQuality[iR+1].count = itemQuality[iR+1].count + 1
		end
	end

	local quality = ""
	local x = 0
	for i = 1, itemQualityNum do
		if itemQuality[i].count > 0 then
			if x == 1 then
				quality = quality..", "
				x = 0
			end
			quality = quality.."|cff"..color.White.hex..itemQuality[i].count.."x|r |cff"..itemQuality[i].color..itemQuality[i].name.."|r"
			x = 1
		end
	end

	SCAN[SCANGUID].itemlevelMin     = minItem
	SCAN[SCANGUID].itemlevelMax     = maxItem
	SCAN[SCANGUID].itemlevelQuality = quality

	local ItemIQcount = SlotCount-ItemCount
	if ItemIQcount > 0 then
		SCAN[SCANGUID].itemlevelMiss = ItemIQcount
	end

	local ItemIQ = math_floor((ItemLevel/SlotCount) * mult2 + 0.5) / mult2
	if type(ItemIQ) == "number" and ItemIQ > 0 then
		SCAN[SCANGUID].itemlevelAvg = ItemIQ
	end

	CharScanner:ShowData(2, SCANGUID, true)
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:SwitchTalentGroup(guid, group)
	if SCAN[guid] and SCAN[guid].talents[group] then
		local activgroup = SCAN[guid].activeTalentGroup
		local button1name = "CharScanner_MainFrame_TalentTree_GroupSelection_Button1"
		local button2name = "CharScanner_MainFrame_TalentTree_GroupSelection_Button2"

		_G[button1name]:SetScript("OnClick", function() CharScanner:SwitchTalentGroup(guid, 1) end)
		_G[button2name]:SetScript("OnClick", function() CharScanner:SwitchTalentGroup(guid, 2) end)


		if SCAN[guid].talentGroups == 1 then
			_G[button1name]:Show()
			_G[button2name]:Hide()
		else
			_G[button1name]:Show()
			_G[button2name]:Show()
		end

		if activgroup == 1 then
			_G[button1name]:SetNormalFontObject(fonts.GameFontWhiteSmall.name)
			_G[button2name]:SetNormalFontObject(fonts.GameFontDisableSmall.name)
		else
			_G[button1name]:SetNormalFontObject(fonts.GameFontDisableSmall.name)
			_G[button2name]:SetNormalFontObject(fonts.GameFontWhiteSmall.name)
		end


		if group == 1 then
			_G[button1name]:Disable()
			if group == activgroup then
				_G[button1name]:SetDisabledFontObject(fonts.GameFontWhiteSmall.name)
			else
				_G[button1name]:SetDisabledFontObject(fonts.GameFontDisableSmall.name)
			end
			local border1 = _G[button1name.."Border"]
			border1:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
			border1:SetPoint("BOTTOMLEFT"  ,sizeGeneral.Border, 0)
			border1:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
			border1:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, 0)

			_G[button2name]:Enable()
			local border2 = _G[button2name.."Border"]
			border2:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
			border2:SetPoint("BOTTOMLEFT"  ,0, sizeGeneral.Border)
			border2:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
			border2:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, sizeGeneral.Border)
		else
			_G[button1name]:Enable()
			local border1 = _G[button1name.."Border"]
			border1:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
			border1:SetPoint("BOTTOMLEFT"  ,0, sizeGeneral.Border)
			border1:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
			border1:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, sizeGeneral.Border)

			_G[button2name]:Disable()
			if group == activgroup then
				_G[button2name]:SetDisabledFontObject(fonts.GameFontWhiteSmall.name)
			else
				_G[button2name]:SetDisabledFontObject(fonts.GameFontDisableSmall.name)
			end

			local border2 = _G[button2name.."Border"]
			border2:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
			border2:SetPoint("BOTTOMLEFT"  ,sizeGeneral.Border, 0)
			border2:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
			border2:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, 0)
		end

		local tier = 1
		local column = 1
		for tab = 1, myMAX_TALENT_TABS do

			local point
			if group == 1 then
				point = select(tab, string.match(SCAN[guid].talentGroup1pointsString, "(.+)/(.+)/(.+)"))
			else
				point = select(tab, string.match(SCAN[guid].talentGroup2pointsString, "(.+)/(.+)/(.+)"))
			end
			local bName   = _G["CharScanner_MainFrame_TalentTree"..tab.."Name"]
			local bPoints = _G["CharScanner_MainFrame_TalentTree"..tab.."Points"]

			bName:SetText(classes[SCAN[guid].classEN].talentTab[tab].tabName)
			bPoints:SetText(point)

			if group == activgroup then
				bName:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
				bPoints:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
			else
				bName:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
				bPoints:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
			end

			for talent = 1, myMAX_NUM_TALENTS do

				if talent == 1 then
					tier = 1
					column = 1
				elseif mod(talent, myMAX_COLUMN) == 1 then
					tier = tier + 1
					column = 1
				else
					column = column + 1
				end

				local buttonName = "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column

				if
					classes[SCAN[guid].classEN].talents[tab] and
					classes[SCAN[guid].classEN].talents[tab][tier] and
					classes[SCAN[guid].classEN].talents[tab][tier][column]
					then

					local spellRank    = SCAN[guid].talents[group][tab][tier][column]
					local spellIcon    = classes[SCAN[guid].classEN].talents[tab][tier][column].icon
					local spellID      = classes[SCAN[guid].classEN].talents[tab][tier][column].id
					local spellName    = classes[SCAN[guid].classEN].talents[tab][tier][column].name
					local talentString = "talent:"..spellID..":"..(spellRank-1)
					local talentLink   = "|cff4e96f7|H"..talentString.."|h["..spellName.."]|h|r"
					_G[buttonName.."IconTexture"]:SetTexture(spellIcon)
					_G[buttonName.."Rank"]:SetText(spellRank)
					_G[buttonName]:Show()
					_G[buttonName]:SetScript("OnEnter", function(self) GameTooltip:SetOwner(self, "ANCHOR_RIGHT") GameTooltip:SetHyperlink(talentString) end)
					_G[buttonName]:SetScript("OnLeave", function() GameTooltip:Hide() end)
					_G[buttonName]:SetScript("OnClick", function() HandleModifiedItemClick(talentLink) end)
					if spellRank == 0 then
						SetItemButtonDesaturated(_G[buttonName], 1, 0.65, 0.65, 0.65)
						_G[buttonName]:SetAlpha(0.5)
						_G[buttonName.."IconTexture"]:SetVertexColor(0.5, 0.5, 0.5)
						_G[buttonName.."RankBorder"]:Hide()
						_G[buttonName.."Rank"]:Hide()
					else
						SetItemButtonDesaturated(_G[buttonName], nil)
						_G[buttonName]:SetAlpha(1)
						_G[buttonName.."IconTexture"]:SetVertexColor(1, 1, 1)
						_G[buttonName.."RankBorder"]:Show()
						_G[buttonName.."Rank"]:Show()
					end
				else
					_G[buttonName]:Hide()
					_G[buttonName]:SetScript("OnEnter", nil)
					_G[buttonName]:SetScript("OnLeave", nil)
				end

			end
		end
	else
		for tab = 1, myMAX_TALENT_TABS do
			_G["CharScanner_MainFrame_TalentTree"..tab.."Name"]:SetText("")
			_G["CharScanner_MainFrame_TalentTree"..tab.."Points"]:SetText("")
		end
		CharScanner_MainFrame_TalentTree_GroupSelection_Button1:Hide()
		CharScanner_MainFrame_TalentTree_GroupSelection_Button1:SetScript("OnClick", nil)
		CharScanner_MainFrame_TalentTree_GroupSelection_Button2:Hide()
		CharScanner_MainFrame_TalentTree_GroupSelection_Button2:SetScript("OnClick", nil)
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ClearData()
	-- Base
	CharScanner_MainFrame_Base_ClassIcon:SetTexCoord(0, 0, 0, 0)
	CharScanner_MainFrame_Base_FactionIcon:SetTexCoord(0, 0, 0, 0)
	CharScanner_MainFrame_Base_Name:SetText("")
	CharScanner_MainFrame_Base_Guild:SetText("")
	CharScanner_MainFrame_Base_Level:SetText("")

	-- Equipment
	CharScanner_MainFrame_Equipment_Title:SetText("")
	CharScanner_MainFrame_Equipment_Text1:SetText("")
	CharScanner_MainFrame_Equipment_Text2:SetText("")

	-- Gear
	for i = 1, #slotButton do
		_G["CharScanner_MainFrame_Gear_Slot"..i.."IconTexture"]:SetTexture(slotButton[i].texture)
		_G["CharScanner_MainFrame_Gear_Slot"..i.."Count"]:SetText("")
		_G["CharScanner_MainFrame_Gear_Slot"..i.."Count"]:Hide()
		_G["CharScanner_MainFrame_Gear_Slot"..i.."EnchTexture"]:Hide()
		_G["CharScanner_MainFrame_Gear_Slot"..i]:SetScript("OnEnter", nil)
		_G["CharScanner_MainFrame_Gear_Slot"..i]:SetScript("OnLeave", nil)
	end

	-- GearStats
	--CharScanner_MainFrame_Equipment_StatsText:SetText("")
	--CharScanner_MainFrame_Equipment_StatsFrame:SetHeight(1)
	--CharScanner_MainFrame_Equipment_StatsFrame:SetWidth(1)

	-- Talent
	CharScanner_MainFrame_Talent_Title:SetText("")
	CharScanner_MainFrame_Talent_TalentGroup1Icon:SetTexture(nil)
	CharScanner_MainFrame_Talent_TalentGroup1Text:SetText("")
	CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(nil)
	CharScanner_MainFrame_Talent_TalentGroup2Text:SetText("")

	-- TalentTree
	CharScanner:SwitchTalentGroup(0, 0)

	local tier = 1
	local column = 1
	for tab = 1, myMAX_TALENT_TABS do
		for talent = 1, myMAX_NUM_TALENTS do
			if talent == 1 then
				tier = 1
				column = 1
			elseif mod(talent, myMAX_COLUMN) == 1 then
				tier = tier + 1
				column = 1
			else
				column = column + 1
			end
			_G["CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column]:Hide()
		end
	end

	-- Stats
	CharScanner_MainFrame_Stats_TitleName:SetText("")
	CharScanner_MainFrame_Stats_TitleAchievement:SetText("")
	CharScanner_MainFrame_Stats_TitleStatistic:SetText("")
	CharScanner_MainFrame_Stats_TitleBossesTotal:SetText("")

	-- level 1 START
	for i = 1, #DATA do
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Name"]:SetText("")
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Achievement"]:SetText("")
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Statistic"]:SetText("")
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."BossesTotal"]:SetText("")
	--end
	-- level 1 END

	-- level 2 START
	--for i = 1, #DATA do
		if DATA[i].statistic then
			for j = 1, #DATA[i].statistic do
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j.."Name"]:SetText("")
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j.."BossesTotal"]:SetText("")
			end
		else
			for j = 1, #DATA[i] do
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Name"]:SetText("")
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Achievement"]:SetText("")
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Statistic"]:SetText("")
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."BossesTotal"]:SetText("")
			--end
		--end
	--end
	-- level 2 END

	-- level 3 START
	--for i = 1, #DATA do
		--if not DATA[i].statistic then
			--for j = 1, #DATA[i] do

				if DATA[i][j].statistic then
					for k = 1, #DATA[i][j].statistic do
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText("")
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."BossesTotal"]:SetText("")
					end
				else
					for k = 1, #DATA[i][j] do
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText("")
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Achievement"]:SetText("")
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Statistic"]:SetText("")
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."BossesTotal"]:SetText("")
					--end
				--end

			--end
		--end
	--end
	-- level 3 END

	-- level 4 START
	--for i = 1, #DATA do
		--if not DATA[i].statistic then
			--for j = 1, #DATA[i] do

				--if not DATA[i][j].statistic then
					--for k = 1, #DATA[i][j] do

						if DATA[i][j][k].statistic then
							for l = 1, #DATA[i][j][k].statistic do
								_G["CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l.."Name"]:SetText("")
								_G["CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l.."BossesTotal"]:SetText("")
							end
						end

					end
				end

			end
		end
	end
	-- level 4 END

	-- summary
	CharScanner_MainFrame_SummaryButton1Achievement:SetText("")
	CharScanner_MainFrame_SummaryButton2Achievement:SetText("")

	CharScanner_MainFrame_SummaryButton1Statistic:SetText("")
	CharScanner_MainFrame_SummaryButton2Statistic:SetText("")

	CharScanner_MainFrame_SummaryButton1BossesTotal:SetText("")
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:Unregister_INSPECT_TALENT_READY()
	CharScanner:UnregisterEvent("INSPECT_TALENT_READY")
	if not InspectFrame.unit then
		ClearInspectPlayer()
	end
end

function CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()
	CharScanner:UnregisterEvent("INSPECT_ACHIEVEMENT_READY")
	-- achievementFunctions workaround START
	-- hack to avoid error messages because of achievementFunctions.selectedCategory
	if achievementFunctions and achievementFunctions.selectedCategory then
		achievementFunctions.selectedCategory = bugfix or "summary"
	end
	-- achievementFunctions workaround END
	ClearAchievementComparisonUnit()
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
local function OnEvent(self, event, ...)
	if event == "PLAYER_REGEN_DISABLED" then
		inCombat = true
		CharScanner_ScanButtonFrame:Hide()
	elseif event == "PLAYER_REGEN_ENABLED" then
		inCombat = false
		if CharScanner_Options.scanButton then
			if UnitIsPlayer("target") then
				CharScanner_ScanButtonFrame:Show()
			end
		end
		if inCombatGrpChk then
			if GetNumRaidMembers() > 1 or GetNumPartyMembers() > 0 then
				inPartyOrRaid = true
			else
				inPartyOrRaid = false
			end
			CharScanner:GroupScanToggle()
			inCombatGrpChk = false
		end
	elseif event == "PLAYER_TARGET_CHANGED" then
		if inCombat then return end
		if CharScanner_Options.scanButton then
			if UnitIsPlayer("target") then
				CharScanner_ScanButtonFrame:Show()
			else
				CharScanner_ScanButtonFrame:Hide()
			end
		end
	elseif event == "INSPECT_TALENT_READY" then
		--if not SCANGUID then Print("INSPECT_TALENT_READY: SCANGUID ERROR!", SCANGUID) end -- TEST
		local talentGroup1pointsTotal = nil
		local talentGroup2pointsTotal = nil
		SCAN[SCANGUID].talentGroup1name = nil
		SCAN[SCANGUID].talentGroup1role = nil
		SCAN[SCANGUID].talentGroup1pointsString = nil
		SCAN[SCANGUID].talentGroup2name = nil
		SCAN[SCANGUID].talentGroup2role = nil
		SCAN[SCANGUID].talentGroup2pointsString = nil
		SCAN[SCANGUID].talentGroups = GetNumTalentGroups(IsUnitInpected, nil)
		SCAN[SCANGUID].activeTalentGroup = GetActiveTalentGroup(IsUnitInpected, nil)

		-- save talent information for each class START
		-- this data is saved to the classes table if a class is scanned the first time
		if not classes[SCAN[SCANGUID].classEN].talentsSaved then
			for tab = 1, myMAX_TALENT_TABS do

				local tabName, _, _, background = GetTalentTabInfo(tab, IsUnitInpected, nil)
				if not classes[SCAN[SCANGUID].classEN].talentTab then classes[SCAN[SCANGUID].classEN].talentTab = {} end
				if not classes[SCAN[SCANGUID].classEN].talentTab[tab] then
					classes[SCAN[SCANGUID].classEN].talentTab[tab] = {}
					classes[SCAN[SCANGUID].classEN].talentTab[tab].tabName = tabName
					--classes[SCAN[SCANGUID].classEN].talentTab[tab].background = background
				end

				for talent=1, myMAX_NUM_TALENTS do
					local name, iconTexture, tier, column, rank = GetTalentInfo(tab, talent, IsUnitInpected, nil, nil)
					local spellLink = GetTalentLink(tab, talent, IsUnitInpected, nil, nil, nil)

					if iconTexture and spellLink then
						local spellID = select(3, spellLink:find("talent:(%d+)"))

						if not classes[SCAN[SCANGUID].classEN].talents then classes[SCAN[SCANGUID].classEN].talents = {} end
						if not classes[SCAN[SCANGUID].classEN].talents[tab] then classes[SCAN[SCANGUID].classEN].talents[tab] = {} end
						if not classes[SCAN[SCANGUID].classEN].talents[tab][tier] then classes[SCAN[SCANGUID].classEN].talents[tab][tier] = {} end
						if not classes[SCAN[SCANGUID].classEN].talents[tab][tier][column] then classes[SCAN[SCANGUID].classEN].talents[tab][tier][column] = {} end

						classes[SCAN[SCANGUID].classEN].talents[tab][tier][column].id   = spellID
						classes[SCAN[SCANGUID].classEN].talents[tab][tier][column].icon = iconTexture
						classes[SCAN[SCANGUID].classEN].talents[tab][tier][column].name = name or "CharScanner_TalentScan_ERROR"
					end
				end
			end
			classes[SCAN[SCANGUID].classEN].talentsSaved = true
		end
		-- save talent information for each class END

		for group = 1, SCAN[SCANGUID].talentGroups do
			if not SCAN[SCANGUID].talents[group] then SCAN[SCANGUID].talents[group] = {} end
			if group == 1 then
				SCAN[SCANGUID].talentGroup1pointsUnspent = GetUnspentTalentPoints(IsUnitInpected, nil, group)
			else
				SCAN[SCANGUID].talentGroup2pointsUnspent = GetUnspentTalentPoints(IsUnitInpected, nil, group)
			end
			local points = 0
			for tab = 1, myMAX_TALENT_TABS do
				if not SCAN[SCANGUID].talents[group][tab] then SCAN[SCANGUID].talents[group][tab] = {} end

				-- --------------------
				for talent=1, myMAX_NUM_TALENTS do
					local _, iconTexture, tier, column, rank = GetTalentInfo(tab, talent, IsUnitInpected, nil, group)
					if not SCAN[SCANGUID].talents[group][tab][tier] then SCAN[SCANGUID].talents[group][tab][tier] = {} end
					if iconTexture then
						SCAN[SCANGUID].talents[group][tab][tier][column] = rank
					end
				end
				-- --------------------

				-- --------------------
				local name, icon, pointsSpent = GetTalentTabInfo(tab, IsUnitInpected, nil, group)
				if name and icon and pointsSpent then
					if tab == 1 then
						if group == 1 then
							points = pointsSpent
							talentGroup1pointsTotal = pointsSpent
							SCAN[SCANGUID].talentGroup1pointsString = pointsSpent
							SCAN[SCANGUID].talentGroup1name         = name
							SCAN[SCANGUID].talentGroup1role         = tab
							SCAN[SCANGUID].talentGroup1icon         = icon
						else
							points = pointsSpent
							talentGroup2pointsTotal = pointsSpent
							SCAN[SCANGUID].talentGroup2pointsString = pointsSpent
							SCAN[SCANGUID].talentGroup2name         = name
							SCAN[SCANGUID].talentGroup2role         = tab
							SCAN[SCANGUID].talentGroup2icon         = icon
						end
					else
						if group == 1 then
							talentGroup1pointsTotal = talentGroup1pointsTotal + pointsSpent
							SCAN[SCANGUID].talentGroup1pointsString = SCAN[SCANGUID].talentGroup1pointsString.."/"..pointsSpent
							if pointsSpent > points then
								points = pointsSpent
								SCAN[SCANGUID].talentGroup1name = name
								SCAN[SCANGUID].talentGroup1role = tab
								SCAN[SCANGUID].talentGroup1icon = icon
							end
						else
							talentGroup2pointsTotal = talentGroup2pointsTotal + pointsSpent
							SCAN[SCANGUID].talentGroup2pointsString = SCAN[SCANGUID].talentGroup2pointsString.."/"..pointsSpent
							if pointsSpent > points then
								points = pointsSpent
								SCAN[SCANGUID].talentGroup2name = name
								SCAN[SCANGUID].talentGroup2role = tab
								SCAN[SCANGUID].talentGroup2icon = icon
							end
						end
					end
				end
				-- --------------------

			end
		end

		-- opposite faction level (lvl ??/-1) check START
		if SCAN[SCANGUID].level < 0 then
			local realLevel = -1
			local totalPoints1 = talentGroup1pointsTotal + SCAN[SCANGUID].talentGroup1pointsUnspent
			local totalPoints2 = 0

			if SCAN[SCANGUID].talentGroups > 1 then
				totalPoints2 = talentGroup2pointsTotal + SCAN[SCANGUID].talentGroup2pointsUnspent
			end

			if totalPoints2 > totalPoints1 then
				realLevel = totalPoints2 + 9
			else
				realLevel = totalPoints1 + 9
			end
	
			-- if we are level 1 we can get the level from chars <= 8
			-- if we can not see a level and totalPoints1 = 0 and totalPoints2 = 0 the scanned char must be level 9
			if totalPoints1 == 0 and totalPoints2 == 0 then
				realLevel = 9
			end
	
			SCAN[SCANGUID].level = realLevel
			CharScanner_MainFrame_Base_Level:SetText(LEVEL.." "..SCAN[SCANGUID].level.." - "..SCAN[SCANGUID].race..", "..SCAN[SCANGUID].class)
		end
		-- opposite faction level (lvl ??/-1) check END

		-- low level (level <= 9) check START
		if talentGroup1pointsTotal                  and talentGroup1pointsTotal                  == 0 and not talentGroup2pointsTotal and
		   SCAN[SCANGUID].talentGroup1pointsUnspent and SCAN[SCANGUID].talentGroup1pointsUnspent == 0 and not SCAN[SCANGUID].talentGroup2pointsUnspent
		   then
			SCAN[SCANGUID].talentGroup1name = nil
			SCAN[SCANGUID].talentGroup1role = nil
			SCAN[SCANGUID].talentGroup1pointsString = nil
			SCAN[SCANGUID].talentGroup2name = nil
			SCAN[SCANGUID].talentGroup2role = nil
			SCAN[SCANGUID].talentGroup2pointsString = nil
			SCAN[SCANGUID].talentGroups = nil
			SCAN[SCANGUID].activeTalentGroup = nil
		end
		-- low level (level <= 9) check END

		CharScanner:ShowData(3, SCANGUID, true)
		CharScanner:Unregister_INSPECT_TALENT_READY()
	elseif event == "INSPECT_ACHIEVEMENT_READY" then
		--if not SCANGUID then Print("INSPECT_TALENT_READY: SCANGUID ERROR!", SCANGUID) end -- TEST
		-- achievement START
		SCAN[SCANGUID].DATA.achievementID = {}
		for i = 1, #DATA do
			SCAN[SCANGUID].DATA[i] = {}
			SCAN[SCANGUID].DATA[i].achievementCount = 0
			if DATA[i].achievement then
				for key, value in pairs(DATA[i].achievement) do
					if GetAchievementComparisonInfo(value, 1) then
						SCAN[SCANGUID].DATA[i].achievementCount = SCAN[SCANGUID].DATA[i].achievementCount + 1
						SCAN[SCANGUID].DATA.achievementID[value] = 1
					end
				end
				SCAN[SCANGUID].statsAchievementTotal = SCAN[SCANGUID].statsAchievementTotal + DATA[i].achievementTotal
				SCAN[SCANGUID].statsAchievementCount = SCAN[SCANGUID].statsAchievementCount + SCAN[SCANGUID].DATA[i].achievementCount
			else

				for j = 1, #DATA[i] do
					SCAN[SCANGUID].DATA[i][j] = {}
					SCAN[SCANGUID].DATA[i][j].achievementCount = 0
					if DATA[i][j].achievement then
						for key, value in pairs(DATA[i][j].achievement) do
							if GetAchievementComparisonInfo(value, 1) then
								SCAN[SCANGUID].DATA[i].achievementCount = SCAN[SCANGUID].DATA[i].achievementCount + 1
								SCAN[SCANGUID].DATA[i][j].achievementCount = SCAN[SCANGUID].DATA[i][j].achievementCount + 1
								SCAN[SCANGUID].DATA.achievementID[value] = 1
							end
						end
						SCAN[SCANGUID].statsAchievementTotal = SCAN[SCANGUID].statsAchievementTotal + DATA[i][j].achievementTotal
						SCAN[SCANGUID].statsAchievementCount = SCAN[SCANGUID].statsAchievementCount + SCAN[SCANGUID].DATA[i][j].achievementCount
					else

						for k = 1, #DATA[i][j] do
							SCAN[SCANGUID].DATA[i][j][k] = {}
							SCAN[SCANGUID].DATA[i][j][k].achievementCount = 0
							if DATA[i][j][k].achievement then
								for key, value in pairs(DATA[i][j][k].achievement) do
									if GetAchievementComparisonInfo(value, 1) then
										SCAN[SCANGUID].DATA[i].achievementCount = SCAN[SCANGUID].DATA[i].achievementCount + 1
										SCAN[SCANGUID].DATA[i][j].achievementCount = SCAN[SCANGUID].DATA[i][j].achievementCount + 1
										SCAN[SCANGUID].DATA[i][j][k].achievementCount = SCAN[SCANGUID].DATA[i][j][k].achievementCount + 1
										SCAN[SCANGUID].DATA.achievementID[value] = 1
									end
								end
								SCAN[SCANGUID].statsAchievementTotal = SCAN[SCANGUID].statsAchievementTotal + DATA[i][j][k].achievementTotal
								SCAN[SCANGUID].statsAchievementCount = SCAN[SCANGUID].statsAchievementCount + SCAN[SCANGUID].DATA[i][j][k].achievementCount
							--else
							end
						end

					end

				end

			end
		end
		-- achievement END

		-- statistic START
		for i = 1, #DATA do
			SCAN[SCANGUID].DATA[i].statisticBoss = 0
			SCAN[SCANGUID].DATA[i].statisticBossAll = 0
			if DATA[i].statistic then
				SCAN[SCANGUID].DATA[i].statisticBossEach = {}
				for key, value in pairs(DATA[i].statistic) do
					local q = BossStatistic(value, IsUnitInpected)
					SCAN[SCANGUID].DATA[i].statisticBossEach[key] = q
					if q > 0 then
						SCAN[SCANGUID].DATA[i].statisticBoss = SCAN[SCANGUID].DATA[i].statisticBoss + 1
						SCAN[SCANGUID].DATA[i].statisticBossAll = SCAN[SCANGUID].DATA[i].statisticBossAll + q
					end
				end
				SCAN[SCANGUID].statsStatisticTotal = SCAN[SCANGUID].statsStatisticTotal + DATA[i].statisticTotal
				SCAN[SCANGUID].statsStatisticCount = SCAN[SCANGUID].statsStatisticCount + SCAN[SCANGUID].DATA[i].statisticBoss
				SCAN[SCANGUID].statsStatisticBossKills = SCAN[SCANGUID].statsStatisticBossKills + SCAN[SCANGUID].DATA[i].statisticBossAll
			else

				for j = 1, #DATA[i] do
					SCAN[SCANGUID].DATA[i][j].statisticBoss = 0
					SCAN[SCANGUID].DATA[i][j].statisticBossAll = 0
					if DATA[i][j].statistic then
						SCAN[SCANGUID].DATA[i][j].statisticBossEach = {}
						for key, value in pairs(DATA[i][j].statistic) do
							local q = BossStatistic(value, IsUnitInpected)
							SCAN[SCANGUID].DATA[i][j].statisticBossEach[key] = q
							if q > 0 then
								SCAN[SCANGUID].DATA[i].statisticBoss = SCAN[SCANGUID].DATA[i].statisticBoss + 1
								SCAN[SCANGUID].DATA[i].statisticBossAll = SCAN[SCANGUID].DATA[i].statisticBossAll + q
								SCAN[SCANGUID].DATA[i][j].statisticBoss = SCAN[SCANGUID].DATA[i][j].statisticBoss + 1
								SCAN[SCANGUID].DATA[i][j].statisticBossAll = SCAN[SCANGUID].DATA[i][j].statisticBossAll + q
							end
						end
						SCAN[SCANGUID].statsStatisticTotal = SCAN[SCANGUID].statsStatisticTotal + DATA[i][j].statisticTotal
						SCAN[SCANGUID].statsStatisticCount = SCAN[SCANGUID].statsStatisticCount + SCAN[SCANGUID].DATA[i][j].statisticBoss
						SCAN[SCANGUID].statsStatisticBossKills = SCAN[SCANGUID].statsStatisticBossKills + SCAN[SCANGUID].DATA[i][j].statisticBossAll
					else

						for k = 1, #DATA[i][j] do
							SCAN[SCANGUID].DATA[i][j][k].statisticBoss = 0
							SCAN[SCANGUID].DATA[i][j][k].statisticBossAll = 0
							if DATA[i][j][k].statistic then
								SCAN[SCANGUID].DATA[i][j][k].statisticBossEach = {}
								for key, value in pairs(DATA[i][j][k].statistic) do
									local q = BossStatistic(value, IsUnitInpected)
									SCAN[SCANGUID].DATA[i][j][k].statisticBossEach[key] = q
									if q > 0 then
										SCAN[SCANGUID].DATA[i].statisticBoss = SCAN[SCANGUID].DATA[i].statisticBoss + 1
										SCAN[SCANGUID].DATA[i].statisticBossAll = SCAN[SCANGUID].DATA[i].statisticBossAll + q
										SCAN[SCANGUID].DATA[i][j].statisticBoss = SCAN[SCANGUID].DATA[i][j].statisticBoss + 1
										SCAN[SCANGUID].DATA[i][j].statisticBossAll = SCAN[SCANGUID].DATA[i][j].statisticBossAll + q
										SCAN[SCANGUID].DATA[i][j][k].statisticBoss = SCAN[SCANGUID].DATA[i][j][k].statisticBoss + 1
										SCAN[SCANGUID].DATA[i][j][k].statisticBossAll = SCAN[SCANGUID].DATA[i][j][k].statisticBossAll + q
									end
								end
								SCAN[SCANGUID].statsStatisticTotal = SCAN[SCANGUID].statsStatisticTotal + DATA[i][j][k].statisticTotal
								SCAN[SCANGUID].statsStatisticCount = SCAN[SCANGUID].statsStatisticCount + SCAN[SCANGUID].DATA[i][j][k].statisticBoss
								SCAN[SCANGUID].statsStatisticBossKills = SCAN[SCANGUID].statsStatisticBossKills + SCAN[SCANGUID].DATA[i][j][k].statisticBossAll
							--else
							end
						end

					end

				end

			end
		end
		-- statistic END

		CharScanner:ShowData(4, SCANGUID, true)
		CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()
	elseif event == "RAID_ROSTER_UPDATE" then
		if inCombat then
			inCombatGrpChk = true
			return
		end
		local raidmembers = GetNumRaidMembers()
		if rosterhack ~= raidmembers then -- avoid execute this block if raidmembers are just moved around
			if raidmembers > 1 then -- 1, not 0. Think the only situation where a 1-member-raid can exist is at the end of a battleground.
				inPartyOrRaid = true
			else
				inPartyOrRaid = false
			end	
			rosterhack = raidmembers
			CharScanner:GroupScanToggle()
		end
	elseif event == "PARTY_MEMBERS_CHANGED" then
		if inCombat then
			inCombatGrpChk = true
			return
		end
		if GetNumRaidMembers() > 0 then return end -- avoid execute this block if we are in raid
		if GetNumPartyMembers() > 0 then
			inPartyOrRaid = true
		else
			inPartyOrRaid = false
		end
		CharScanner:GroupScanToggle()
	elseif event == "PLAYER_ENTERING_WORLD" then
		if GroupScanIsActive then
			CharScanner:StopGroupScan()
		end
		if SingleScanIsActive then
			CharScanner:StopSingleScan(true)
		end
		if not CharScanner_Options.StartMessage then
			Print(L["Please read CharScanner.lua for more information."])
			Print(L["Scan access via SCAN-Button, Minimap-Button, LDB, Blizzard Interface Options or '/scan' resp. '/scan options'."])
			CharScanner_Options.StartMessage = true
		end
	elseif event == "PLAYER_LOGIN" then
		LoadAddOn("Blizzard_InspectUI") -- we need this
		--LoadAddOn("Blizzard_DebugTools") -- FOR TEST/DEBUG ONLY - DevTools_Dump
		playerGUID = UnitGUID("player")
		BuildIDTable()
		CharScanner:InitOptions()
		CharScanner:CreateInterfaceOptions()
		CharScanner:LDBcheck()
		CharScanner:CreateMinimapButton()
		CharScanner:CreateMainFrame()
		CharScanner:InitializeFrame()
		table.insert(UISpecialFrames, "CharScanner_MainFrame")
		CharScanner:UnregisterEvent("PLAYER_LOGIN")
	end
end
-- ---------------------------------------------------------------------------------------------------------------------
CharScanner:RegisterEvent("PLAYER_LOGIN")
CharScanner:RegisterEvent("PLAYER_ENTERING_WORLD")
CharScanner:RegisterEvent("PLAYER_REGEN_DISABLED")
CharScanner:RegisterEvent("PLAYER_REGEN_ENABLED")
CharScanner:RegisterEvent("PLAYER_TARGET_CHANGED")
CharScanner:RegisterEvent("RAID_ROSTER_UPDATE")
CharScanner:RegisterEvent("PARTY_MEMBERS_CHANGED")
CharScanner:SetScript("OnEvent", OnEvent)